/******************************************************************************\
|*                                                                            *|
|*                    Neurotechnologija Finger Record 1.1                     *|
|*                                                                            *|
|* NFRecord.h                                                                 *|
|* Header file for NFRecord module                                            *|
|*                                                                            *|
|* Copyright (C) 2005-2006 Neurotechnologija                                  *|
|*                                                                            *|
\******************************************************************************/

#ifndef NF_RECORD_H_INCLUDED
#define NF_RECORD_H_INCLUDED

#include <NCore.h>

#ifdef N_CPP
extern "C"
{
#endif

typedef enum NFPosition_
{
	nfpUnknown = 0,
	nfpRightThumb = 1,
	nfpRightIndex = 2,
	nfpRightMiddle = 3,
	nfpRightRing = 4,
	nfpRightLittle = 5,
	nfpLeftThumb = 6,
	nfpLeftIndex = 7,
	nfpLeftMiddle = 8,
	nfpLeftRing = 9,
	nfpLeftLittle = 10
} NFPosition;

typedef enum NFImpressionType_
{
	nfitLiveScanPlain = 0,
	nfitLiveScanRolled = 1,
	nfitNonliveScanPlain = 2,
	nfitNonliveScanRolled = 3,
	nfitLatentImpression = 4,
	nfitLatentTracing = 5,
	nfitLatentPhoto = 6,
	nfitLatentLift = 7,
	nfitSwipe = 8,
	nfitLiveScanContactless = 9
} NFImpressionType;

typedef enum NFPatternClass_
{
	nfpcUnknown = 0,
	nfpcPlainArch = 1,
	nfpcTentedArch = 2,
	nfpcRadialLoop = 3,
	nfpcUlnarLoop = 4,
	nfpcPlainWhorl = 5,
	nfpcCentralPocketLoop = 6,
	nfpcDoubleLoop = 7,
	nfpcAccidentalWhorl = 8,
	nfpcWhorl = 9,
	nfpcRightSlantLoop = 10,
	nfpcLeftSlantLoop = 11,
	nfpcScar = 12,
	nfpcAmputation = 15
} NFPatternClass;

#define NFR_RESOLUTION 500

#define NFR_MAX_DIMENSION 2047
#define NFR_BLOCK_SIZE 16
#define NFR_MAX_BLOCKED_ORIENTS_DIMENSION 128

#define NFR_MAX_MINUTIA_COUNT 255
#define NFR_MAX_CORE_COUNT 15
#define NFR_MAX_DELTA_COUNT 15
#define NFR_MAX_DOUBLE_CORE_COUNT 15

typedef enum NFMinutiaFormat_ // Flags
{
	nfmfHasQuality = 1,
	nfmfHasCurvature = 2,
	nfmfHasG = 4
} NFMinutiaFormat;

typedef enum NFMinutiaType_
{
	nfmtUnknown = 0,
	nfmtEnd = 1,
	nfmtBifurcation = 2
} NFMinutiaType;

typedef struct NFMinutia_
{
	NUShort X;
	NUShort Y;
	NFMinutiaType Type;
	NByte Angle;
	NByte Quality;
	NByte Curvature;
	NByte G;
} NFMinutia;

typedef enum NFRidgeCountsType_
{
	nfrctNone = 0,
	nfrctFourNeighbours = 1,
	nfrctEightNeighbours = 2,
	nfrctFourNeighboursWithIndexes = 5,
	nfrctEightNeighboursWithIndexes = 6,
	nfrctUnspecified = 128 + 4
} NFRidgeCountsType;

typedef struct NFMinutiaNeighbour_
{
	NInt Index;
	NByte RidgeCount;
} NFMinutiaNeighbour;

typedef struct NFCore_
{
	NUShort X;
	NUShort Y;
	NInt Angle;
} NFCore;

typedef struct NFDelta_
{
	NUShort X;
	NUShort Y;
	NInt Angle1;
	NInt Angle2;
	NInt Angle3;
} NFDelta;

typedef struct NFDoubleCore_
{
	NUShort X;
	NUShort Y;
} NFDoubleCore;

N_DECLARE_HANDLE(HNFRecord)

NResult N_API NFRecordGetMaxSize(NFMinutiaFormat minutiaFormat, NInt minutiaCount, NFRidgeCountsType ridgeCountsType,
	NInt coreCount, NInt deltaCount, NInt doubleCoreCount, NInt boWidth, NInt boHeight, NSizeType * pSize);
NResult N_API NFRecordCheck(const void * buffer, NSizeType bufferSize);

NResult N_API NFRecordGetWidthMem(const void * buffer, NSizeType bufferSize, NUShort * pValue);
NResult N_API NFRecordGetHeightMem(const void * buffer, NSizeType bufferSize, NUShort * pValue);
NResult N_API NFRecordGetHorzResolutionMem(const void * buffer, NSizeType bufferSize, NUShort * pValue);
NResult N_API NFRecordGetVertResolutionMem(const void * buffer, NSizeType bufferSize, NUShort * pValue);
NResult N_API NFRecordGetPositionMem(const void * buffer, NSizeType bufferSize, NFPosition * pValue);
NResult N_API NFRecordGetImpressionTypeMem(const void * buffer, NSizeType bufferSize, NFImpressionType * pValue);
NResult N_API NFRecordGetPatternClassMem(const void * buffer, NSizeType bufferSize, NFPatternClass * pValue);
NResult N_API NFRecordGetQualityMem(const void * buffer, NSizeType bufferSize, NByte * pValue);
NResult N_API NFRecordGetGMem(const void * buffer, NSizeType bufferSize, NByte * pValue);
NResult N_API NFRecordGetCbeffProductTypeMem(const void * buffer, NSizeType bufferSize, NUShort * pValue);

typedef struct NFRecordInfo_
{
	NByte MajorVersion;
	NByte MinorVersion;
	NUShort Size;
	NByte HeaderSize;
} NFRecordInfo;

void N_API NFRecordInfoDispose(NFRecordInfo * pInfo);

#define NFR_SKIP_RIDGE_COUNTS    0x00010000 // Load/Save
#define NFR_SKIP_SINGULAR_POINTS 0x00020000 // Load/Save/SaveV1
#define NFR_SKIP_BLOCKED_ORIENTS 0x00040000 // Load
#define NFR_SAVE_BLOCKED_ORIENTS 0x00040000 // Save/SaveV1
#define NFR_SKIP_QUALITIES       0x00100000 // Load/Save
#define NFR_SKIP_CURVATURES      0x00200000 // Load/Save/SaveV1
#define NFR_SKIP_GS              0x00400000 // Load/Save/SaveV1

NResult N_API NFRecordCreate(NUShort width, NUShort height,
	NUShort horzResolution, NUShort vertResolution, NUInt flags, HNFRecord * pHRecord);
NResult N_API NFRecordCreateFromMemory(const void * buffer, NSizeType bufferSize,
	NUInt flags, NFRecordInfo * pInfo, HNFRecord * pHRecord);

void N_API NFRecordFree(HNFRecord hRecord);

NResult N_API NFRecordGetMinutiaCount(HNFRecord hRecord, NInt * pValue);
NResult N_API NFRecordGetMinutia(HNFRecord hRecord, NInt index, NFMinutia * pValue);
NResult N_API NFRecordSetMinutia(HNFRecord hRecord, NInt index, const NFMinutia * pValue);
NResult N_API NFRecordGetMinutiae(HNFRecord hRecord, NFMinutia * arMinutiae);
NResult N_API NFRecordGetMinutiaCapacity(HNFRecord hRecord, NInt * pValue);
NResult N_API NFRecordSetMinutiaCapacity(HNFRecord hRecord, NInt value);
NResult N_API NFRecordAddMinutia(HNFRecord hRecord, const NFMinutia * pValue);
NResult N_API NFRecordInsertMinutia(HNFRecord hRecord, NInt index, const NFMinutia * pValue);
NResult N_API NFRecordRemoveMinutia(HNFRecord hRecord, NInt index);
NResult N_API NFRecordClearMinutiae(HNFRecord hRecord);

NResult N_API NFRecordGetMinutiaNeighbourCount(HNFRecord hRecord, NInt minutiaIndex, NInt * pValue);
NResult N_API NFRecordGetMinutiaNeighbour(HNFRecord hRecord, NInt minutiaIndex, NInt index, NFMinutiaNeighbour * pValue);
NResult N_API NFRecordSetMinutiaNeighbour(HNFRecord hRecord, NInt minutiaIndex, NInt index, const NFMinutiaNeighbour * pValue);
NResult N_API NFRecordGetMinutiaNeighbours(HNFRecord hRecord, NInt minutiaIndex, NFMinutiaNeighbour * arMinutiaNeighbours);

NResult N_API NFRecordGetCoreCount(HNFRecord hRecord, NInt * pValue);
NResult N_API NFRecordGetCore(HNFRecord hRecord, NInt index, NFCore * pValue);
NResult N_API NFRecordSetCore(HNFRecord hRecord, NInt index, const NFCore * pValue);
NResult N_API NFRecordGetCores(HNFRecord hRecord, NFCore * arCores);
NResult N_API NFRecordGetCoreCapacity(HNFRecord hRecord, NInt * pValue);
NResult N_API NFRecordSetCoreCapacity(HNFRecord hRecord, NInt value);
NResult N_API NFRecordAddCore(HNFRecord hRecord, const NFCore * pValue);
NResult N_API NFRecordInsertCore(HNFRecord hRecord, NInt index, const NFCore * pValue);
NResult N_API NFRecordRemoveCore(HNFRecord hRecord, NInt index);
NResult N_API NFRecordClearCores(HNFRecord hRecord);

NResult N_API NFRecordGetDeltaCount(HNFRecord hRecord, NInt * pValue);
NResult N_API NFRecordGetDelta(HNFRecord hRecord, NInt index, NFDelta * pValue);
NResult N_API NFRecordSetDelta(HNFRecord hRecord, NInt index, const NFDelta * pValue);
NResult N_API NFRecordGetDeltas(HNFRecord hRecord, NFDelta * arDeltas);
NResult N_API NFRecordGetDeltaCapacity(HNFRecord hRecord, NInt * pValue);
NResult N_API NFRecordSetDeltaCapacity(HNFRecord hRecord, NInt value);
NResult N_API NFRecordAddDelta(HNFRecord hRecord, const NFDelta * pValue);
NResult N_API NFRecordInsertDelta(HNFRecord hRecord, NInt index, const NFDelta * pValue);
NResult N_API NFRecordRemoveDelta(HNFRecord hRecord, NInt index);
NResult N_API NFRecordClearDeltas(HNFRecord hRecord);

NResult N_API NFRecordGetDoubleCoreCount(HNFRecord hRecord, NInt * pValue);
NResult N_API NFRecordGetDoubleCore(HNFRecord hRecord, NInt index, NFDoubleCore * pValue);
NResult N_API NFRecordSetDoubleCore(HNFRecord hRecord, NInt index, const NFDoubleCore * pValue);
NResult N_API NFRecordGetDoubleCores(HNFRecord hRecord, NFDoubleCore * arDoubleCores);
NResult N_API NFRecordGetDoubleCoreCapacity(HNFRecord hRecord, NInt * pValue);
NResult N_API NFRecordSetDoubleCoreCapacity(HNFRecord hRecord, NInt value);
NResult N_API NFRecordAddDoubleCore(HNFRecord hRecord, const NFDoubleCore * pValue);
NResult N_API NFRecordInsertDoubleCore(HNFRecord hRecord, NInt index, const NFDoubleCore * pValue);
NResult N_API NFRecordRemoveDoubleCore(HNFRecord hRecord, NInt index);
NResult N_API NFRecordClearDoubleCores(HNFRecord hRecord);

NResult N_API NFRecordClone(HNFRecord hRecord, HNFRecord * pHClonedRecord);
NResult N_API NFRecordGetSize(HNFRecord hRecord, NUInt flags, NSizeType * pSize);
NResult N_API NFRecordSaveToMemory(HNFRecord hRecord, void * buffer, NSizeType bufferSize, NUInt flags, NSizeType * pSize);

NResult N_API NFRecordGetWidth(HNFRecord hRecord, NUShort * pValue);
NResult N_API NFRecordGetHeight(HNFRecord hRecord, NUShort * pValue);
NResult N_API NFRecordGetHorzResolution(HNFRecord hRecord, NUShort * pValue);
NResult N_API NFRecordGetVertResolution(HNFRecord hRecord, NUShort * pValue);
NResult N_API NFRecordGetPosition(HNFRecord hRecord, NFPosition * pValue);
NResult N_API NFRecordSetPosition(HNFRecord hRecord, NFPosition value);
NResult N_API NFRecordGetImpressionType(HNFRecord hRecord, NFImpressionType * pValue);
NResult N_API NFRecordSetImpressionType(HNFRecord hRecord, NFImpressionType value);
NResult N_API NFRecordGetPatternClass(HNFRecord hRecord, NFPatternClass * pValue);
NResult N_API NFRecordSetPatternClass(HNFRecord hRecord, NFPatternClass value);
NResult N_API NFRecordGetQuality(HNFRecord hRecord, NByte * pValue);
NResult N_API NFRecordSetQuality(HNFRecord hRecord, NByte value);
NResult N_API NFRecordGetG(HNFRecord hRecord, NByte * pValue);
NResult N_API NFRecordSetG(HNFRecord hRecord, NByte value);
NResult N_API NFRecordGetCbeffProductType(HNFRecord hRecord, NUShort * pValue);
NResult N_API NFRecordSetCbeffProductType(HNFRecord hRecord, NUShort value);
NResult N_API NFRecordGetRidgeCountsType(HNFRecord hRecord, NFRidgeCountsType * pValue);
NResult N_API NFRecordSetRidgeCountsType(HNFRecord hRecord, NFRidgeCountsType value);
NResult N_API NFRecordGetMinutiaFormat(HNFRecord hRecord, NFMinutiaFormat * pValue);
NResult N_API NFRecordSetMinutiaFormat(HNFRecord hRecord, NFMinutiaFormat value);

#ifdef N_CPP
}
#endif

#endif // !NF_RECORD_H_INCLUDED
