#include "System.h"
#include "AboutForm.h"
#include "AboutForm.rh"
#include "VFingerX.h"
#include "ScanManX.h"

////////////////
//            //
// CAboutForm //
//            //
////////////////

CAboutForm::CAboutForm(const string &engineName)
	: CForm(), engineName(engineName)
{
	SetMinimizeBox(false);
	SetMaximizeBox(false);
	SetFormBorder(fbDialog);
	SetFormPosition(fpOwnerCenter);
	SetText(STRING("About"));

	bitmap_logo = Application()->LoadBitmap(MAKEINTRESOURCE(IDB_LOGO));
	CPicture *picture_logo = new CPicture();
	picture_logo->SetLocation(8, 8);
	picture_logo->SetBitmap(bitmap_logo);

	int x = picture_logo->X() + picture_logo->Width() + 8;
	int y = 8;
	int w = 420;
	int h = y + picture_logo->Height();
	
	CLabel *label_app = new CLabel();
	label_app->SetLocation(x, y);
	label_app->SetWidth(w);
	y += label_app->Height();
	CLabel *label_appcopyright = new CLabel();
	label_appcopyright->SetLocation(x, y);
	label_appcopyright->SetWidth(w);
	y += label_appcopyright->Height() + 4;
	ShowVersion(Application()->Name(), Application()->VersionHigh(), Application()->VersionLow(),
		Application()->Copyright(), label_app, label_appcopyright);

	int xx = x + 8, yy, ww = w - 16;

	CGroupBox *groupbox_os = new CGroupBox();
	groupbox_os->SetLocation(x, y);
	groupbox_os->SetWidth(w);
	groupbox_os->SetText(STRING("Operating system"));
	yy = y + 16;
	CLabel *label_os = new CLabel();
	label_os->SetLocation(xx, yy);
	label_os->SetWidth(ww);
	label_os->SetText(Format(TEXT("%s. Version %lu.%lu (Build %lu)"), (LPCTSTR)Application()->OSName(),
		Application()->OSMajorVersion(), Application()->OSMinorVersion(), Application()->OSBuild()));
	yy += label_os->Height();
	CLabel *label_ossp = new CLabel();
	label_ossp->SetLocation(xx, yy);
	label_ossp->SetWidth(ww);
	label_ossp->SetText(Format(TEXT("Service Pack Version %lu.%lu"),
		Application()->OSSPMajorVersion(), Application()->OSSPMinorVersion()));
	yy += label_ossp->Height() + 8;
	groupbox_os->SetHeight(yy - y);
	y += groupbox_os->Height();

	CGroupBox *groupbox_vf = new CGroupBox();
	groupbox_vf->SetLocation(x, y);
	groupbox_vf->SetWidth(w);
	yy = y + 12;
	CLabel *label_vf = new CLabel();
	label_vf->SetLocation(xx, yy);
	label_vf->SetWidth(ww);
	yy += label_vf->Height();
	CLabel *label_vfcopyright = new CLabel();
	label_vfcopyright->SetLocation(xx, yy);
	label_vfcopyright->SetWidth(ww);
	yy += label_vfcopyright->Height() + 6;
	ShowVersion(VFGetName(), VFGetVersionHigh(), VFGetVersionLow(), VFGetCopyright(),
		label_vf, label_vfcopyright);
	label_is_registered = new CLabel();
	label_is_registered->SetLocation(xx, yy);
	label_is_registered->SetWidth(100);
	label_is_registered->SetText(STRING("Is registered: ") + (VFRegistrationType() == VF_RT_UNREGISTERED ? TEXT("No") : TEXT("Yes")));
	yy += label_is_registered->Height() + 10;
	groupbox_vf->SetHeight(yy - y);
	y += groupbox_vf->Height();
	
	CGroupBox *groupbox_sm = new CGroupBox();
	groupbox_sm->SetLocation(x, y);
	groupbox_sm->SetSize(w, h - y);
	yy = y + 12;
	CLabel *label_sm = new CLabel();
	label_sm->SetLocation(xx, yy);
	label_sm->SetWidth(ww);
	yy += label_sm->Height();
	CLabel *label_smcopyright = new CLabel();
	label_smcopyright->SetLocation(xx, yy);
	label_smcopyright->SetWidth(ww);
	yy += label_smcopyright->Height() + 4;
	ShowVersion(SMGetName(), SMGetVersionHigh(), SMGetVersionLow(), SMGetCopyright(),
		label_sm, label_smcopyright);
	
	y = h + 8;
	CButton *button_ok = new CButton();
	button_ok->SetLocation(x + w - button_ok->Width(), y);
	button_ok->SetDefault(true);
	button_ok->SetCancel(true);
	button_ok->SetModalResult(IDOK);
	button_ok->SetText(STRING("OK"));
	y += button_ok->Height();

	x += w + 8;
	y += 8;
	SetClientSize(x, y);

	AddControl(picture_logo);

	AddControl(label_app);
	AddControl(label_appcopyright);

	AddControl(label_os);
	AddControl(label_ossp);
	AddControl(groupbox_os);

	AddControl(label_vf);
	AddControl(label_vfcopyright);
	AddControl(label_is_registered);
	AddControl(groupbox_vf);

	AddControl(label_sm);
	AddControl(label_smcopyright);
	AddControl(groupbox_sm);

	AddControl(button_ok);
}

CAboutForm::~CAboutForm()
{
	DeleteBitmap(bitmap_logo);
}

void CAboutForm::OnCreate()
{
	CForm::OnCreate();
}

void CAboutForm::ShowVersion(const string &name, DWORD version_high, DWORD version_low, const string &copyright,
	CLabel *label_name, CLabel *label_copyright)
{
	label_name->SetText(Format(TEXT("%s    Version %u.%u.%u.%u"), (LPCTSTR)name,
		HIWORD(version_high), LOWORD(version_high), HIWORD(version_low), LOWORD(version_low)));
	label_copyright->SetText(copyright);
}

//////////////
//          //
// AboutBox //
//          //
//////////////

void AboutBox(CForm *owner, const string &engineName)
{
	CAboutForm *form = new CAboutForm(engineName);
	form->SetParent(owner);
	form->ShowModal();
	delete form;
}
