#include "System.h"
#include "Application.h"
#include "Utils.h"

CApplication * application = NULL;

#ifndef VER_NT_WORKSTATION
#define VER_NT_WORKSTATION			0x0000001
#endif
#ifndef VER_NT_DOMAIN_CONTROLLER
#define VER_NT_DOMAIN_CONTROLLER	0x0000002
#endif
#ifndef VER_NT_SERVER
#define VER_NT_SERVER				0x0000003
#endif

#ifndef VER_SERVER_NT
#define VER_SERVER_NT                       0x80000000
#endif
#ifndef VER_WORKSTATION_NT
#define VER_WORKSTATION_NT                  0x40000000
#endif
#ifndef VER_SUITE_SMALLBUSINESS
#define VER_SUITE_SMALLBUSINESS             0x00000001
#endif
#ifndef VER_SUITE_ENTERPRISE
#define VER_SUITE_ENTERPRISE                0x00000002
#endif
#ifndef VER_SUITE_BACKOFFICE
#define VER_SUITE_BACKOFFICE                0x00000004
#endif
#ifndef VER_SUITE_COMMUNICATIONS
#define VER_SUITE_COMMUNICATIONS            0x00000008
#endif
#ifndef VER_SUITE_TERMINAL
#define VER_SUITE_TERMINAL                  0x00000010
#endif
#ifndef VER_SUITE_SMALLBUSINESS_RESTRICTED
#define VER_SUITE_SMALLBUSINESS_RESTRICTED  0x00000020
#endif
#ifndef VER_SUITE_EMBEDDEDNT
#define VER_SUITE_EMBEDDEDNT                0x00000040
#endif
#ifndef VER_SUITE_DATACENTER
#define VER_SUITE_DATACENTER                0x00000080
#endif
#ifndef VER_SUITE_SINGLEUSERTS
#define VER_SUITE_SINGLEUSERTS              0x00000100
#endif
#ifndef VER_SUITE_PERSONAL
#define VER_SUITE_PERSONAL                  0x00000200
#endif
#ifndef VER_SUITE_BLADE
#define VER_SUITE_BLADE                     0x00000400
#endif

CApplication::CApplication(HINSTANCE handle, bool initCom)
	: versionHigh(MAKEDWORD(0, 1)), versionLow(MAKEDWORD(0, 0)), locale(0), running(false)
{
	// Register application
	if(application) throw exception("Another instance of application already exists");
	application = this;
	
	// Store instance handle and show command
	this->handle = handle == NULL ? GetModuleHandle(NULL) : handle;

	// Retrieve executable file name and set title
	const int maxModuleFileName = 1000;
	exeName.SetLength(maxModuleFileName);
	GetModuleFileName(this->handle, exeName, maxModuleFileName);
	exeName.Update();
	name = GetFileNameWithoutExtension(exeName);

	// Get operating system version information
	OSVERSIONINFOEX vie;
	vie.dwOSVersionInfoSize = sizeof(vie);
	if(GetVersionEx((LPOSVERSIONINFO)&vie))
	{
		osMajorVersion = vie.dwMajorVersion;
		osMinorVersion = vie.dwMinorVersion;
		osPlatformId = vie.dwPlatformId;
		osBuild = vie.dwBuildNumber;
		osUpdateName.Assign(vie.szCSDVersion);
		osSPMajorVersion = vie.wServicePackMajor;
		osSPMinorVersion = vie.wServicePackMinor;
#if WINVER > 0x0400
		osSuite = vie.wSuiteMask;
		osType = vie.wProductType;
#else
		osSuite = vie.wReserved[0];
		osType = LOBYTE(vie.wReserved[1]);
#endif
	}
	else
	{
		osSPMajorVersion = 0;
		osSPMinorVersion = 0;
		osSuite = 0;
		osType = 0;
		OSVERSIONINFO vi;
		vi.dwOSVersionInfoSize = sizeof(vi);
		if(GetVersionEx(&vi))
		{
			osMajorVersion = vi.dwMajorVersion;
			osMinorVersion = vi.dwMinorVersion;
			osPlatformId = vi.dwPlatformId;
			osBuild = vi.dwBuildNumber;
			osUpdateName.Assign(vi.szCSDVersion);
		}
		else
		{
			osMajorVersion = 0;
			osMinorVersion = 0;
			osPlatformId = 0;
			osBuild = 0;
		}
	}
	if(osPlatformId != VER_PLATFORM_WIN32_NT) osBuild = LOWORD(osBuild);

	// OS name
	if(osPlatformId == VER_PLATFORM_WIN32s) osName = TEXT("Win32s");
	else if(osPlatformId == VER_PLATFORM_WIN32_WINDOWS)
	{
		if(osMajorVersion > 4) osName = TEXT("Windows");
		else if(osMajorVersion == 4)
		{
			if(osMinorVersion >= 90) osName = TEXT("Windows Me");
			else if(osMinorVersion >= 10) osName = TEXT("Windows 98");
			else osName = TEXT("Windows 95");
		}
		else osName = Format(TEXT("Windows %ul.%ul"), osMajorVersion, osMinorVersion);
	}
	else if(osPlatformId == VER_PLATFORM_WIN32_NT)
	{
		if(osMajorVersion > 5) osName = TEXT("Windows NT");
		else if(osMajorVersion == 5)
		{
			if(osMinorVersion >= 1)
			{
				if(osType == VER_NT_SERVER && osMinorVersion >= 2)
				{
					if(osSuite & VER_SUITE_BLADE) osName = TEXT("Windows Server 2003, Web Edition");
					if(osSuite & VER_SUITE_DATACENTER) osName = TEXT("Windows Server 2003, Datacenter Edition");
					if(osSuite & VER_SUITE_ENTERPRISE) osName = TEXT("Windows Server 2003, Enterprise Edition");
					else osName = TEXT("Windows Server 2003 family");
				}
				else if(osType == VER_NT_WORKSTATION)
				{
					if(osSuite & VER_SUITE_PERSONAL) osName = TEXT("Windows XP Home Edition");
					else osName = TEXT("Windows XP Professional");
				}
				else
				{
					osName = TEXT("Windows XP/Server");
				}
			}
			else
			{
				if(osType == VER_NT_SERVER)
				{
					if(osSuite & VER_SUITE_DATACENTER) osName = TEXT("Windows 2000 DataCenter Server");
					else if(osSuite & VER_SUITE_ENTERPRISE) osName = TEXT("Windows 2000 Advanced Server");
					else osName = TEXT("Windows 2000 Server");
				}
				else if(osType == VER_NT_WORKSTATION) osName = TEXT("Windows 2000 Professional");
				else osName = TEXT("Windows 2000");
			}
		}
		else if(osMajorVersion == 4)
		{
			if(osType == VER_NT_SERVER) osName = TEXT("Windows NT 4.0 Server");
			else if(osType == VER_NT_WORKSTATION) osName = TEXT("Windows NT 4.0 Workstation");
			else osName = TEXT("Windows NT 4.0");
		}
		else osName = Format(TEXT("Windows NT %ul.%ul"), osMajorVersion, osMinorVersion);
	}
	else osName = TEXT("[Unknown]");
	if(!osUpdateName.IsEmpty())
		if(osPlatformId == VER_PLATFORM_WIN32_WINDOWS) osName += osUpdateName;
		else osName += STRING(", ") + osUpdateName;

	// Initialize COM library
	if(initCom && FAILED(::CoInitialize(NULL))) throw exception("Can not initialize COM library");
	uninitCom = initCom;
}

CApplication::~CApplication()
{
	// Uninitialize COM library
	if(uninitCom) ::CoUninitialize();
}

void CApplication::OnStartRunning()
{
	if(running) throw exception("Application is already running");
	running = true;
}

void CApplication::OnEndRunning()
{
	if(!running) throw exception("Application is not running");
	running = false;
}

void CApplication::CheckRunning()
{
	if(running) throw exception("Application is running");
}

void CApplication::SetName(const string &value)
{
	CheckRunning();
	name = value;
}

void CApplication::SetVersion(DWORD versionHigh, DWORD versionLow)
{
	CheckRunning();
	this->versionHigh = versionHigh;
	this->versionLow = versionLow;
}

void CApplication::SetCopyright(const string &value)
{
	CheckRunning();
	copyright = value;
}

void CApplication::SetCompany(const string &value)
{
	CheckRunning();
	company = value;
}

void CApplication::SetRegistryKey(const string &value)
{
	CheckRunning();
	registryKey = value;
}

void CApplication::SetLocale(LCID value)
{
	if(locale != value)
	{
		if(value != 0)
		{
			LCID *i;
			for(i = locales.Begin(); i < locales.End(); i++)
				if(*i == value) break;
			if(i == locales.End()) throw exception("Locale not available");
		}
		locale = value;
		if(!running) SetThreadLocale(locale ? locale : LOCALE_USER_DEFAULT);
	}
}

int CApplication::LocaleCount()
{
	return locales.Length();
}

LCID CApplication::Locale(int index)
{
	return locales[index];
}

void CApplication::AddLocale(LCID value)
{
	CheckRunning();
	locales.Add(value);
}

string CApplication::LoadString(DWORD id, ...)
{
	va_list list;
	va_start(list, id);
	string s = LoadStringV(id, list);
	va_end(list);
	return s;
}

string CApplication::LoadStringV(DWORD id, va_list list)
{
	const int max_load_string = 1000;
	TCHAR buf[max_load_string];
	if(!::LoadString(handle, id, buf, max_load_string))
		throw EXCEPTION(Format(TEXT("Can not find string with resource id %u"), id));
	return FormatV(buf, list);
}

HICON CApplication::LoadIcon(LPCTSTR iconName)
{
	return ::LoadIcon(handle, iconName);
}

HBITMAP CApplication::LoadBitmap(LPCTSTR bitmapName)
{
	return ::LoadBitmap(handle, bitmapName);
}

CApplication * Application()
{
	return application;
}

bool IsApplication()
{
	return Application() != NULL;
}

void CheckApplication()
{
	if(!IsApplication()) throw exception("No application");
}
