#include "System.h"
#include "Controls.h"
#include "Utils.h"
#include "GUIApplication.h"

//////////////
//          //
// CControl //
//          //
//////////////

CControl::CControl(bool subclass)
	: handle(NULL), parent(NULL), tab_stop(false), group(false), x(0), y(0), width(0), height(0),
	enabled(true), visible(true),
	control_count(0), controls(NULL), client_edge(false), window_edge(false), static_edge(false)
{
	this->is_subclass = subclass;
	font = CreateFont(-MulDiv(8, GetDeviceCaps(GetDC(NULL), LOGPIXELSY), 72),
		0, 0, 0, 0, FALSE, FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE,
		TEXT("MS Sans Serif"));
}

CControl::~CControl()
{
	while(control_count) delete controls[0];
	DestroyWnd();
	SetParent(NULL);
	if(font) DeleteFont(font);
}

string CControl::GetWndClassName()
{
	string s = AnsiToString(typeid(*this).name());
	const TCHAR *ss = TEXT("class ");
	int i = s.IndexOf(ss);
	if(i != -1)
	{
		s.Remove(0, _tcslen(ss));
		if(!s.IsEmpty() && s[0] == TEXT('C')) s.Remove(0);
	}
	return s;
}

void CControl::GetWndClass(WNDCLASSEX &wcex)
{
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
	wcex.lpfnWndProc = (WNDPROC)WndProc;
	wcex.hInstance = Application()->Handle();
	wcex.hbrBackground = GetBackgroundBrush();
	wcex.hCursor = GetCursor();
}

void CControl::RegisterClass()
{
	WNDCLASSEX wcex;
	memset(&wcex, 0, sizeof(wcex));
	wnd_class = GetWndClassName();
	wcex.lpszClassName = wnd_class;
	if(!is_subclass)
	{
		GetWndClass(wcex);
		RegisterClassEx(&wcex);
	}
}

DWORD CControl::GetWndStyle()
{
	return WS_CHILD | WS_CLIPCHILDREN | WS_CLIPSIBLINGS
		| (visible ? WS_VISIBLE : 0)
		| (enabled ? 0 : WS_DISABLED)
		| (tab_stop ? WS_TABSTOP : 0)
		| (group ? WS_GROUP : 0);
}

DWORD CControl::GetWndExStyle()
{
	return (client_edge ? WS_EX_CLIENTEDGE : 0)
		| (window_edge ? WS_EX_WINDOWEDGE : 0)
		| (static_edge ? WS_EX_STATICEDGE : 0);
}

void CControl::CreateWnd()
{
	if(!HandleAllocated())
	{
		RegisterClass();
		handle = CreateWindowEx(GetWndExStyle(), wnd_class, text, GetWndStyle(),
			x, y, width, height, ParentHandle(), NULL, Application()->Handle(), NULL);
		SetWindowLong(handle, GWL_USERDATA, (LONG)this);
		if(is_subclass) old_wnd_proc = SubclassWindow(handle, (WNDPROC)WndProc);
		RECT rect;
		GetWindowRect(handle, &rect);
		if(parent && parent->HandleAllocated())
		{
			POINT pt = { rect.left, rect.top };
			parent->ScreenToClient(pt);
			x = pt.x;
			y = pt.y;
		}
		width = rect.right - rect.left;
		height = rect.bottom - rect.top;
		SetFont(font);
		for(int i = 0; i < control_count; i++)
			controls[i]->CreateWnd();
		OnCreate();
	}
}

void CControl::DestroyWnd()
{
	if(HandleAllocated())
	{
		OnDestroy();
		for(int i = 0; i < control_count; i++) controls[i]->DestroyWnd();
		DestroyWindow(handle);
		handle = NULL;
	}
}

void CControl::RecreateWnd()
{
	if(HandleAllocated())
	{
		DestroyWnd();
		CreateWnd();
	}
}

LRESULT CALLBACK CControl::WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	CControl *control = (CControl*)GetWindowLong(hWnd, GWL_USERDATA);
	LRESULT r = control ? control->OnMessage(message, wParam, lParam) : DefWindowProc(hWnd, message, wParam, lParam);
	if(message == WM_DESTROY && control) control->handle = NULL;
	return r;
}

LRESULT CControl::DefWndProc(UINT message, WPARAM wParam, LPARAM lParam)
{
	return is_subclass ? CallWindowProc(old_wnd_proc, handle, message, wParam, lParam)
		: DefWindowProc(handle, message, wParam, lParam);
}

void CControl::RestoreMessage(MSG &msg, UINT message, WPARAM wParam, LPARAM lParam)
{
	msg.hwnd = handle;
	msg.message = message;
	msg.wParam = wParam;
	msg.lParam = lParam;
	msg.time = GetMessageTime();
	DWORD pos = GetMessagePos();
	msg.pt.x = GET_X_LPARAM(pos);
	msg.pt.y = GET_Y_LPARAM(pos);
}

CForm* CControl::ParentForm()
{
	CControl *control = this;
	CForm *form = NULL;
	do
	{
		form = dynamic_cast<CForm*>(control);
		if(control->parent) control = control->parent;
		else break;
	} while(!form);
	return form;
}

#define CControl_DefProc return DefWndProc(message, wParam, lParam)

LRESULT CControl::OnMessage(UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
		case WM_COMMAND:
			{
				WORD id = GET_WM_COMMAND_ID(wParam, lParam);
				WORD event = GET_WM_COMMAND_CMD(wParam, lParam);
				HWND hwnd = GET_WM_COMMAND_HWND(wParam, lParam);
				CControl *control = GetControl(hwnd);
				if(control) control->OnCommandSent(event);
				ParentForm()->OnCommand(id, event, control);
			}
			CControl_DefProc;
		case WM_NOTIFY:
			{
				LPNMHDR nmhdr = (LPNMHDR)lParam;
				HWND hwnd = nmhdr->hwndFrom;
				CControl *control = GetControl(hwnd);
				if(control) control->OnNotifySent(nmhdr);
				LRESULT r = ParentForm()->OnNotify(nmhdr, control);
				if(!r) CControl_DefProc;
				return r;
			}
		case WM_HSCROLL:
			{
				WORD code = GET_WM_HSCROLL_CODE(wParam, lParam);
				WORD pos = GET_WM_HSCROLL_POS(wParam, lParam);
				HWND hwnd = GET_WM_HSCROLL_HWND(wParam, lParam);
				CControl *control = GetControl(hwnd);
				if(control) control->OnHScrollSent(code, pos);
				ParentForm()->OnHScroll(code, pos, control);
			}
			CControl_DefProc;
		case WM_VSCROLL:
			{
				WORD code = GET_WM_VSCROLL_CODE(wParam, lParam);
				WORD pos = GET_WM_VSCROLL_POS(wParam, lParam);
				HWND hwnd = GET_WM_VSCROLL_HWND(wParam, lParam);
				CControl *control = GetControl(hwnd);
				if(control) control->OnVScrollSent(code, pos);
				ParentForm()->OnVScroll(code, pos, control);
			}
			CControl_DefProc;
		case WM_WINDOWPOSCHANGED:
			{
				LPWINDOWPOS wp = (LPWINDOWPOS)lParam;
				if(!(wp->flags & SWP_NOMOVE))
				{
					x = wp->x;
					y = wp->y;
				}
				if(!(wp->flags & SWP_NOSIZE))
				{
					width = wp->cx;
					height = wp->cy;
				}
				if((wp->flags & SWP_HIDEWINDOW) || (wp->flags & SWP_SHOWWINDOW))
				{
					bool new_visible = (wp->flags & SWP_HIDEWINDOW) == 0;
					if(new_visible != visible)
					{
						visible = new_visible;
						DoVisibleChanged();
					}
				}
			}
			CControl_DefProc;
		case WM_ENABLE:
			{
				enabled = wParam != 0;
				LRESULT r = DefWndProc(message, wParam, lParam);
				DoEnabledChanged();
				return r;
			}
		case WM_SETTEXT:
			{
				text.Assign((LPCTSTR)lParam);
				LRESULT r = DefWndProc(message, wParam, lParam);
				DoTextChanged();
				return r;
			}
		case WM_SETFONT:
			{
				HFONT font = (HFONT)wParam;
				if(font != this->font)
				{
					if(this->font) DeleteFont(this->font);
					if(font)
					{
						LOGFONT lf;
						GetObject(font, sizeof(lf), &lf);
						this->font = CreateFontIndirect(&lf);
					}
					else this->font = NULL;
					LRESULT r = DefWndProc(message, wParam, lParam);
					OnFontChanged();
					return r;
				}
			}
			CControl_DefProc;
		case WM_GETFONT: return (LRESULT)font;
		case WM_SIZE: return OnSize((DWORD)wParam, LOWORD(lParam), HIWORD(lParam));
		case WM_PAINT: return OnPaint(wParam, lParam);
		case WM_ERASEBKGND: return OnEraseBackground((HDC)wParam, lParam);
		case WM_LBUTTONDBLCLK: return OnLButtonDblClk((DWORD)wParam, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
		case WM_LBUTTONDOWN: return OnLButtonDown((DWORD)wParam, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
		case WM_LBUTTONUP: return OnLButtonUp((DWORD)wParam, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
		case WM_MBUTTONDBLCLK: return OnMButtonDblClk((DWORD)wParam, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
		case WM_MBUTTONDOWN: return OnMButtonDown((DWORD)wParam, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
		case WM_MBUTTONUP: return OnMButtonUp((DWORD)wParam, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
		case WM_RBUTTONDBLCLK: return OnRButtonDblClk((DWORD)wParam, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
		case WM_RBUTTONDOWN: return OnRButtonDown((DWORD)wParam, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
		case WM_RBUTTONUP: return OnRButtonUp((DWORD)wParam, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
		case WM_MOUSEMOVE: return OnMouseMove((DWORD)wParam, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
		case WM_GETDLGCODE: return OnGetDlgCode(wParam, (MSG*)lParam);
		case WM_KEYDOWN: return OnKeyDown((DWORD)wParam, (DWORD)lParam);
		case WM_KEYUP: return OnKeyUp((DWORD)wParam, (DWORD)lParam);
		case WM_KILLFOCUS:
			ParentForm()->OnControlKillFocus((HWND)wParam, this);
			return OnKillFocus((HWND)wParam, lParam);
		case WM_SETFOCUS:
			ParentForm()->OnControlSetFocus((HWND)wParam, this);
			return OnSetFocus((HWND)wParam, lParam);
		case WM_SYSKEYDOWN: return OnSysKeyDown((DWORD)wParam, (DWORD)lParam);
		case WM_SYSKEYUP: return OnSysKeyUp((DWORD)wParam, (DWORD)lParam);
		default: CControl_DefProc;
	}
}

void CControl::DoEnabledChanged()
{
	CForm *form = ParentForm();
	if(form) form->OnControlEnabledChanged(this);
	OnEnabledChanged();
}

void CControl::DoVisibleChanged()
{
	CForm *form = ParentForm();
	if(form) form->OnControlVisibleChanged(this);
	OnVisibleChanged();
}

void CControl::OnActiveChanged()
{
	for(int i = 0; i < control_count; i++)
		controls[i]->OnActiveChanged();
}

void CControl::OnActiveControlChanged()
{
	for(int i = 0; i < control_count; i++)
		controls[i]->OnActiveControlChanged();
}

void CControl::OnControlParentChanged(CControl *control)
{
	for(int i = 0; i < control_count; i++)
		controls[i]->OnControlParentChanged(control);
}

bool CControl::OnDefault()
{
	for(int i = 0; i < control_count; i++)
		if(controls[i]->OnDefault()) return true;
	return false;
}

bool CControl::OnCancel()
{
	for(int i = 0; i < control_count; i++)
		if(controls[i]->OnCancel()) return true;
	return false;
}

LRESULT CControl::SendMessage(UINT message, WPARAM wParam, LPARAM lParam)
{
	CheckHandle();
	return ::SendMessage(handle, message, wParam, lParam);
}

bool CControl::PostMessage(UINT message, WPARAM wParam, LPARAM lParam)
{
	CheckHandle();
	return ::PostMessage(handle, message, wParam, lParam) != 0;
}

LRESULT CControl::SendCommand(WORD event)
{
	if(HandleAllocated() && parent && parent->HandleAllocated())
		return parent->SendMessage(WM_COMMAND, GET_WM_COMMAND_MPS(0, handle, event));
	else return 0;
}

LRESULT CControl::SendNotify(NMHDR * nmh)
{
	if(HandleAllocated() && parent && parent->HandleAllocated())
	{
		nmh->hwndFrom = handle;
		nmh->idFrom = 0;
		return parent->SendMessage(WM_NOTIFY, 0, (LPARAM)nmh);
	}
	else return 0;
}

LRESULT CControl::SendNotify(UINT code)
{
	NMHDR nmh;
	nmh.code = code;
	return SendNotify(&nmh);
}

void CControl::OnCommandSent(WORD /*event*/)
{
}

void CControl::OnNotifySent(LPNMHDR /*nmh*/)
{
}

void CControl::OnHScrollSent(WORD /*code*/, WORD /*pos*/)
{
}

void CControl::OnVScrollSent(WORD /*code*/, WORD /*pos*/)
{
}

void CControl::SetBounds(int x, int y, int width, int height)
{
	this->x = x;
	this->y = y;
	this->width = width;
	this->height = height;
	if(HandleAllocated())
		SetWindowPos(handle, NULL, x, y, width, height, SWP_NOACTIVATE | SWP_NOREPOSITION | SWP_NOZORDER);
}

void CControl::SetEnabled(bool value)
{
	if(enabled != value)
	{
		if(HandleAllocated()) EnableWindow(handle, value);
		else
		{
			enabled = value;
			DoEnabledChanged();
		}
	}
}

void CControl::SetVisible(bool value)
{
	if(value) Show(); else Hide();
}

void CControl::SetParent(CControl *value)
{
	if(parent != value)
	{
		if(parent) parent->RemoveControl(this);
		if(value) value->AddControl(this);
	}
}

void CControl::SetText(const string &value)
{
	if(HandleAllocated())
	{
		SetWindowText(handle, value);
	}
	else
	{
		text = value;
		OnTextChanged();
	}
}

void CControl::Show(int cmd_show)
{
	if(HandleAllocated()) ShowWindow(handle, cmd_show);
	else
	{
		bool new_visible = cmd_show != SW_HIDE;
		if(new_visible != visible)
		{
			visible = new_visible;
			DoVisibleChanged();
		}
	}
}

void CControl::Hide()
{
	Show(SW_HIDE);
}

void CControl::Update()
{
	if(HandleAllocated()) UpdateWindow(handle);
}

void CControl::Invalidate(RECT *rect, bool erase)
{
	if(HandleAllocated()) InvalidateRect(handle, rect, erase);
}

void CControl::SetRedrawState(bool value)
{
	if(HandleAllocated()) SetWindowRedraw(handle, value);
}

void CControl::AddControl(CControl *control)
{
	if(!control) throw exception("Can not add a NULL control");
	if(control->parent) throw exception("Control already has a parent");
	CControl **new_controls = new CControl*[control_count + 1];
	for(int i = 0; i < control_count; i++) new_controls[i] = controls[i];
	new_controls[control_count] = control;
	control_count++;
	if(controls) delete[] controls;
	controls = new_controls;
	control->parent = this;
	if(HandleAllocated() && control->HandleAllocated()) ::SetParent(control->handle, handle);
	CForm *form = ParentForm();
	if(form) form->OnControlParentChanged(control);
}

void CControl::RemoveControl(CControl *control)
{
	int i;
	for(i = 0; i < control_count; i++)
		if(controls[i] == control) break;
	if(i != control_count)
	{
		control->parent = NULL;
		if(control->HandleAllocated()) ::SetParent(control->handle, NULL);
		control_count--;
		CControl **new_controls = new CControl*[control_count];
		int j;
		for(j = 0; j < i; j++) new_controls[j] = controls[j];
		for(; j < control_count; j++) new_controls[j] = controls[j + 1];
		delete[] controls;
		controls = new_controls;
		CForm *form = ParentForm();
		if(form) form->OnControlParentChanged(control);
	}
}

CControl* CControl::GetControl(HWND handle)
{
	if(!handle) return NULL;
	for(int i = 0; i < control_count; i++)
		if(controls[i]->handle == handle)
			return controls[i];
	return NULL;
}

int CControl::IndexOfControl(CControl *control)
{
	int index = -1;
	for(int i = 0; i < control_count; i++)
		if(controls[i] == control)
		{
			index = i;
			break;
		}
	return index;
}

bool CControl::ContainsControl(CControl *control)
{
	for(int i = 0; i < control_count; i++)
		if(controls[i] == control || controls[i]->ContainsControl(control)) return true;
	return false;
}

CControl* CControl::FindControl(HWND handle)
{
	if(!handle) return NULL;
	int i;
	for(i = 0; i < control_count; i++)
		if(controls[i]->handle == handle) return controls[i];
	for(i = 0; i < control_count; i++)
	{
		CControl *control = controls[i]->FindControl(handle);
		if(control) return control;
	}
	return NULL;
}

void CControl::GetClientRect(RECT &rect)
{
	if(HandleAllocated()) ::GetClientRect(handle, &rect);
	else
	{
		rect.left = 0;
		rect.top = 0;
		rect.right = width;
		rect.bottom = height;
		AdjustRect(rect);
		rect.right = width - (-rect.left + (rect.right - width));
		rect.bottom = height - (-rect.top + (rect.bottom - height));
		rect.left = 0;
		rect.top = 0;
	}
}

int CControl::ClientWidth()
{
	RECT rect;
	GetClientRect(rect);
	return rect.right - rect.left;
}

int CControl::ClientHeight()
{
	RECT rect;
	GetClientRect(rect);
	return rect.bottom - rect.top;
}

bool CControl::Capture()
{
	return HandleAllocated() && ::GetCapture() == handle;
}

void CControl::SetCapture(bool value)
{
	if(!HandleAllocated()) return;
	if(Capture()) { if(!value) ::ReleaseCapture(); }
	else {	if(value) ::SetCapture(handle); }
}

void CControl::SetClientEdge(bool value)
{
	if(client_edge != value)
	{
		client_edge = value;
		RecreateWnd();
	}
}

void CControl::SetWindowEdge(bool value)
{
	if(window_edge != value)
	{
		window_edge = value;
		RecreateWnd();
	}
}

void CControl::SetStaticEdge(bool value)
{
	if(static_edge != value)
	{
		static_edge = value;
		RecreateWnd();
	}
}

void CControl::SetFont(HFONT value)
{
	if(HandleAllocated())
	{
		SetWindowFont(handle, value, true);
	}
	else font = value;
}

bool CControl::Focused()
{
	return HandleAllocated() && GetFocus() == handle;
}

void CControl::SetFocus()
{
	if(HandleAllocated()) ::SetFocus(handle);
}

bool CControl::CanFocus()
{
	if(!HandleAllocated()) return false;
	CControl *control = this;
	CForm *form = ParentForm();
	do
	{
		if(!control->visible || !control->enabled) return false;
		control = control->parent;
	}
	while(control != form);
	return true; 
}

LONG CControl::ClassLong(int index)
{
	CheckHandle();
	return GetClassLong(handle, index);
}

DWORD CControl::ClassStyle()
{
	return (DWORD)ClassLong(GCL_STYLE);
}

HBRUSH CControl::BackgroundBrush()
{
	return (HBRUSH)ClassLong(GCL_HBRBACKGROUND);
}

HCURSOR CControl::Cursor()
{
	return (HCURSOR)ClassLong(GCL_HCURSOR);
}

LONG CControl::WndLong(int index)
{
	CheckHandle();
	return GetWindowLong(handle, index);
}

DWORD CControl::WndStyle()
{
	CheckHandle();
	return GetWindowStyle(handle);
}

DWORD CControl::WndExStyle()
{
	CheckHandle();
	return GetWindowExStyle(handle);
}

DWORD CControl::GetDlgCode(MSG *message)
{
	CheckHandle();
	return (DWORD)SendMessage(WM_GETDLGCODE, 0, (LPARAM)message);
}

void CControl::SetTabStop(bool value)
{
	if(tab_stop != value)
	{
		tab_stop = value;
		if(HandleAllocated())
		{
			DWORD style = WndStyle() & ~WS_TABSTOP;
			if(value) style |= WS_TABSTOP;
			SetWndStyle(style);
		}
	}
}

void CControl::SetGroup(bool value)
{
	if(group != value)
	{
		group = value;
		if(HandleAllocated())
		{
			DWORD style = WndStyle() & ~WS_GROUP;
			if(value) style |= WS_GROUP;
			SetWndStyle(style);
		}
	}
}

void CControl::SetWndLong(int index, LONG value)
{
	CheckHandle();
	SetWindowLong(handle, index, value);
}

void CControl::SetWndStyle(DWORD value)
{
	SetWndLong(GWL_STYLE, value);
}

void CControl::SetWndExStyle(DWORD value)
{
	SetWndLong(GWL_EXSTYLE, value);
}

void CControl::SetClientSize(int width, int height)
{
	RECT rect;
	rect.left = 0;
	rect.top = 0;
	rect.right = width;
	rect.bottom = height;
	AdjustRect(rect);
	SetSize(rect.right - rect.left, rect.bottom - rect.top);
}

void CControl::AdjustRect(RECT &rect)
{
	DWORD style = HandleAllocated() ? WndStyle() : GetWndStyle();
	DWORD ex_style = HandleAllocated() ? WndExStyle() : GetWndExStyle();
	AdjustWindowRectEx(&rect, style, FALSE, ex_style);
}

void CControl::ClientToScreen(POINT &pt)
{
	CheckHandle();
	::ClientToScreen(handle, &pt);
}

void CControl::ScreenToClient(POINT &pt)
{
	CheckHandle();
	::ScreenToClient(handle, &pt);
}

///////////////
//           //
// CMainMenu //
//           //
///////////////

CMainMenu::CMainMenu()
	: CMenu(), form(NULL)
{
}

CMainMenu::CMainMenu(HMENU hMenu)
	: CMenu(hMenu), form(NULL)
{
}

CMainMenu::CMainMenu(LPCTSTR resourceName)
	: CMenu(resourceName), form(NULL)
{
}

void CMainMenu::OnChanged()
{
	CMenu::OnChanged();
	if(form && form->HandleAllocated())
		if(!DrawMenuBar(form->Handle()))
			throw exception("Can not draw menu bar");
}

bool CMainMenu::TranslateMessage(MSG &msg)
{
	if(!form || !form->HandleAllocated()) return false;
	if(IsMnemonic(MnemonicFromMessage(msg)))
	{
		::TranslateMessage(&msg);
		DispatchMessage(&msg);
		return true;
	}
	CMenuItem * item = GetMenuItemWithShortcut(Shortcut(msg));
	if(item)
	{
		SendMessage(form->Handle(), WM_COMMAND, GET_WM_COMMAND_MPS(item->Id, NULL, 1));
		return true;
	}
	return false;
}

///////////
//       //
// CForm //
//       //
///////////

CForm::CForm()
	: CControl(), showed_once(false),
	in_modal_loop(false), exiting_modal_loop(false), modal_result(0),
	has_caption(true), minimize_box(true), maximize_box(true), system_menu(true), is_popup(false),
	form_border(fbThick), double_border(true), tool_window(false), form_position(fpCurrent), form_state(fsNormal),
	active_control(NULL), active(false),
	form_count(0), forms(NULL), mainMenu(NULL)
{
	SetBounds(0, 0, 300, 200);
	SetVisible(false);
}

CForm::~CForm()
{
	while(form_count) delete *forms;
	CMainMenu * oldMainMenu = mainMenu;
	MainMenu = NULL;
	delete oldMainMenu;
	SetParent(NULL);
}

void CForm::GetWndClass(WNDCLASSEX &wcex)
{
	CControl::GetWndClass(wcex);
	wcex.hIcon = GetIcon();
	wcex.hIconSm = GetIconSmall();
	wcex.lpszMenuName = NULL;
}

DWORD CForm::GetWndStyle()
{
	return CControl::GetWndStyle() & ~WS_CHILD & ~WS_TABSTOP & ~WS_GROUP & ~WS_VISIBLE
		| (has_caption ? WS_CAPTION : 0)
		| (minimize_box ? WS_MINIMIZEBOX : 0)
		| (maximize_box ? WS_MAXIMIZEBOX : 0)
		| (system_menu ? WS_SYSMENU : 0)
		| (is_popup ? WS_POPUP : WS_OVERLAPPED)
		| form_border
		| form_state;
}

DWORD CForm::GetWndExStyle()
{
	return CControl::GetWndExStyle()
		| (double_border ? WS_EX_DLGMODALFRAME : 0)
		| (tool_window ? WS_EX_TOOLWINDOW : 0);
}

void CForm::SetHasCaption(bool value)
{
	if(has_caption != value)
	{
		has_caption = value;
		RecreateWnd();
	}
}

void CForm::SetMinimizeBox(bool value)
{
	if(minimize_box != value)
	{
		minimize_box = value;
		RecreateWnd();
	}
}

void CForm::SetMaximizeBox(bool value)
{
	if(maximize_box != value)
	{
		maximize_box = value;
		RecreateWnd();
	}
}

void CForm::SetSystemMenu(bool value)
{
	if(system_menu != value)
	{
		system_menu = value;
		RecreateWnd();
	}
}

void CForm::SetIsPopup(bool value)
{
	if(is_popup != value)
	{
		is_popup = value;
		RecreateWnd();
	}
}

void CForm::SetFormBorder(::FormBorder value)
{
	if(form_border != value)
	{
		form_border = value;
		RecreateWnd();
	}
}

void CForm::SetDoubleBorder(bool value)
{
	if(double_border != value)
	{
		double_border = value;
		RecreateWnd();
	}
}

void CForm::SetToolWindow(bool value)
{
	if(tool_window != value)
	{
		tool_window = value;
		RecreateWnd();
	}
}

void CForm::SetFormPosition(::FormPosition value)
{
	if(!HandleAllocated()) form_position = value;
}

void CForm::SetFormState(::FormState value)
{
	if(!HandleAllocated()) form_state = value;
}

void CForm::CreateWnd()
{
	bool create = !HandleAllocated();
	if(create)
	{
		if(form_position == fpDefault || form_position == fpDefaultPosition)
			SetLocation(CW_USEDEFAULT, 0);
		if(form_position == fpDefault || form_position == fpDefaultSize)
			SetSize(CW_USEDEFAULT, 0);
		if(form_position == fpOwnerCenter && Owner())
			SetLocation(Owner()->X() + (Owner()->Width() - Width()) / 2, Owner()->Y() + (Owner()->Height() - Height()) / 2);
	}
	CControl::CreateWnd();
	RECT rect;
	GetWindowRect(handle, &rect);
	x = rect.left;
	y = rect.top;
	width = rect.right - rect.left;
	height = rect.bottom - rect.top;
	if(create && visible)
	{
		Show();
	}
	for(int i = 0; i < form_count; i++)
		forms[i]->CreateWnd();
}

void CForm::DestroyWnd()
{
	for(int i = 0; i < form_count; i++)
		forms[i]->DestroyWnd();
	CControl::DestroyWnd();
}

LRESULT CForm::OnMessage(UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
		case WM_DESTROY:
			{
				CGUIApplication * guiApp = GUIApplication();
				if(guiApp && guiApp->MainForm() == this) PostQuitMessage(0);
			}
			break;
		case WM_CLOSE:
			if(in_modal_loop)
			{
				if(!modal_result) modal_result = IDCLOSE;
				if(!OnCloseQuery()) modal_result = 0;
				else OnClose();
				exiting_modal_loop = false;
			}
			else
			{
				if(Owner()) SetForegroundWindow(Owner()->handle);
				Hide();
			}
			return 0;
		case WM_WINDOWPOSCHANGED:
			if((((LPWINDOWPOS)lParam)->flags & SWP_SHOWWINDOW) == SWP_SHOWWINDOW && !showed_once)
			{
				showed_once = true;
				PostMessage(CM_FIRSTSHOW, 0, 0);
			}
			break;
		case WM_ACTIVATE:
			active = wParam != WA_INACTIVE;
			OnActiveChanged();
			break;
		case CM_FIRSTSHOW: OnFirstShow(); return 0;
		case WM_HELP: OnHelp(); return TRUE;
	}
	return CControl::OnMessage(message, wParam, lParam);
}

void CForm::OnCommand(WORD /*id*/, WORD /*event*/, CControl * /*control*/)
{
}

LRESULT CForm::OnNotify(LPNMHDR /*nmh*/, CControl * /*control*/)
{
	return 0;//DefWndProc(WM_NOTIFY, nmh->idFrom, (LPARAM)nmh);
}

void CForm::OnHScroll(WORD /*code*/, WORD /*pos*/, CControl * /*control*/)
{
}

void CForm::OnVScroll(WORD /*code*/, WORD /*pos*/, CControl * /*control*/)
{
}

void CForm::ExitModal(int modal_result)
{
	if(in_modal_loop)
	{
		this->modal_result = modal_result;
		exiting_modal_loop = true;
		Close();
	}
}

void CForm::OnControlKillFocus(HWND /*hwnd*/, CControl * /*control*/)
{
}

void CForm::OnControlSetFocus(HWND /*hwnd*/, CControl *control)
{
	if(control == this)
	{
		if(active_control && active_control->CanFocus()) active_control->SetFocus();
		else active_control = NULL;
	}
	else
	{
		active_control = control;
		OnActiveControlChanged();
	}
}

void CForm::OnControlEnabledChanged(CControl *control)
{
	if(control == active_control && !control->CanFocus())
	{
		active_control = NULL;
		OnActiveControlChanged();
	}
}

void CForm::OnControlVisibleChanged(CControl *control)
{
	if(control == active_control && !control->CanFocus())
	{
		active_control = NULL;
		OnActiveControlChanged();
	}
}

void CForm::OnControlParentChanged(CControl *control)
{
	if(control == active_control && !ContainsControl(control))
	{
		active_control = NULL;
		OnActiveControlChanged();
	}
	CControl::OnControlParentChanged(control);
}

void CForm::SetActiveControl(CControl *value)
{
	if(active_control == value) return;
	if(value && !ContainsControl(value)) throw exception("Form does not contain the control");
	if(HandleAllocated())
	{
		if(value)
		{
			if(value->CanFocus()) value->SetFocus();
			else throw exception("Can not focus a control");
		}
		else
		{
			active_control = NULL;
			SetFocus();
		}
	}
	else
	{
		active_control = value;
		OnActiveControlChanged();
	}
}

void CForm::OnCreate()
{
	LPCTSTR menuName = GetMenuName();
	LPCTSTR accelName = GetAccelName();
	if(menuName)
	{
		CMainMenu * oldMainMenu = mainMenu;
		CMainMenu * newMainMenu = new CMainMenu(menuName);
		if(accelName) newMainMenu->AssignAccelerators(accelName);
		MainMenu = newMainMenu;
		delete oldMainMenu;
	}
	CControl::OnCreate();
	UpdateMainMenu();
}

void CForm::OnDestroy()
{
	if(!SetMenu(handle, NULL)) throw exception("Can not remove HMENU");
	CControl::OnDestroy();
}

void CForm::SetParent(CControl *value)
{
	if(parent != value)
	{
		if(value && !dynamic_cast<CForm*>(value)) throw exception("Form parent must be a form");
		if(parent) ((CForm*)parent)->RemoveForm(this);
		if(value) ((CForm*)value)->AddForm(this);
	}
}

bool CForm::IsMenuMessage(MSG &msg)
{
	//msg.message == WM_SYSKEYDOWN;
	CForm * form = FindForm(GetForegroundWindow());
	CMainMenu * mainMenu = form && form->mainMenu ? form->mainMenu : this->mainMenu;
	return mainMenu ? mainMenu->TranslateMessage(msg) : false;
}

bool CForm::IsDialogMessage(MSG &msg)
{
	if(msg.message == WM_KEYDOWN && (msg.wParam == VK_RETURN || msg.wParam == VK_ESCAPE))
	{
		if(::SendMessage(msg.hwnd, WM_GETDLGCODE, 0, (LPARAM)&msg) & (DLGC_WANTMESSAGE | DLGC_WANTALLKEYS))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		else if(msg.wParam == VK_RETURN) OnDefault(); else OnCancel();
		return true;
	}
	else return ::IsDialogMessage(handle, &msg) != 0;
}

void CForm::Show(int cmd_show)
{
	CControl::Show(cmd_show);
}

void CForm::DisableForms(vector<CForm*> &formList, CForm * owner, CForm * &foreground)
{
	if(owner != this)
	{
		int c = owner->FormCount();
		for(int i = 0; i < c; i++)
			DisableForms(formList, owner->GetForm(i), foreground);
		if(owner->Enabled())
		{
			CControl *control = Owner()->FindControl(GetCapture());
			if(control) control->SendMessage(WM_CANCELMODE, 0, 0);
			formList.Add(owner);
			if(!foreground && GetForegroundWindow() == owner->handle) foreground = owner;
			owner->SetEnabled(false);
		}
	}
}

void CForm::EnableForms(vector<CForm*> &formList, CForm * foreground)
{
	for(CForm ** i = formList.Begin(); i < formList.End(); i++)
		(*i)->SetEnabled(true);
	if(foreground) SetForegroundWindow(foreground->handle);
}

int CForm::ShowModal(int cmd_show)
{
	if(in_modal_loop) throw exception("Form is already showing modal");
	if(visible) throw exception("Can not show modal - form visible");
	modal_result = 0;
	in_modal_loop = true;
	
	vector<CForm*> formList;
	CForm * foreground = NULL;
	if(Owner())
	{
		CForm * owner = Owner();
		while(owner->Owner()) owner = owner->Owner();
		DisableForms(formList, owner, foreground);
	}

	CreateWnd();
	Show(cmd_show);
	Update();
	
	MSG msg;
	while(GetMessage(&msg, NULL, 0, 0))
	{
		try
		{
			if(!IsMenuMessage(msg) && !IsDialogMessage(msg))
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
		catch(exception &e)
		{
			ErrorBox(this, e);
		}
		catch(_com_error &e)
		{
			ErrorBox(this, e);
		}
		if(modal_result && !exiting_modal_loop) break;
	}

	EnableForms(formList, foreground);

	Hide();
	DestroyWnd();
	
	in_modal_loop = false;
	return modal_result;
}

void CForm::Close()
{
	PostMessage(WM_CLOSE, 0, 0);
}

void CForm::DrawMenuBar()
{
	if(HandleAllocated())
	{
		::DrawMenuBar(handle);
	}
}

void CForm::EnableMenuItem(WORD id, bool value)
{
	if(mainMenu)
	{
		CMenuItem * menuItem = mainMenu->GetMenuItemWithId(id);
		if(menuItem)
		{
			menuItem->Enabled = value;
		}
	}
}

void CForm::CheckMenuItem(WORD id, bool value, bool radio)
{
	if(mainMenu)
	{
		CMenuItem * menuItem = mainMenu->GetMenuItemWithId(id);
		if(menuItem)
		{
			menuItem->RadioCheck = radio;
			menuItem->Checked = value;
		}
	}
}

void CForm::SetMenuItemText(WORD id, const string &value)
{
	if(mainMenu)
	{
		CMenuItem * menuItem = mainMenu->GetMenuItemWithId(id);
		if(menuItem)
		{
			menuItem->Text = value;
		}
	}
}

bool CForm::Minimized()
{
	return HandleAllocated() ? IsMinimized(handle) != 0 : form_state == fsMinimized;
}

bool CForm::Maximized()
{
	return HandleAllocated() ? IsMaximized(handle) != 0 : form_state == fsMaximized;
}

bool CForm::Restored()
{
	return HandleAllocated() ? IsRestored(handle) != 0 : form_state == fsNormal;
}

void CForm::Minimize()
{
	CheckHandle();
	CloseWindow(handle);
}

void CForm::Maximize()
{
	CheckHandle();
	Show(SW_MAXIMIZE);
}

void CForm::Restore()
{
	CheckHandle();
	if(Minimized()) OpenIcon(handle);
	else if(Maximized()) Show(SW_RESTORE);
}

HICON CForm::Icon()
{
	return (HICON)ClassLong(GCL_HICON);
}

HICON CForm::SmallIcon()
{
	return (HICON)ClassLong(GCL_HICONSM);
}

LPCTSTR CForm::MenuName()
{
	return (LPCTSTR)ClassLong(GCL_MENUNAME);
}

void CForm::AdjustRect(RECT &rect)
{
	DWORD style = HandleAllocated() ? WndStyle() : GetWndStyle();
	DWORD ex_style = HandleAllocated() ? WndExStyle() : GetWndExStyle();
	bool menu = mainMenu != NULL;
	AdjustWindowRectEx(&rect, style, menu, ex_style);
}

void CForm::AddForm(CForm * form)
{
	if(!form) throw exception("Can not add a NULL form");
	if(form->parent) throw exception("Form already has a parent");
	CForm **new_forms = new CForm*[form_count + 1];
	for(int i = 0; i < form_count; i++) new_forms[i] = forms[i];
	new_forms[form_count] = form;
	form_count++;
	if(forms) delete[] forms;
	forms = new_forms;
	form->parent = this;
	if(HandleAllocated() && form->HandleAllocated()) ::SetParent(form->handle, handle);
}

void CForm::RemoveForm(CForm * form)
{
	int i;
	for(i = 0; i < form_count; i++)
		if(forms[i] == form) break;
	if(i != form_count)
	{
		form->parent = NULL;
		if(form->HandleAllocated()) ::SetParent(form->handle, NULL);
		form_count--;
		CForm **new_forms = new CForm*[form_count];
		int j;
		for(j = 0; j < i; j++) new_forms[j] = forms[j];
		for(; j < form_count; j++) new_forms[j] = forms[j + 1];
		delete[] forms;
		forms = new_forms;
	}
}

CForm * CForm::FindForm(HWND handle)
{
	if(!handle) return NULL;
	if(handle == this->handle) return this;
	for(int i = 0; i < form_count; i++)
	{
		CForm * form = forms[i]->FindForm(handle);
		if(form) return form;
	}
	return NULL;
}

void CForm::UpdateMainMenu()
{
	if(HandleAllocated())
		if(!SetMenu(Handle(), mainMenu ? mainMenu->Handle : NULL))
			throw exception("Can not set HMENU");
}

void CForm::set_MainMenu(CMainMenu * value)
{
	if(mainMenu != value)
	{
		if(value && value->form) throw exception("MainMenu already belongs to a Form");
		if(mainMenu) mainMenu->form = NULL;
		mainMenu = value;
		if(mainMenu) mainMenu->form = this;
		UpdateMainMenu();
	}
}

///////////////
//           //
// Functions //
//           //
///////////////

int MessageBox(CForm *owner, const string &text, const string &caption, UINT type)
{
	return MessageBox(owner ? owner->Handle() : NULL, (LPCTSTR)text, (LPCTSTR)caption, type);
}

int MessageBox(CForm *owner, const string &text, UINT type)
{
	return MessageBox(owner, text, Application() ? Application()->Name() : emptyString, type);
}

int ErrorBox(CForm *owner, exception &e, UINT type)
{
	return MessageBox(owner, AnsiToString(e.what()), type);
}

int ErrorBox(CForm *owner, exception &e, const string &caption, UINT type)
{
	return MessageBox(owner, AnsiToString(e.what()), caption, type);
}

int ErrorBox(CForm *owner, _com_error &e, UINT type)
{
	return MessageBox(owner, FormatComError(e), type);
}

int ErrorBox(CForm *owner, _com_error &e, const string &caption, UINT type)
{
	return MessageBox(owner, FormatComError(e), caption, type);
}
