#include "System.h"
#include "Options.h"
#include "Utils.h"
#include "VFDatabase.h"
#include "VFFileList.h"

VFDemoOptions vfDemoDefaultOptions;

void VFDemoInitDefaultOptions()
{
	vfDemoDefaultOptions.useLowQualityImages = true;

	vfDemoDefaultOptions.imagesFromFile.resolution = VF_IMAGE_RESOLUTION;
	vfDemoDefaultOptions.imagesFromFile.fingerId = default_file_finger_id;
	vfDemoDefaultOptions.imagesFromFile.prompt = true;

	vfDemoDefaultOptions.imagesFromScanner.fingerId = default_scanner_finger_id;
	vfDemoDefaultOptions.imagesFromScanner.prompt = false;

	vfDemoDefaultOptions.enrollment.checkMinimalMinutiaCount = true;
	vfDemoDefaultOptions.enrollment.minimalMinutiaCount = 10;
	vfDemoDefaultOptions.enrollment.searchForDuplicates = true;
	vfDemoDefaultOptions.enrollment.warnDifferentFingerId = true;

	vfDemoDefaultOptions.identification.checkForEmptyDatabase = true;
	vfDemoDefaultOptions.identification.useG = true;
	vfDemoDefaultOptions.identification.databasePercent = 100;
	vfDemoDefaultOptions.identification.matchAllDatabase = false;
	vfDemoDefaultOptions.identification.showAllResults = false;
}

static void VFDemoLoadOptionsImages(MSXML2::IXMLDOMNodePtr node, VFDemoOptionsImages &value)
{
	MSXML2::IXMLDOMNamedNodeMapPtr attributes = node->attributes;
	XmlGetAttributeMemberValue(attributes, value, fingerId, string);
	XmlGetAttributeMemberValue(attributes, value, prompt, StringToBool);
}

static void VFDemoLoadOptionsImagesFromFile(MSXML2::IXMLDOMNodePtr node, VFDemoOptionsImagesFromFile &value)
{
	VFDemoLoadOptionsImages(node, value);
	MSXML2::IXMLDOMNamedNodeMapPtr attributes = node->attributes;
	XmlGetAttributeMemberValue(attributes, value, resolution, StringToInt);
}

static void VFDemoLoadOptionsImagesFromScanner(MSXML2::IXMLDOMNodePtr node, VFDemoOptionsImagesFromScanner &value)
{
	VFDemoLoadOptionsImages(node, value);
}

static void VFDemoLoadOptionsEnrollment(MSXML2::IXMLDOMNodePtr node, VFDemoOptionsEnrollment &value)
{
	MSXML2::IXMLDOMNamedNodeMapPtr attributes = node->attributes;
	XmlGetAttributeMemberValue(attributes, value, checkMinimalMinutiaCount, StringToBool);
	XmlGetAttributeMemberValue(attributes, value, minimalMinutiaCount, StringToInt);
	XmlGetAttributeMemberValue(attributes, value, searchForDuplicates, StringToBool);
	XmlGetAttributeMemberValue(attributes, value, warnDifferentFingerId, StringToBool);
}

static void VFDemoLoadOptionsIdentification(MSXML2::IXMLDOMNodePtr node, VFDemoOptionsIdentification &value)
{
	MSXML2::IXMLDOMNamedNodeMapPtr attributes = node->attributes;
	XmlGetAttributeMemberValue(attributes, value, checkForEmptyDatabase, StringToBool);
	XmlGetAttributeMemberValue(attributes, value, useG, StringToBool);
	XmlGetAttributeMemberValue(attributes, value, databasePercent, StringToInt);
	XmlGetAttributeMemberValue(attributes, value, matchAllDatabase, StringToBool);
	XmlGetAttributeMemberValue(attributes, value, showAllResults, StringToBool);
}

void VFDemoLoadOptions(MSXML2::IXMLDOMNodePtr node, VFDemoOptions &value)
{
	MSXML2::IXMLDOMNamedNodeMapPtr attributes = node->attributes;
	XmlGetAttributeMemberValue(attributes, value, useLowQualityImages, StringToBool);

	MSXML2::IXMLDOMNodePtr imagesFromFileNode = node->selectSingleNode("ImagesFromFile");
	if((bool)imagesFromFileNode)
		VFDemoLoadOptionsImagesFromFile(imagesFromFileNode, value.imagesFromFile);
	
	MSXML2::IXMLDOMNodePtr imagesFromScannerNode = node->selectSingleNode("ImagesFromScanner");
	if((bool)imagesFromScannerNode)
		VFDemoLoadOptionsImagesFromScanner(imagesFromScannerNode, value.imagesFromScanner);
	
	MSXML2::IXMLDOMNodePtr enrollmentNode = node->selectSingleNode("Enrollment");
	if((bool)enrollmentNode)
		VFDemoLoadOptionsEnrollment(enrollmentNode, value.enrollment);
	
	MSXML2::IXMLDOMNodePtr identificationNode = node->selectSingleNode("Identification");
	if((bool)identificationNode)
		VFDemoLoadOptionsIdentification(identificationNode, value.identification);
}

static void VFDemoSaveOptionsImages(MSXML2::IXMLDOMDocumentPtr document, MSXML2::IXMLDOMNodePtr node, const VFDemoOptionsImages &value)
{
	MSXML2::IXMLDOMNamedNodeMapPtr attributes = node->attributes;
	XmlSetAttributeMemberValue(document, attributes, value, fingerId, string);
	XmlSetAttributeMemberValue(document, attributes, value, prompt, BoolToString);
}

static void VFDemoSaveOptionsImagesFromFile(MSXML2::IXMLDOMDocumentPtr document, MSXML2::IXMLDOMNodePtr node, const VFDemoOptionsImagesFromFile &value)
{
	VFDemoSaveOptionsImages(document, node, value);
	MSXML2::IXMLDOMNamedNodeMapPtr attributes = node->attributes;
	XmlSetAttributeMemberValue(document, attributes, value, resolution, IntToString);
}

static void VFDemoSaveOptionsImagesFromScanner(MSXML2::IXMLDOMDocumentPtr document, MSXML2::IXMLDOMNodePtr node, const VFDemoOptionsImagesFromScanner &value)
{
	VFDemoSaveOptionsImages(document, node, value);
}

static void VFDemoSaveOptionsEnrollment(MSXML2::IXMLDOMDocumentPtr document, MSXML2::IXMLDOMNodePtr node, const VFDemoOptionsEnrollment &value)
{
	MSXML2::IXMLDOMNamedNodeMapPtr attributes = node->attributes;
	XmlSetAttributeMemberValue(document, attributes, value, checkMinimalMinutiaCount, BoolToString);
	XmlSetAttributeMemberValue(document, attributes, value, minimalMinutiaCount, IntToString);
	XmlSetAttributeMemberValue(document, attributes, value, searchForDuplicates, BoolToString);
	XmlSetAttributeMemberValue(document, attributes, value, warnDifferentFingerId, BoolToString);
}

static void VFDemoSaveOptionsIdentification(MSXML2::IXMLDOMDocumentPtr document, MSXML2::IXMLDOMNodePtr node, const VFDemoOptionsIdentification &value)
{
	MSXML2::IXMLDOMNamedNodeMapPtr attributes = node->attributes;
	XmlSetAttributeMemberValue(document, attributes, value, checkForEmptyDatabase, BoolToString);
	XmlSetAttributeMemberValue(document, attributes, value, useG, BoolToString);
	XmlSetAttributeMemberValue(document, attributes, value, databasePercent, IntToString);
	XmlSetAttributeMemberValue(document, attributes, value, matchAllDatabase, BoolToString);
	XmlSetAttributeMemberValue(document, attributes, value, showAllResults, BoolToString);
}

void VFDemoSaveOptions(MSXML2::IXMLDOMDocumentPtr document, MSXML2::IXMLDOMNodePtr node, const VFDemoOptions &value)
{
	MSXML2::IXMLDOMNamedNodeMapPtr attributes = node->attributes;
	XmlSetAttributeMemberValue(document, attributes, value, useLowQualityImages, BoolToString);

	MSXML2::IXMLDOMElementPtr imagesFromFileNode = document->createElement("ImagesFromFile");
	VFDemoSaveOptionsImagesFromFile(document, imagesFromFileNode, value.imagesFromFile);
	node->appendChild(imagesFromFileNode);

	MSXML2::IXMLDOMElementPtr imagesFromScannerNode = document->createElement("ImagesFromScanner");
	VFDemoSaveOptionsImagesFromScanner(document, imagesFromScannerNode, value.imagesFromScanner);
	node->appendChild(imagesFromScannerNode);

	MSXML2::IXMLDOMElementPtr enrollmentNode = document->createElement("Enrollment");
	VFDemoSaveOptionsEnrollment(document, enrollmentNode, value.enrollment);
	node->appendChild(enrollmentNode);

	MSXML2::IXMLDOMElementPtr identificationNode = document->createElement("Identification");
	VFDemoSaveOptionsIdentification(document, identificationNode, value.identification);
	node->appendChild(identificationNode);
}
