#include "System.h"
#include "Streams.h"

/////////////
//         //
// CStream //
//         //
/////////////

CStream::CStream()
	: otherByteOrder(false)
{
}

CStream::~CStream()
{
}

void CStream::Write(const void * buffer, DWORD count)
{
	if(WriteBuffer(buffer, count) != count) throw exception("Can not write to file");
}

void CStream::Read(void * buffer, DWORD count)
{
	if(ReadBuffer(buffer, count) != count) throw exception("Can not read from file");
}

void CStream::SetOtherByteOrder(bool value)
{
	otherByteOrder = value;
}

void CStream::WriteByte(BYTE value)
{
	Write(&value, sizeof(BYTE));
}

void CStream::WriteSByte(signed char value)
{
	Write(&value, sizeof(signed char));
}

void CStream::WriteUShort(USHORT value)
{
	USHORT v = otherByteOrder ? MAKEWORD(HIBYTE(value), LOBYTE(value)) : value;
	Write(&v, sizeof(USHORT));
}

void CStream::WriteShort(short value)
{
	short v = otherByteOrder ? MAKEWORD(HIBYTE(value), LOBYTE(value)) : value;
	Write(&v, sizeof(short));
}

void CStream::WriteUInt(UINT value)
{
	UINT v = otherByteOrder ? MAKEDWORD(MAKEWORD(HIBYTE(HIWORD(value)), LOBYTE(HIWORD(value))), MAKEWORD(HIBYTE(LOWORD(value)), LOBYTE(LOWORD(value)))) : value;
	Write(&v, sizeof(UINT));
}

void CStream::WriteInt(int value)
{
	int v = otherByteOrder ? MAKEDWORD(MAKEWORD(HIBYTE(HIWORD(value)), LOBYTE(HIWORD(value))), MAKEWORD(HIBYTE(LOWORD(value)), LOBYTE(LOWORD(value)))) : value;
	Write(&v, sizeof(int));
}

void CStream::WriteBool(bool value)
{
	Write(&value, sizeof(bool));
}

void CStream::WriteChar(char value)
{
	Write(&value, sizeof(char));
}

void CStream::WriteFloat(float value)
{
	if(otherByteOrder) throw exception("Other byte order is not supported for float in stream");
	Write(&value, sizeof(float));
}

void CStream::WriteDouble(double value)
{
	if(otherByteOrder) throw exception("Other byte order is not supported for double in stream");
	Write(&value, sizeof(double));
}

BYTE CStream::ReadByte()
{
	BYTE value;
	Read(&value, sizeof(BYTE));
	return value;
}

signed char CStream::ReadSByte()
{
	signed char value;
	Read(&value, sizeof(signed char));
	return value;
}

USHORT CStream::ReadUShort()
{
	USHORT value;
	Read(&value, sizeof(USHORT));
	return otherByteOrder ? MAKEWORD(HIBYTE(value), LOBYTE(value)) : value;
}

short CStream::ReadShort()
{
	short value;
	Read(&value, sizeof(short));
	return otherByteOrder ? MAKEWORD(HIBYTE(value), LOBYTE(value)) : value;
}

UINT CStream::ReadUInt()
{
	UINT value;
	Read(&value, sizeof(UINT));
	return otherByteOrder ? MAKEDWORD(MAKEWORD(HIBYTE(HIWORD(value)), LOBYTE(HIWORD(value))), MAKEWORD(HIBYTE(LOWORD(value)), LOBYTE(LOWORD(value)))) : value;
}

int CStream::ReadInt()
{
	int value;
	Read(&value, sizeof(int));
	return otherByteOrder ? MAKEDWORD(MAKEWORD(HIBYTE(HIWORD(value)), LOBYTE(HIWORD(value))), MAKEWORD(HIBYTE(LOWORD(value)), LOBYTE(LOWORD(value)))) : value;
}

bool CStream::ReadBool()
{
	bool value;
	Read(&value, sizeof(bool));
	return value;
}

char CStream::ReadChar()
{
	char value;
	Read(&value, sizeof(char));
	return value;
}

float CStream::ReadFloat()
{
	if(otherByteOrder) throw exception("Other byte order is not supported for float in stream");
	float value;
	Read(&value, sizeof(float));
	return value;
}

double CStream::ReadDouble()
{
	if(otherByteOrder) throw exception("Other byte order is not supported for double in stream");
	double value;
	Read(&value, sizeof(double));
	return value;
}

void CStream::WriteString(const string &value)
{
	ansiString s = StringToAnsi(value);
	Write((LPCSTR)s, value.Length() * sizeof(char));
}

string CStream::ReadString(int length)
{
	ansiString s(length);
	Read((LPSTR)s, length * sizeof(char));
	return AnsiToString(s);
}

bool CStream::ReadLine(string &value)
{
	const int sizeofChar = sizeof(char);
	value.Clear();
	bool wasChar = false;
	CHAR ch;
	ansiString str;
	while(ReadBuffer(&ch, sizeofChar) == sizeofChar)
	{
		if(ch == '\r' || ch == '\n')
		{
			if(ch == '\r' && ReadBuffer(&ch, sizeofChar) == sizeofChar)
			{
				if(ch != '\n') SetPosition(-sizeofChar, soCurrent);
			}
			ch = '\0';
		}
		wasChar = true;
		if(!ch) break;
		str.Append(ch);
	}
	if(!wasChar) return false;
	value = AnsiToString(str);
	return true;
}

void CStream::WriteLine(const string &value)
{
	WriteString(value + newLine);
}

void CStream::WriteStringL(const string &value)
{
	WriteInt(value.Length());
	WriteString(value);
}

string CStream::ReadStringL()
{
	return ReadString(ReadInt());
}

///////////////////
//               //
// CHandleStream //
//               //
///////////////////

#ifndef INVALID_SET_FILE_POINTER
#define INVALID_SET_FILE_POINTER ((DWORD)-1)
#endif

CHandleStream::CHandleStream()
	: CStream(), handle(NULL)
{
}

CHandleStream::~CHandleStream()
{
	Close();
}

void CHandleStream::Close()
{
	if(handle)
	{
		if(!CloseHandle(handle)) throw exception("Can not close file handle");
		handle = NULL;
	}
}

bool CHandleStream::Opened()
{
	return handle != NULL;
}

void CHandleStream::SetHandle(HANDLE handle, FileAccessMode accessMode)
{
	if(!handle) throw exception("Can not set NULL handle");
	if(((accessMode & faReadWrite) == 0) || ((accessMode & ~faReadWrite) != 0)) throw exception("Invalid file access mode");
	this->handle = handle;
	this->accessMode = accessMode;
}

void CHandleStream::CheckOpen(FileAccessMode requiredAccessMode)
{
	if(!handle) throw exception("File handle not opened");
	if((accessMode & requiredAccessMode) != requiredAccessMode)
		throw exception("File handle opened with access mode that denies the operation");
}

DWORD CHandleStream::WriteBuffer(const void * buffer, DWORD count)
{
	CheckOpen(faWrite);
	DWORD bytesWriten;
	if(!WriteFile(handle, buffer, count, &bytesWriten, NULL)) throw exception("Can not write to file");
	return bytesWriten;
}

DWORD CHandleStream::ReadBuffer(void * buffer, DWORD count)
{
	CheckOpen(faRead);
	DWORD bytesRead;
	if(!ReadFile(handle, buffer, count, &bytesRead, NULL)) throw exception("Can not read from file");
	return bytesRead;
}

DWORD CHandleStream::Position()
{
	CheckOpen();
	DWORD v = SetFilePointer(handle, 0, NULL, FILE_CURRENT);
	if(v == INVALID_SET_FILE_POINTER) throw exception("Can not retrieve file position");
	return v;
}

DWORD CHandleStream::SetPosition(int value, SeekOrigin origin)
{
	CheckOpen();
	DWORD moveMethod;
	if(origin == soBegin) moveMethod = FILE_BEGIN;
	else if(origin == soCurrent) moveMethod = FILE_CURRENT;
	else if(origin == soEnd) moveMethod = FILE_END;
	else throw exception("Invalid seek origin");
	DWORD v = SetFilePointer(handle, value, NULL, moveMethod);
	if(v == INVALID_SET_FILE_POINTER) throw exception("Can not set file position");
	return v;
}

void CHandleStream::Flush()
{
	CheckOpen(faWrite);
	if(!FlushFileBuffers(handle)) throw exception("Can not flush file");
}

/////////////////
//             //
// CFileStream //
//             //
/////////////////

CFileStream::CFileStream()
	: CHandleStream()
{
}

CFileStream::CFileStream(const string &fileName, FileOpenMode openMode)
	: CHandleStream()
{
	Open(fileName, openMode);
}

CFileStream::CFileStream(const string &fileName, FileOpenMode openMode, FileAccessMode accessMode, FileShareMode shareMode)
	: CHandleStream()
{
	Open(fileName, openMode, accessMode, shareMode);
}

void CFileStream::Open(const string &fileName, FileOpenMode openMode)
{
	Open(fileName, openMode, openMode == foOpen || openMode == foOpenOrCreate ? faRead : faWrite);
}

void CFileStream::Open(const string &fileName, FileOpenMode openMode, FileAccessMode accessMode, FileShareMode shareMode)
{
	Close();
	if((openMode == foAppend || openMode == foAppendExisting || openMode == foTruncate || openMode == foTruncateAlways) && (accessMode & faRead))
		throw exception("File Append or truncate open mode can not be used with read access mode");
	DWORD creationDisposition;
	if(openMode == foCreateNew) creationDisposition = CREATE_NEW;
	else if(openMode == foCreate) creationDisposition = CREATE_ALWAYS;
	else if(openMode == foOpen || openMode == foAppendExisting) creationDisposition = OPEN_EXISTING;
	else if(openMode == foAppend || openMode == foOpenOrCreate || openMode == foTruncateAlways) creationDisposition = OPEN_ALWAYS;
	else if(openMode == foTruncate) creationDisposition = TRUNCATE_EXISTING;
	else throw exception("Invalid file open mode");
	DWORD desiredAccess = 0;
	if(accessMode & faRead) desiredAccess |= GENERIC_READ;
	if(accessMode & faWrite) desiredAccess |= GENERIC_WRITE;
	if(((accessMode & faReadWrite) == 0) || ((accessMode & ~faReadWrite) != 0)) throw exception("Invalid file access mode");
	DWORD fileShareMode = 0;
	if(shareMode & fsRead) fileShareMode |= FILE_SHARE_READ;
	if(shareMode & fsWrite) fileShareMode |= FILE_SHARE_WRITE;
	if((shareMode & ~fsReadWrite) != 0) throw exception("Invalid file share mode");
	HANDLE handle = CreateFile(fileName, desiredAccess, fileShareMode, NULL, creationDisposition, 0, NULL);
	if(handle == INVALID_HANDLE_VALUE)
	{
		handle = NULL;
		throw exception("Can not open file");
	}
	SetHandle(handle, accessMode);
	if(openMode == foAppend || openMode == foAppendExisting)
	{
		if(SetFilePointer(handle, 0, NULL, FILE_END) == INVALID_SET_FILE_POINTER)
		{
			CloseHandle(handle);
			handle = NULL;
			throw exception("Can not append file");
		}
	}
	else if(openMode == foTruncateAlways)
	{
		SetEndOfFile();
	}
}

void CFileStream::SetEndOfFile()
{
	CheckOpen(faWrite);
	if(!::SetEndOfFile(Handle())) throw exception("Can not set end of file");
}

void CFileStream::SetSize(int value)
{
	SetPosition(value);
	SetEndOfFile();
}

void CFileStream::Truncate()
{
	SetSize(0);
}
