#pragma once

enum SeekOrigin { soBegin, soCurrent, soEnd };

class CStream
{
private:
	bool otherByteOrder;

public:
	CStream();
	virtual ~CStream();

	virtual void Close() = 0;

	virtual DWORD WriteBuffer(const void * buffer, DWORD count) = 0;
	virtual DWORD ReadBuffer(void * buffer, DWORD count) = 0;

	virtual DWORD Position() = 0;
	virtual DWORD SetPosition(int value, SeekOrigin origin = soBegin) = 0;

	virtual void Flush() = 0;

	void Write(const void * buffer, DWORD count);
	void Read(void * buffer, DWORD count);

	bool OtherByteOrder() { return otherByteOrder; }
	void SetOtherByteOrder(bool value);

	void WriteByte(BYTE value);
	void WriteSByte(signed char value);
	void WriteUShort(USHORT value);
	void WriteShort(short value);
	void WriteUInt(UINT value);
	void WriteInt(int value);
	void WriteBool(bool value);
	void WriteChar(char value);
	void WriteFloat(float value);
	void WriteDouble(double value);

	BYTE ReadByte();
	signed char ReadSByte();
	USHORT ReadUShort();
	short ReadShort();
	UINT ReadUInt();
	int ReadInt();
	bool ReadBool();
	char ReadChar();
	float ReadFloat();
	double ReadDouble();

	void WriteString(const string &value);
	string ReadString(int length);

	bool ReadLine(string &value);
	void WriteLine(const string &value);

	void WriteStringL(const string &value);
	string ReadStringL();
};

enum FileOpenMode { foAppend, foAppendExisting, foCreate, foCreateNew, foOpen, foOpenOrCreate, foTruncate, foTruncateAlways};
enum FileAccessMode { faRead = 1, faWrite = 2, faReadWrite = faRead | faWrite};
enum FileShareMode { fsNone = 0, fsRead = 1, fsWrite = 2, fsReadWrite = fsRead | fsWrite};

class CHandleStream: public CStream
{
private:
	HANDLE handle;
	FileAccessMode accessMode;

protected:
	void SetHandle(HANDLE handle, FileAccessMode accessMode);
	void CheckOpen(FileAccessMode requiredAccessMode = (FileAccessMode)0);

public:
	CHandleStream();
	~CHandleStream();

	void Close();

	HANDLE Handle() { return handle; }

	bool Opened();

	DWORD WriteBuffer(const void * buffer, DWORD count);
	DWORD ReadBuffer(void * buffer, DWORD count);

	DWORD Position();
	DWORD SetPosition(int value, SeekOrigin origin = soBegin);

	void Flush();
};

class CFileStream: public CHandleStream
{
public:
	CFileStream();
	CFileStream(const string &fileName, FileOpenMode openMode = foOpen);
	CFileStream(const string &fileName, FileOpenMode openMode, FileAccessMode accessMode, FileShareMode shareMode = fsNone);

	void Open(const string &fileName, FileOpenMode openMode = foOpen);
	void Open(const string &fileName, FileOpenMode openMode, FileAccessMode acessMode, FileShareMode shareMode = fsNone);

	void SetEndOfFile();
	void SetSize(int value);
	void Truncate();
};
