#pragma once

#include "Vectors.h"

// Only CHAR and WCHAR are supported

inline CHAR ToUpper(CHAR c)
{
	return (CHAR)toupper(c);
}

inline WCHAR ToUpper(WCHAR c)
{
	return towupper(c);
}

inline CHAR ToLower(CHAR c)
{
	return (CHAR)tolower(c);
}

inline WCHAR ToLower(WCHAR c)
{
	return towlower(c);
}

inline bool IsDigit(CHAR c)
{
	return isdigit(c) != 0;
}

inline bool IsDigit(WCHAR c)
{
	return iswdigit(c) != 0;
}

inline int StringLength(const CHAR * s)
{
	return strlen(s);
}

inline int StringLength(const WCHAR * s)
{
	return wcslen(s);
}

inline int CompareStrings(const CHAR * s1, const CHAR * s2, bool ignoreCase = false)
{
	return ignoreCase ? stricmp(s1, s2) : strcmp(s1, s2);
}

inline int CompareStrings(const WCHAR * s1, const WCHAR * s2, bool ignoreCase = false)
{
	return ignoreCase ? wcsicmp(s1, s2) : wcscmp(s1, s2);
}

inline bool StringsEqual(const CHAR * s1, const CHAR * s2, bool ignoreCase = false)
{
	return CompareStrings(s1, s2, ignoreCase) == 0;
}

inline bool StringsEqual(const WCHAR * s1, const WCHAR * s2, bool ignoreCase = false)
{
	return CompareStrings(s1, s2, ignoreCase) == 0;
}

inline void CopyString(CHAR * dst, const CHAR * src, int l)
{
	memcpy(dst, src, l * sizeof(CHAR));
}

inline void CopyString(WCHAR * dst, const WCHAR * src, int l)
{
	memcpy(dst, src, l * sizeof(WCHAR));
}

inline void MoveString(CHAR * dst, CHAR * src, int l)
{
	memmove(dst, src, l * sizeof(CHAR));
}

inline void MoveString(WCHAR * dst, WCHAR * src, int l)
{
	memmove(dst, src, l * sizeof(WCHAR));
}

#define baseString_AlignLength(l, a) ((l / a + (l % a ? 1 : 0)) * a)
#define baseString_AlignCapacity(l, m, a) l <= m ? baseString_AlignLength(l, a) :

template<class T> class baseString
{
private:
	typedef baseString<T> InstanceType;

	DWORD * ref;
	T * str, * end;
	int length;
	int capacity;
	static T * emptyString;

	void SetCapacity(int l)
	{
		if(!l)
		{
			if(!emptyString)
			{
				DWORD * emptyRef = (DWORD *)malloc(sizeof(T) + sizeof(DWORD));
				*emptyRef = 0;
				emptyString = (T *)(emptyRef + 1);
			}
			capacity = 1;
			str = emptyString;
			ref = ((DWORD *)emptyString - 1);
			(*ref)++;
		}
		else
		{
			l++;
			int c =
				baseString_AlignCapacity(l, 8, 1)
				baseString_AlignCapacity(l, 16, 2)
				baseString_AlignCapacity(l, 32, 4)
				baseString_AlignCapacity(l, 64, 8)
				baseString_AlignCapacity(l, 128, 16)
				baseString_AlignCapacity(l, 256, 32)
				baseString_AlignCapacity(l, 512, 64)
				baseString_AlignCapacity(l, 1024, 128)
				baseString_AlignLength(l, 256);
			if(c != capacity || !str)
			{
				ref = (DWORD *)realloc(str ? ref : NULL, c * sizeof(T) + sizeof(DWORD));
				if(!ref) throw exception("Out of memory in string");
				str = (T *)(ref + 1);
				*ref = 1;
				capacity = c;
			}
		}
	}

	void MakeUnique(int l = -1)
	{
		if(!str || *ref > 1 || str == emptyString)
		{
			T * s;
			if(str)
			{
				(*ref)--;
				if(str == emptyString && !(*ref))
				{
					free(ref);
					emptyString = NULL;
					s = NULL;
				}
				else s = str;
				str = NULL;
			}
			else
			{
				s = NULL;
				length = 0;
			}
			if(l < 0) l = length;
			SetCapacity(l);
			CopyString(str, s, length > l ? l : length);
			length = l;
			end = str + l;
			*end = 0;
		}
		else
		{
			if(l >= 0 && l != length)
			{
				SetCapacity(l);
				length = l;
				end = str + l;
				*end = 0;
			}
		}
	}

	void CheckIndex(int i, bool end = true) const
	{
		if(i < 0 || (end && i > length) || (!end && i >= length)) throw exception("String index out of bounds");
	}

	void AppendString(const T * s, int l)
	{
		if(!l) return;
		MakeUnique(length + l);
		CopyString(str + length - l, s, l);
	}

	void InsertString(int o, const T * s, int l)
	{
		CheckIndex(o);
		if(!l) return;
		MakeUnique(length + l);
		MoveString(str + o + l, str + o, length - o - l);
		CopyString(str + o, s, l);
	}

	int IndexOfString(const T * s, int l, int o) const
	{
		CheckIndex(o);
		if(!l) return -1;
		const T * ss, * s1, * s2;
		int i, ii;
		for(ss = str, i = o; i < length; ss++, i++)
			for(s1 = ss, s2 = s, ii = 0; ii <= l; s1++, s2++, ii++)
				if(!*s2) return i;
				else if(*s1 != *s2) break;
		return -1;
	}

	int LastIndexOfString(const T * s, int l, int o) const
	{
		CheckIndex(o);
		if(!l) return -1;
		const T * ss, * s1, * s2;
		int i, ii, j = -1;
		for(ss = str, i = o; i < length; ss++, i++)
			for(s1 = ss, s2 = s, ii = 0; ii <= l; s1++, s2++, ii++)
				if(!*s2)
				{
					j = i;
					break;
				}
				else if(*s1 != *s2) break;
		return j;
	}

	bool StartsWithString(const T * s, int l) const
	{
		if(l > length) return false;
		const T * e = s + l;
		for(const T * i = str, * j = s; j < e; i++, j++)
			if(*i != *j) return false;
		return true;
	}

	bool EndsWithString(const T * s, int l) const
	{
		if(l > length) return false;
		for(const T * i = end - 1, * j = s + l - 1; j >= s; i--, j--)
			if(*i != *j) return false;
		return true;
	}

public:
	baseString()
		: str(NULL)
	{
		Clear();
	}

	explicit baseString(int length)
		: str(NULL)
	{
		SetLength(length);
	}

	baseString(int length, T value)
		: str(NULL)
	{
		Assign(length, value);
	}

	baseString(const T * value)
		: str(NULL)
	{
		Assign(value);
	}

	baseString(const InstanceType &value)
		: str(NULL)
	{
		Assign(value);
	}

	baseString(const _bstr_t &value)
		: str(NULL)
	{
		Assign((const T *)value);
	}

	~baseString()
	{
		if(!(--(*ref))) free(ref);
	}

	bool IsEmpty() const
	{
		return length == 0;
	}

	int Length() const
	{
		return length;
	}

	void SetLength(int value)
	{
		MakeUnique(value);
	}

	void Update()
	{
		SetLength(StringLength(str));
	}

	T * Data()
	{
		MakeUnique();
		return str;
	}

	const T * Data() const
	{
		return str;
	}

	void Clear()
	{
		SetLength(0);
	}

	void Assign(int l, T c)
	{
		SetLength(l);
		Fill(c);
	}

	void Assign(const T * s)
	{
		if(s == str) return;
		SetLength(StringLength(s));
		CopyString(str, s, length);
	}

	void Assign(const InstanceType &s)
	{
		if(&s == this) return;
		if(str && !(--(*ref))) free(ref);
		ref = s.ref;
		str = s.str;
		end = s.end;
		length = s.length;
		capacity = s.capacity;
		(*ref)++;
	}

	void Copy(T * s, int o = 0, int l = -1) const
	{
		int ll = length - o;
		if(l < 0 || l > ll) l = ll;
		CopyString(s, str + o, l);
		s[l] = 0;
	}

	operator T * ()
	{
		MakeUnique();
		return str;
	}

	operator const T * () const
	{
		return str;
	}

	T * Begin()
	{
		MakeUnique();
		return str;
	}
	
	const T * Begin() const
	{
		return str;
	}

	T * End()
	{
		MakeUnique();
		return end;
	}

	const T * End() const
	{
		return end;
	}

	T & operator[](int i)
	{
		CheckIndex(i, false);
		MakeUnique();
		return str[i];
	}

	const T & operator[](int i) const
	{
		CheckIndex(i, false);
		return str[i];
	}

	T & First()
	{
		CheckIndex(0, false);
		return *str;
	}

	const T & First() const
	{
		CheckIndex(0, false);
		return *str;
	}

	T & Last()
	{
		CheckIndex(length - 1, false);
		return *(end - 1);
	}

	const T & Last() const
	{
		CheckIndex(length - 1, false);
		return *(end - 1);
	}

	void Fill(T c, int o = 0, int l = -1)
	{
		if(!c) return;
		MakeUnique();
		int e = o + (l < 0 ? length : l);
		if(e > length) e = length;
		T * es = str + e;
		for(T * s = str + o; s < es; s++) *s = c;
	}

	InstanceType Substring(int o = 0, int l = -1) const
	{
		CheckIndex(o);
		int ll = length - o;
		ll = l < 0 || l >= ll ? ll : l;
		InstanceType s(ll);
		CopyString(s.str, str + o, ll);
		return s;
	}

	InstanceType LeftSubstring(int c) const
	{
		return Substring(0, c);
	}

	InstanceType RightSubstring(int c) const
	{
		return Substring(length - c, c);
	}

	void Append(T c)
	{
		if(!c) return;
		MakeUnique(length + 1);
		str[length - 1] = c;
	}

	void Append(const T * s)
	{
		AppendString(s, StringLength(s));
	}

	void Append(const InstanceType &s)
	{
		AppendString(s.str, s.length);
	}

	InstanceType & operator = (T c)
	{
		Assign(1, c);
		return *this;
	}

	InstanceType & operator = (const T * s)
	{
		Assign(s);
		return *this;
	}

	InstanceType & operator = (const InstanceType &s)
	{
		Assign(s);
		return *this;
	}

	InstanceType & operator += (T c)
	{
		Append(c);
		return *this;
	}

	InstanceType & operator += (const T * s)
	{
		Append(s);
		return *this;
	}

	InstanceType & operator += (const InstanceType &s)
	{
		Append(s);
		return *this;
	}

	int IndexOf(T c, int o = 0) const
	{
		CheckIndex(o);
		if(!c) return -1;
		T * s = str + o;
		for(int i = o; i < length; i++, s++)
			if(*s == c) return i;
		return -1;
	}

	int IndexOf(const T * s, int o = 0) const
	{
		return IndexOfString(s, StringLength(s), o);
	}

	int IndexOf(const InstanceType &s, int o = 0) const
	{
		return IndexOfString(s.str, s.length, o);
	}

	int LastIndexOf(T c, int o = 0) const
	{
		CheckIndex(o);
		if(!c) return -1;
		T * s = end - 1;
		for(int i = length - 1; i >= o; i--, s--)
			if(*s == c) return i;
		return -1;
	}

	int LastIndexOf(const T * s, int o = 0) const
	{
		return LastIndexOfString(s, StringLength(s), o);
	}

	int LastIndexOf(const InstanceType &s, int o = 0) const
	{
		return LastIndexOfString(s.str, s.length, o);
	}

	void Insert(int o, T c)
	{
		CheckIndex(o);
		if(c == TEXT('\0')) return;
		MakeUnique(length + 1);
		MoveString(str + o + 1, str + o, length - o - 1);
		str[o] = c;
	}

	void Insert(int o, const T * s)
	{
		InsertString(o, s, StringLength(s));
	}

	void Insert(int o, const InstanceType &s)
	{
		InsertString(o, s.str, s.length);
	}

	void AppendFront(T c)
	{
		Insert(0, c);
	}

	void AppendFront(const T * s)
	{
		Insert(0, s);
	}

	void AppendFront(const InstanceType &s)
	{
		Insert(0, s);
	}

	void Remove(int i, int c = 1)
	{
		if(c <= 0) return;
		CheckIndex(i, false);
		MakeUnique();
		MoveString(str + i, str + i + c, length - i - c);
		SetLength(length - c);
	}

	void RemoveFirst(int c = 1)
	{
		Remove(0, c);
	}

	void RemoveLast(int c = 1)
	{
		Remove(length - c, c);
	}

	void TrimLeft()
	{
		int c = 0;
		for(T * i = str; i < end; i++)
			if(*i != ' ') break;
			else c++;
		if(c) Remove(0, c);
	}

	void TrimRight()
	{
		int c = 0;
		for(T * i = end - 1; i >= str; i--)
			if(*i != ' ') break;
			else c++;
		if(c) Remove(length - c, c);
	}

	void Trim()
	{
		TrimLeft();
		TrimRight();
	}

	void Split(const InstanceType &separator, vector<InstanceType> &ss)
	{
		ss.Clear();
		int i = 0, j = -1;
		for(T * s = str; s <= end; s++, i++)
			if(s == end || separator.IndexOf(*s) != -1)
			{
				ss.Add(Substring(j + 1, i - j - 1));
				j = i;
			}
	}

	void Split(T separator, vector<InstanceType> &ss)
	{
		ss.Clear();
		int i = 0, j = -1;
		for(T * s = str; s <= end; s++, i++)
			if(s == end || separator == *s)
			{
				ss.Add(Substring(j + 1, i - j - 1));
				j = i;
			}
	}

	void ToUpper()
	{
		for(T * i = str; i < end; i++)
			*i = ::ToUpper(*i);
	}

	void ToLower()
	{
		for(T * i = str; i < end; i++)
			*i = ::ToLower(*i);
	}

	void ToToggle()
	{
		for(T * i = str; i < end; i++)
			*i = i == str || *(i - 1) == ' ' ? ::ToUpper(*i) : ::ToLower(*i);
	}

	bool Contains(T value) const
	{
		return IndexOf(value) != -1;
	}

	bool Contains(const T * value) const
	{
		return IndexOf(value) != -1;
	}

	bool StartsWith(T value) const
	{
		return !IsEmpty() && First() == value;
	}

	bool EndsWith(T value) const
	{
		return !IsEmpty() && Last() == value;
	}

	bool StartsWith(const T * value) const
	{
		return StartsWithString(value, StringLength(value));
	}

	bool EndsWith(const T * value) const
	{
		return EndsWithString(value, StringLength(value));
	}

	bool StartsWith(const InstanceType &value) const
	{
		return StartsWithString(value, value.Length());
	}

	bool EndsWith(const InstanceType &value) const
	{
		return EndsWithString(value, value.Length());
	}

	void PadRight(int l, T c = ' ')
	{
		if(l > length)
		{
			int oldLen = length;
			Length = l;
			if(!c) return;
			for(T * i = str + oldLen; i < end; i++) *i = c;
		}
	}

	void PadLeft(int l, T c = ' ')
	{
		if(l > length)
		{
			int oldLen = length;
			Length = l;
			T * e = str + length - oldLen;
			MoveString(e, str, oldLen);
			if(!c) return;
			for(T * i = str; i < e; i++) *i = c;
		}
	}

	int Compare(const T * value, bool ignoreCase = false) const
	{
		return CompareStrings(str, value, ignoreCase);
	}

	bool Equals(const T * value, bool ignoreCase = false) const
	{
		return StringsEqual(str, value, ignoreCase);
	}

	operator _bstr_t() const
	{
		return _bstr_t(str);
	}

};

inline baseString<CHAR> ConcatenateStrings(const CHAR * s1, int l1, const CHAR * s2, int l2)
{
	baseString<CHAR> s(l1 + l2);
	CopyString(s.Begin(), s1, l1);
	CopyString(s.Begin() + l1, s2, l2);
	return s;
}

inline baseString<WCHAR> ConcatenateStrings(const WCHAR * s1, int l1, const WCHAR * s2, int l2)
{
	baseString<WCHAR> s(l1 + l2);
	CopyString(s.Begin(), s1, l1);
	CopyString(s.Begin() + l1, s2, l2);
	return s;
}

template<class T> inline baseString<T> operator+(const baseString<T> &s, T c)
{
	int l = s.Length();
	baseString<T> ns(l + 1);
	CopyString(ns.Begin(), s.Begin(), l);
	ns.Last() = c;
	*ns.End() = 0;
	return ns;
}

template<class T> inline baseString<T> operator+(T c, const baseString<T> &s)
{
	int l = s.Length();
	baseString<T> ns(l + 1);
	ns.First() = c;
	CopyString(ns.Begin() + 1, s.Begin(), l);
	return ns;
}

template<class T> inline baseString<T> operator + (const baseString<T> &s1, const T * s2)
{
	return ConcatenateStrings(s1.Begin(), s1.Length(), s2, StringLength(s2));
}

template<class T> inline baseString<T> operator + (const T * s1, const baseString<T> &s2)
{
	return ConcatenateStrings(s1, StringLength(s1), s2.Begin(), s2.Length());
}

template<class T> inline baseString<T> operator + (const baseString<T> &s1, const baseString<T> &s2)
{
	return ConcatenateStrings(s1.Begin(), s1.Length(), s2.Begin(), s2.Length());
}

template<class T> inline bool operator == (const baseString<T> &s1, const baseString<T> &s2)
{
	return StringsEqual(s1.Data(), s2.Data());
}

template<class T> inline bool operator == (const baseString<T> &s1, const T * s2)
{
	return StringsEqual(s1.Data(), s2);
}

template<class T> inline bool operator == (const T * s1, const baseString<T> &s2)
{
	return StringsEqual(s1, s2.Data());
}

template<class T> inline bool operator != (const baseString<T> &s1, const baseString<T> &s2)
{
	return !StringsEqual(s1.Data(), s2.Data());
}

template<class T> inline bool operator != (const baseString<T> &s1, const T * s2)
{
	return !StringsEqual(s1.Data(), s2);
}

template<class T> inline bool operator != (const T * s1, const baseString<T> &s2)
{
	return !StringsEqual(s1, s2.Data());
}

template<class T> inline bool operator > (const baseString<T> &s1, const baseString<T> &s2)
{
	return CompareStrings(s1.Data(), s2.Data()) > 0;
}

template<class T> inline bool operator > (const baseString<T> &s1, const T * s2)
{
	return CompareStrings(s1.Data(), s2) > 0;
}

template<class T> inline bool operator > (const T * s1, const baseString<T> &s2)
{
	return CompareStrings(s1, s2.Data()) > 0;
}

template<class T> inline bool operator >= (const baseString<T> &s1, const baseString<T> &s2)
{
	return CompareStrings(s1.Data(), s2.Data()) >= 0;
}

template<class T> inline bool operator >= (const baseString<T> &s1, const T * s2)
{
	return CompareStrings(s1.Data(), s2) >= 0;
}

template<class T> inline bool operator >= (const T * s1, const baseString<T> &s2)
{
	return CompareStrings(s1, s2.Data()) >= 0;
}

template<class T> inline bool operator < (const baseString<T> &s1, const baseString<T> &s2)
{
	return CompareStrings(s1.Data(), s2.Data()) < 0;
}

template<class T> inline bool operator < (const baseString<T> &s1, const T * s2)
{
	return CompareStrings(s1.Data(), s2) < 0;
}

template<class T> inline bool operator < (const T * s1, const baseString<T> &s2)
{
	return CompareStrings(s1, s2.Data()) < 0;
}

template<class T> inline bool operator <= (const baseString<T> &s1, const baseString<T> &s2)
{
	return CompareStrings(s1.Data(), s2.Data()) <= 0;
}

template<class T> inline bool operator <= (const baseString<T> &s1, const T * s2)
{
	return CompareStrings(s1.Data(), s2) <= 0;
}

template<class T> inline bool operator <= (const T * s1, const baseString<T> &s2)
{
	return CompareStrings(s1, s2.Data()) <= 0;
}

typedef baseString<CHAR> ansiString;
typedef baseString<WCHAR> wideString;

#ifdef _UNICODE
typedef wideString string;
#else
typedef ansiString string;
#endif

#define STRING(s) string(TEXT(s))

inline wideString AnsiToWideString(const CHAR * s)
{
	int c = MultiByteToWideChar(CP_ACP, 0, s, -1, NULL, 0);
	wideString str(c);
	MultiByteToWideChar(CP_ACP, 0, s, -1, str, c);
	str.Update();
	return str;
}

inline ansiString WideToAnsiString(const WCHAR * s)
{
	int c = WideCharToMultiByte(CP_ACP, 0, s, -1, NULL, 0, NULL, NULL);
	ansiString str(c);
	WideCharToMultiByte(CP_ACP, 0, s, -1, str, c, NULL, NULL);
	str.Update();
	return str;
}

inline string AnsiToString(const CHAR * s)
{
#ifdef _UNICODE
	return AnsiToWideString(s);
#else
	return s;
#endif
}

inline string AnsiToString(const ansiString &s)
{
#ifdef _UNICODE
	return AnsiToWideString(s);
#else
	return s;
#endif
}

inline ansiString StringToAnsi(const TCHAR * s)
{
#ifdef _UNICODE
	return WideToAnsiString(s);
#else
	return s;
#endif
}

inline ansiString StringToAnsi(const string &s)
{
#ifdef _UNICODE
	return WideToAnsiString(s);
#else
	return s;
#endif
}

inline string WideToString(const WCHAR * s)
{
#ifdef _UNICODE
	return s;
#else
	return WideToAnsiString(s);
#endif
}

inline string WideToString(const wideString &s)
{
#ifdef _UNICODE
	return s;
#else
	return WideToAnsiString(s);
#endif
}

inline wideString StringToWide(const TCHAR * s)
{
#ifdef _UNICODE
	return s;
#else
	return AnsiToWideString(s);
#endif
}

inline wideString StringToWide(const string &s)
{
#ifdef _UNICODE
	return s;
#else
	return AnsiToWideString(s);
#endif
}

extern const string emptyString;

extern const string newLine;

typedef vector<string> strings;
