#include "System.h"
#include "Utils.h"
#include <math.h>
#include <shlobj.h>

// String routines

static const int max_format_string = 1000;

string Format(LPCTSTR format, ...)
{
	string s(max_format_string);
	va_list list;
	va_start(list, format);
	_vsntprintf(s, max_format_string, format, list);
	va_end(list);
	s.Update();
	return s;
}

string Format(DWORD id, ...)
{
	CheckApplication();
	va_list list;
	va_start(list, id);
	string s = Application()->LoadString(id, list);
	va_end(list);
	return s;
}

string FormatV(LPCTSTR format, va_list list)
{
	string s(max_format_string);
	_vsntprintf(s, max_format_string, format, list);
	s.Update();
	return s;
}

string FormatV(DWORD id, va_list list)
{
	CheckApplication();
	return Application()->LoadString(id, list);
}

string AdjustLineBreaks(const string &s)
{
	string d;
	for(const TCHAR * i = s.Begin(); i < s.End() && *i; i++)
	{
		if(*i != '\r' || *(i + 1) != '\n')
		{
			if(*i == '\n' || *i == '\r')
			{
				d.Append('\r');
				d.Append('\n');
			}
			else d.Append(*i);
		}
	}
	return d;
}

string GetOneLine(const string &s)
{
	int j = s.Length(), j1;
	j1 = s.IndexOf('\r');
	if(j1 != -1 && j1 < j) j = j1;
	j1 = s.IndexOf('\n');
	if(j1 != -1 && j1 < j) j = j1;
	return s.LeftSubstring(j);
}

// File name and path routines

const TCHAR directorySeparator = TEXT('\\');
const TCHAR altDirectorySeparator = TEXT('/');
const TCHAR volumeSeparator = TEXT(':');

bool IsValidPathChar(TCHAR c)
{
	return c > 31 && c != TEXT('<') && c != TEXT('>') && c != TEXT(':') && c != TEXT('\"')
		&& c != TEXT('/') && c != TEXT('\\') && c != TEXT('|')
		/*&& c != TEXT('*') && c != TEXT('?')*/; // theese are valid when making filter
}

void CheckValidPathChar(TCHAR c)
{
	if(!IsValidPathChar(c)) throw exception("Invalid char in path");
}

void CheckValidPathToken(const string &s)
{
	for(const TCHAR * i = s.Begin(); i < s.End(); i++)
		CheckValidPathChar(*i);
}

string GetCleanedPath(const string &path)
{
	string s(path);
	int i;
	string ss;
	bool wasSeparator = true;
	bool wasChar = false;
	int e = -1;
	for(i = s.Length() - 1; i >= -1; i--)
	{
		// Change slash to backslash
		if(i != -1 && s[i] == altDirectorySeparator) s[i] = directorySeparator;
		// Remove space
		if(i != -1 && s[i] == ' ')
		{
			if(e == -1) e = i;
		}
		else if(i == -1 || s[i] == directorySeparator)
		{
			// Remove space from separator to char or separator
			if(e != -1)
			{
				s.Remove(i + 1, e - i);
				e = -1;
			}
			wasSeparator = true;
			wasChar = false;
		}
		else
		{
			// Remove space from char to separator
			if(e != -1 && wasSeparator)
			{
				s.Remove(i + 1, e - i);
			}
			e = -1;
			wasSeparator = false;
			wasChar = true;
		}
	}
	bool isUNC = false;
	// Starts with two backslashes?
	if(s.Length() >= 2)
	{
		if(s[0] == directorySeparator && s[1] == directorySeparator)
		{
			isUNC = true;
			// Text until backslash
			wasSeparator = false;
			if(s.Length() > 2)
			{
				i = s.IndexOf(directorySeparator, 2);
				if(i == -1)
				{
					ss = s;
					i = s.Length();
				}
				else
				{
					ss = s.Substring(2, i - 2);
					i++;
					wasSeparator = true;
				}
			}
			else
			{
				i = s.Length();
				ss = emptyString;
			}
			// Question?
			if(wasSeparator && ss.Equals(TEXT("?"), true))
			{
				// Remove it
				s.Remove(0, i);
				// Text until backslash
				wasSeparator = false;
				if(s.IsEmpty())
				{
					ss = emptyString;
					i = s.Length();
				}
				else
				{
					i = s.IndexOf(directorySeparator);
					if(i == -1)
					{
						ss = s;
						i = s.Length();
					}
					else
					{
						ss = s.LeftSubstring(i);
						i++;
						wasSeparator = true;
					}
				}
				// UNC?
				if(wasSeparator && ss.Equals(TEXT("UNC"), true))
				{
					// Remove it
					s.RemoveFirst(i);
					// Add to backslashes
					s.AppendFront(directorySeparator);
					s.AppendFront(directorySeparator);
				}
				else isUNC = false;
			}
		}
	}
	if(isUNC)
	{
		if(s.Length() > 2)
		{
			i = s.IndexOf(directorySeparator, 2);
			if(i == -1) i = s.Length();
			ss = s.Substring(2, i - 2);
			ss.ToToggle();
			s = emptyString + directorySeparator + directorySeparator + ss + s.Substring(i);
			i = 1 + 1 + ss.Length() + 1; // plus two directorySeparator + 1
		}
		else
		{
			i = s.Length();
			ss = emptyString;
		}
	}
	else
	{
		if(!s.IsEmpty())
		{
			i = s.IndexOf(directorySeparator);
			if(i == -1) i = s.Length();
			ss = s.LeftSubstring(i);
			if(ss.Length() > 0 && ss[ss.Length() - 1] == volumeSeparator)
			{
				ss.RemoveLast();
				ss.TrimRight();
				ss.ToToggle();
				s = ss + volumeSeparator + s.Substring(i);
				i = ss.Length() + 1 + 1; // plus volumeSeparator + 1
				if(i != s.Length()) i++; // plus directorySeparator
			}
		}
		else
		{
			i = 0;
			ss = emptyString;
		}
	}
	CheckValidPathToken(ss);
	for(TCHAR * c = s.Begin() + i; c < s.End(); c++)
		if(*c != directorySeparator) CheckValidPathChar(*c);
	return s;
}

string GetCurrentDirectory()
{
	int l = ::GetCurrentDirectory(0, NULL);
	string s(l - 1);
	::GetCurrentDirectory(l, s);
	return s;
}

string GetFileExtension(const string &path)
{
	string s = GetCleanedPath(path);
	int i = s.LastIndexOf('.');
	int j = s.LastIndexOf(directorySeparator);
	return i <= j /* also if i = -1 */ ? emptyString : s.RightSubstring(s.Length() - i - 1);
}

string GetFilePathWithoutExtension(const string &path)
{
	string s = GetCleanedPath(path);
	int i = s.LastIndexOf('.');
	int j = s.LastIndexOf(directorySeparator);
	return i <= j /* also if i = -1 */ ? s : s.LeftSubstring(i);
}

bool FileNameHasNoExtension(const string &path)
{
	string s = GetCleanedPath(path);
	int i = s.LastIndexOf(TEXT('.'));
	int j = s.LastIndexOf(directorySeparator);
	return i <= j /* also if i = -1 */ || i == s.Length() - 1;
}

string ChangeFileExtension(const string &path, const string &ext)
{
	string e(ext);
	while(!e.IsEmpty() && e.First() == TEXT('.')) e.RemoveFirst();
	while(!e.IsEmpty() && e.Last() == TEXT('.')) e.RemoveLast();
	string s = GetFilePathWithoutExtension(path);
	if(!e.IsEmpty()) s += TEXT('.') + e;
	return s;
}

string GetFileName(const string &path)
{
	string s = GetCleanedPath(path);
	int i = s.LastIndexOf(directorySeparator);
	return i == -1 ? s : s.Substring(i + 1, s.Length() - i - 1);
}

string GetFileNameWithoutExtension(const string &path)
{
	string s = GetCleanedPath(path);
	int i = s.LastIndexOf(directorySeparator);
	if(i == -1) i = 0; else i++;
	int c = s.LastIndexOf(TEXT('.'));
	if(c == -1) c = s.Length() - i; else c -= i;
	return s.Substring(i, c);
}

bool IsUNCPath(const string &path)
{
	string s = GetCleanedPath(path);
	return s.Length() >= 2 && s[0] == directorySeparator && s[1] == directorySeparator;
}

string GetDriveName(const string &path)
{
	string s = GetCleanedPath(path);
	if(IsUNCPath(path))
	{
		int i = s.IndexOf(directorySeparator, 2);
		return s.Substring(2, i == -1 ? -1 : i - 2);
	}
	else
	{
		int i = s.IndexOf(TEXT(':'));
		return i == -1 ? emptyString : s.LeftSubstring(i);
	}
}

string GetDirectoryName(const string &path)
{
	string s = GetCleanedPath(path);
	int o = IsUNCPath(path) ? 2 : 0;
	int i = s.LastIndexOf(directorySeparator, o);
	if(i == -1) i = s.Length();
	return s.Substring(0, i);
}

string GetSubdirectories(const string &path)
{
	string s = GetCleanedPath(path);
	int i;
	if(IsUNCPath(path))
	{
		i = s.IndexOf(directorySeparator, 2);
		if(i == -1) i = s.Length(); else i++;
	}
	else
	{
		i = s.IndexOf(volumeSeparator);
		if(i == -1) i = 0;
		else
		{
			i++;
			if(s.Length() > i && s[i] == directorySeparator) i++;
		}
	}
	int j = s.LastIndexOf(directorySeparator, i);
	return s.Substring(i, (j == -1 ? i : j) - i);
}

void GetSubdirectories(const string &filename, strings &dirs)
{
	dirs.Clear();
	string s = GetSubdirectories(filename);
	if(!s.IsEmpty()) s.Split(directorySeparator, dirs);
}

string GetSubdirectoryName(const string &filename, int index)
{
	string s = GetSubdirectories(filename);
	if(index == 0 || s.IsEmpty()) return s;
	if(index > 0)
	{
		s.Append(directorySeparator);
		for(const TCHAR *i = s.Begin(), *j = i; i < s.End(); i++)
		{
			if(*i == directorySeparator)
			{
				index--;
				if(!index)
				{
					return s.Substring(j - s.Begin(), i - j);
				}
				else j = i + 1;
			}
		}
	}
	else
	{
		s.Insert(0, directorySeparator);
		if(!s.IsEmpty())
			for(const TCHAR *i = s.End() - 1, *j = i; i >= s.Begin(); i--)
			{
				if(*i == directorySeparator)
				{
					index++;
					if(!index)
					{
						return s.Substring(i - s.Begin() + 1, j - i);
					}
					else j = i - 1;
				}
			}
	}
	return emptyString;
}

string MakeFullFileName(const string &path, const string &filename)
{
	string p = GetCleanedPath(path);
	string f = GetCleanedPath(filename);
	AddTrailingBackslash(p);
	if(!GetDriveName(f).IsEmpty()) return f;
	strings dirsF; GetSubdirectories(f, dirsF);
	strings dirsP; GetSubdirectories(p, dirsP);
	string *i;
	for(i = dirsF.Begin(); i < dirsF.End(); i++)
		if(*i != TEXT("."))
			if(*i == TEXT(".."))
			{
				if(!dirsP.IsEmpty()) dirsP.Remove(dirsP.Length() - 1);
			}
			else dirsP.Add(*i);
	string s = GetDriveName(p);
	if(!s.IsEmpty())
	{
		if(!IsUNCPath(p)) s += volumeSeparator;
		else s.AppendFront(string(2, directorySeparator));
		s += directorySeparator;
	}
	for(i = dirsP.Begin(); i < dirsP.End(); i++)
	{
		s += *i;
		s += directorySeparator;
	}
	s += GetFileName(f);
	return s;
}

string MakeRelativeFileName(const string &path, const string &filename)
{
	string p = GetCleanedPath(path);
	string f = GetCleanedPath(filename);
	AddTrailingBackslash(p);
	if(!GetDriveName(p).Equals(GetDriveName(f), true) || IsUNCPath(p) != IsUNCPath(f)) return f;
	strings dirsF; GetSubdirectories(f, dirsF);
	strings dirsP; GetSubdirectories(p, dirsP);
	strings dirs;
	int i = 0;
	int j = 0;
	while(i < dirsF.Length() || j < dirsP.Length())
	{
		if(i < dirsF.Length() && j < dirsP.Length())
			if(!dirsF[i].Equals(dirsP[i], true))
			{
				for(; j < dirsP.Length(); j++)
					dirs.Add(TEXT(".."));
				continue;
			}
		if(j >= dirsP.Length()) dirs.Add(dirsF[i]);
		else if(i >= dirsF.Length()) dirs.Add(TEXT(".."));
		i++; j++;
	}
	string s;
	for(string * k = dirs.Begin(); k < dirs.End(); k++)
	{
		s += *k;
		s += directorySeparator;
	}
	s += GetFileName(f);
	return s;
}

string GetFullFileName(const string &filename)
{
	return MakeFullFileName(GetCurrentDirectory(), filename);
}

string GetRelativeFileName(const string &filename)
{
	return MakeRelativeFileName(GetCurrentDirectory(), filename);
}

bool FileExists(const string &path)
{
	WIN32_FIND_DATA fd;
	HANDLE handle = FindFirstFile(path, &fd);
	if(handle != INVALID_HANDLE_VALUE)
	{
		do if((fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
		{
			FindClose(handle);
			return true;
		}
		while(FindNextFile(handle, &fd));
		FindClose(handle);
	}
	return false;
}

void AddTrailingBackslash(string &path)
{
	if(!path.IsEmpty() && path.Last() != directorySeparator) path.Append(directorySeparator);
}

void RemoveTrailingBackslash(string &path)
{
	while(!path.IsEmpty() && path.Last() == directorySeparator) path.RemoveLast();
}

string FindFile(const string &path, const string &filter, bool search_subdirectories)
{
	string cur_path = path;
	AddTrailingBackslash(cur_path);
	string filename;
	WIN32_FIND_DATA fd;
	string fn = cur_path + TEXT('*');
	HANDLE h = FindFirstFile(fn, &fd);
	if(h != INVALID_HANDLE_VALUE)
	{
		do
		{
			if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
			{
				string s = fd.cFileName;
				if(FileFilterHasExtension(filter, GetFileExtension(s)))
				{
					filename = cur_path + s;
					break;
				}
			}
		} while(FindNextFile(h, &fd));
		FindClose(h);
	}
	if(filename.IsEmpty() && search_subdirectories)
	{
		h = FindFirstFile(fn, &fd);
		if(h != INVALID_HANDLE_VALUE)
		{
			do
			{
				if(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY
					&& !StringsEqual(fd.cFileName, TEXT(".")) && !StringsEqual(fd.cFileName, TEXT("..")))
				{
					filename = FindFile(cur_path + fd.cFileName, filter, true);
					if(!filename.IsEmpty()) break;
				}
			} while(FindNextFile(h, &fd));
			FindClose(h);
		}
	}
	return filename;
}

// File filter routines

string MakeFileFilterString(const string &name, const string &filter, bool filter_in_name)
{
	return name + (filter_in_name ? TEXT(" (") + filter + TEXT(")") : TEXT("")) + TEXT('|') + filter;
}

string AllFilesFilter()
{
	return TEXT("*.*");
}

string AllFilesFilterName()
{
	return TEXT("All files");
}

string GetAllFilesFilterString(bool filter_in_name)
{
	return MakeFileFilterString(AllFilesFilterName(), AllFilesFilter(), filter_in_name);
}

bool FileFilterHasExtension(const string &filter, const string &ext)
{
	int j = ext.LastIndexOf(TEXT('.'));
	if(j == -1) j = 0; else j++;
	string e = ext.Substring(j, ext.Length() - j);
	int l = filter.Length();
	int s = -1;
	for(int i = 0; i <= l; i++)
		if(s != -1 && (i == l || filter[i] == TEXT(';')))
		{
			if(filter.Substring(s, i - s) == e) return true;
			s = -1;
		}
		else if(i != l && filter[i] == TEXT('.')) s = i + 1;
	return false;
}

int ExtensionsInFileFilter(const string &filter)
{
	int l = filter.Length();
	bool s = false;
	int c = 0;
	for(int i = 0; i <= l; i++)
		if(s && (i == l || filter[i] == TEXT(';')))
		{
			c++;
			s = false;
		}
		else if(i != l && filter[i] == TEXT('.')) s = true;
	return c;
}

string GetExtensionFromFileFilter(const string &filter, int index)
{
	int l = filter.Length();
	int s = -1;
	for(int i = 0; i <= l; i++)
		if(s != -1 && (i == l || filter[i] == TEXT(';')))
		{
			if(index == 0) return filter.Substring(s, i - s);
			index--;
			s = -1;
		}
		else if(i != l && filter[i] == TEXT('.')) s = i + 1;
	return emptyString;
}

void AddFileFilterString(string &s, const string &filter_string)
{
	if(!s.IsEmpty()) s += TEXT('|');
	s += filter_string;
}

void AddFileFilterString(string &s, const string &name, const string &filter, bool filter_in_name)
{
	AddFileFilterString(s, MakeFileFilterString(name, filter, filter_in_name));
}

void AddAllFilesFilterString(string &s, bool filter_in_name)
{
	AddFileFilterString(s, GetAllFilesFilterString(filter_in_name));
}

// Convertion to string routines

const string falseString("False");
const string trueString("True");

string DateToString()
{
	TCHAR buf[9];
	return string(_tstrdate(buf));
}

string TimeToString()
{
	TCHAR buf[9];
	return string(_tstrtime(buf));
}

string IntToString(int value, int base)
{
	TCHAR buf[33];
	return string(_itot(value, buf, base));
}

string HexToString(int value)
{
	return IntToString(value, 16);
}

string OctToString(int value)
{
	return IntToString(value, 8);
}

string BinToString(int value)
{
	return IntToString(value, 2);
}

string IntToString(int value, const string &one, const string &many, const string &no)
{
	return (value == 0 && !no.IsEmpty() ? no : IntToString(value)) + TEXT(" ")
		+ (value % 10 == 1 && value % 100 != 11 ? one
		: many.IsEmpty() ? (one.IsEmpty() ? emptyString : ToLower(one.Last()) == TEXT('s') ? one + TEXT("es") : one + TEXT("s"))
			: many);
}

string BoolToString(bool value, const string &trueValue, const string &falseValue)
{
	return value ? trueValue : falseValue;
}

string FloatToString(double value)
{
	TCHAR buf[100];
	_stprintf(buf, TEXT("%f"), value);
	size_t l = _tcslen(buf);
	// Remove trailing zeros
	for(int i = l - 1; i >= 0; i--)
		if(buf[i] != TEXT('0'))
		{
			l = i + 1;
			buf[l] = TEXT('\0');
			break;
		}
	// Remove trailing period
	if(l > 0 && buf[l - 1] == TEXT('.'))
	{
		l--;
		buf[l] = TEXT('\0');
	}
	string s(buf);
	// Add zero in the beginning
	if(s.Length() > 0 && s[0] == TEXT('.')) s = TEXT('0') + s;
	return s;
}

string FloatToString(double value, int digits)
{
	const int maxDigits = 10;
	if(digits < 0) digits = 0;
	if(digits > maxDigits) digits = maxDigits;
	int p = 1;
	for(int i = 0; i < digits; i++) p *= 10;
	return FloatToString(Round(value * p) / (double)p);
}

string FloatToStringS(double value, int digits)
{
	const int maxDigits = 10;
	if(digits < 1) digits = 1;
	if(digits > maxDigits) digits = maxDigits;
	int v = 1;
	for(int i = 1; i < digits; i++) v *= 10;
	int v1 = v * 10;
	int pC = 0;
	double p = 1.0;
	while(value < v && pC < maxDigits)
	{
		value *= 10;
		p /= 10;
		pC++;
	}
	while(value >= v1 && pC > -maxDigits)
	{
		value /= 10;
		p *= 10;
		pC--;
	}
	return FloatToString(Round(value) * p);
}

string MateValuesToString(const vector<MateValue> &mateValues)
{
	string s;
	for(const MateValue * i = mateValues.Begin(); i < mateValues.End(); i++)
	{
		if(i != mateValues.Begin()) s += TEXT(' ');
		s += i->ToString();
	}
	return s;
}

void MinimizeMateValues(const vector<MateValue> &mateValues, vector<MateValue> &dstMateValues, int maxMates)
{
	dstMateValues.Clear();
	if(mateValues.IsEmpty()) throw exception("Can not minimize empty make values");
	if(maxMates < 1) maxMates = 1;
	const MateValue * first = NULL;
	bool add = false;
	for(const MateValue * i = mateValues.Begin(); i < mateValues.End(); i++)
	{
		bool last = i == mateValues.End() - 1;
		if(dstMateValues.IsEmpty() && (last || i->Value() != 0))
	 	{
			if(i != mateValues.Begin()) first = i - 1;
		}
		if(!dstMateValues.IsEmpty() || last || i->Value() != 0)
		{
			dstMateValues.Add(*i);
			if(--maxMates == 0)
			{
				if(!last)
				{
					const MateValue * next = i + 1;
					add = next->MaxValue() != -1
						&& next->Value() >= (next->MaxValue() / 2) + (next->MaxValue() % 2 == 0 ? 0 : 1);
				}
				break;
			}
		}
	}
	for(MateValue * cur = dstMateValues.End() - 1; add && cur >= dstMateValues.Begin(); cur--)
	{
		if(cur->MaxValue() != -1 && cur->Value() + 1 == cur->MaxValue())
		{
			cur->SetValue(0);
			add = true;
		}
		else
		{
			cur->SetValue(cur->Value() + 1);
			add = false;
		}
	}
	if(add)
	{
		string name = first ? first->Name()
			: TEXT('x') + IntToString(dstMateValues.First().MaxValue()) + TEXT(' ') + dstMateValues.First().Name();
		int maxValue = first ? first->MaxValue() : -1;
		dstMateValues.AddFront(MateValue(name, maxValue, 1));
		if(maxMates == 0) dstMateValues.RemoveLast();
	}
}

string MateValuesToString(const vector<MateValue> &mateValues, int maxMates)
{
	vector<MateValue> theMateValues;
	MinimizeMateValues(mateValues, theMateValues, maxMates);
	return MateValuesToString(theMateValues);
}

void SetMateValues(DWORD value, vector<MateValue> &mateValues)
{
	if(mateValues.IsEmpty()) return;
	for(MateValue * i = mateValues.End() - 1; i >= mateValues.Begin(); i--)
	{
		if(i->MaxValue() == -1)
		{
			i->SetValue((int)value);
			value = 0;
		}
		else
		{
			i->SetValue((int)(value % i->MaxValue()));
			value /= i->MaxValue();
		}
	}
	if(value != 0) throw exception("Overflow in mate values");
}

void SetMateValues(unsigned __int64 value, vector<MateValue> &mateValues)
{
	if(mateValues.IsEmpty()) return;
	for(MateValue * i = mateValues.End() - 1; i >= mateValues.Begin(); i--)
	{
		if(i->MaxValue() == -1)
		{
			i->SetValue((int)value);
			value = 0;
		}
		else
		{
			i->SetValue((int)(value % i->MaxValue()));
			value /= i->MaxValue();
		}
	}
	if(value != 0) throw exception("Overflow in mate values");
}

string MatedValueToString(DWORD value, vector<MateValue> &mateValues, int maxMates)
{
	SetMateValues(value, mateValues);
	return MateValuesToString(mateValues, maxMates);
}

string MatedValueToString(unsigned __int64 value, vector<MateValue> &mateValues, int maxMates)
{
	SetMateValues(value, mateValues);
	return MateValuesToString(mateValues, maxMates);
}

string TimeToString(DWORD ms, int maxMates)
{
	static bool initialized = false;
	static vector<MateValue> mateValues;
	if(!initialized)
	{
		mateValues.Add(MateValue(TEXT("d"), -1));
		mateValues.Add(MateValue(TEXT("h"), 24));
		mateValues.Add(MateValue(TEXT("m"), 60));
		mateValues.Add(MateValue(TEXT("s"), 60));
		mateValues.Add(MateValue(TEXT("ms"), 1000));
		initialized = true;
	}
	return MatedValueToString(ms, mateValues, maxMates);
}

string TimeToString(double sec, int maxMates)
{
	static bool initialized = false;
	static vector<MateValue> mateValues;
	if(!initialized)
	{
		mateValues.Add(MateValue(TEXT("d"), -1));
		mateValues.Add(MateValue(TEXT("h"), 24));
		mateValues.Add(MateValue(TEXT("m"), 60));
		mateValues.Add(MateValue(TEXT("s"), 60));
		mateValues.Add(MateValue(TEXT("ms"), 1000));
		mateValues.Add(MateValue(TEXT("mks"), 1000));
		initialized = true;
	}
	bool negative = sec < 0.0;
	if(negative) sec *= -1;
	sec *= 1000000;
	return (negative ? STRING("-") : emptyString) + MatedValueToString((unsigned __int64)(sec + 0.5), mateValues, maxMates);
}

string SizeToString(DWORD bytes, int maxMates)
{
	static bool initialized = false;
	static vector<MateValue> mateValues;
	if(!initialized)
	{
		mateValues.Add(MateValue(TEXT("GB"), -1));
		mateValues.Add(MateValue(TEXT("MB"), 1024));
		mateValues.Add(MateValue(TEXT("KB"), 1024));
		mateValues.Add(MateValue(TEXT("byte"), emptyString, 1024));
		initialized = true;
	}
	return MatedValueToString(bytes, mateValues, maxMates);
}

// Convertion from string routines

bool StringToBool(const string &value, const string &trueValue, const string &falseValue)
{
	string s = value;
	s.Trim();
	if(s.Equals(trueValue, true)) return true;
	else if(s.Equals(falseValue, true)) return false;
	else throw exception("Can not convert string to bool");
}

int StringToInt(const string &value, const string &end, int base)
{
	TCHAR *e;
	errno = 0;
	int v = (int)_tcstol(value, &e, base);
	if(errno == ERANGE) throw exception("Overflow when converting string to int");
	string s(e);
	s.Trim();
	string se = end;
	se.Trim();
	string vv(value);
	vv.Trim();
	if(vv.IsEmpty() || (!s.IsEmpty() && s != se)) throw exception("Can not convert string to int");
	return v;
}

int StringToHex(const string &value, const string &end)
{
	return StringToInt(value, end, 16);
}

int StringToOct(const string &value, const string &end)
{
	return StringToInt(value, end, 8);
}

int StringToBin(const string &value, const string &end)
{
	return StringToInt(value, end, 2);
}

double StringToFloat(const string &value, const string &end)
{
	TCHAR *e;
	double v = _tcstod(value, &e);
	if(v == HUGE_VAL || v == -HUGE_VAL) throw exception("Overflow when converting string to float");
	string s(e);
	s.Trim();
	string se = end;
	se.Trim();
	string vv(value);
	vv.Trim();
	if(vv.IsEmpty() || (!s.IsEmpty() && s != se)) throw exception("Can not convert string to float");
	return v;
}
