#pragma once

#include "Controls.h"
#include "VFImage.h"
#include "VFFeatures.h"

class CVFView: public CControl
{
private:
	CVFImage * image;
	COLORREF minColor;
	COLORREF maxColor;
	int resolution;
	CVFFeatures * features;
	bool showFeatures;
	COLORREF minutiaColor;
	bool showSingularPoints;
	COLORREF singularPointColor;
	bool showBlockedOrientations;
	COLORREF blockedOrientationsColor;
	int zoom;

	HBITMAP bitmap;
	int bitmapWidth, bitmapHeight;

	void FreeBitmap();

	void GetImageRect(int &x, int &y, int &w, int &h);

	void PaintBO(HDC dc, int x, int y, BYTE dir);
	void PaintDir(HDC dc, int x, int y, BYTE dir, int length);

	int FeatToLocal(int value)
	{
		return value * VF_IMAGE_RESOLUTION * zoom / (VF_FEATURES_RESOLUTION * 100);
	}

	double FeatToLocalF(double value)
	{
		return value * VF_IMAGE_RESOLUTION * zoom / (VF_FEATURES_RESOLUTION * 100.0);
	}

protected:
	LRESULT OnGetDlgCode(WPARAM /*wParam*/, MSG * /*message*/) { return DLGC_STATIC; }

	LRESULT OnEraseBackground(HDC dc, LPARAM lParam);
	LRESULT OnPaint(WPARAM wParam, LPARAM lParam);

public:
	CVFView();
	~CVFView();

	CVFImage * get_Image() { return image; }
	void set_Image(CVFImage * value)
	{
		image = value;
		UpdateImage();
	}
	__declspec(property(get = get_Image, put = set_Image)) CVFImage * Image;
	
	COLORREF get_MinColor() { return minColor; }
	void set_MinColor(COLORREF value)
	{
		if(minColor != value)
		{
			minColor = value;
			UpdateImage();
		}
	}
	__declspec(property(get = get_MinColor, put = set_MinColor)) COLORREF MinColor;

	COLORREF get_MaxColor() { return maxColor; }
	void set_MaxColor(COLORREF value)
	{
		if(maxColor != value)
		{
			maxColor = value;
			UpdateImage();
		}
	}
	__declspec(property(get = get_MaxColor, put = set_MaxColor)) COLORREF MaxColor;

	int get_Resolution() { return resolution; }
	void set_Resolution(int value)
	{
		if(value < 1) value = 1;
		if(resolution != value)
		{
			resolution = value;
			Invalidate();
		}
	}
	__declspec(property(get = get_Resolution, put = set_Resolution)) int Resolution;

	CVFFeatures * get_Features() { return features; }
	void set_Features(CVFFeatures * value)
	{
		features = value;
		UpdateFeatures();
	}
	__declspec(property(get = get_Features, put = set_Features)) CVFFeatures * Features;
	
	bool get_ShowFeatures() { return showFeatures; }
	void set_ShowFeatures(bool value)
	{
		if(showFeatures != value)
		{
			showFeatures = value;
			Invalidate();
		}
	}
	__declspec(property(get = get_ShowFeatures, put = set_ShowFeatures)) bool ShowFeatures;

	COLORREF get_MinutiaColor() { return minutiaColor; }
	void set_MinutiaColor(COLORREF value)
	{
		if(minutiaColor != value)
		{
			minutiaColor = value;
			Invalidate();
		}
	}
	__declspec(property(get = get_MinutiaColor, put = set_MinutiaColor)) COLORREF MinutiaColor;
	
	bool get_ShowSingularPoints() { return showSingularPoints; }
	void set_ShowSingularPoints(bool value)
	{
		if(showSingularPoints != value)
		{
			showSingularPoints = value;
			Invalidate();
		}
	}
	__declspec(property(get = get_ShowSingularPoints, put = set_ShowSingularPoints)) bool ShowSingularPoints;
	
	COLORREF get_SingularPointColor() { return singularPointColor; }
	void set_SingularPointColor(COLORREF value)
	{
		if(singularPointColor != value)
		{
			singularPointColor = value;
			Invalidate();
		}
	}
	__declspec(property(get = get_SingularPointColor, put = set_SingularPointColor)) COLORREF SingularPointColor;
	
	bool get_ShowBlockedOrientations() { return showBlockedOrientations; }
	void set_ShowBlockedOrientations(bool value)
	{
		if(showBlockedOrientations != value)
		{
			showBlockedOrientations = value;
			Invalidate();
		}
	}
	__declspec(property(get = get_ShowBlockedOrientations, put = set_ShowBlockedOrientations)) bool ShowBlockedOrientations;
	
	COLORREF get_BlockedOrientationsColor() { return blockedOrientationsColor; }
	void set_BlockedOrientationsColor(COLORREF value)
	{
		if(blockedOrientationsColor != value)
		{
			blockedOrientationsColor = value;
			Invalidate();
		}
	}
	__declspec(property(get = get_BlockedOrientationsColor, put = set_BlockedOrientationsColor)) COLORREF BlockedOrientationsColor;
	
	int get_Zoom() { return zoom; }
	void set_Zoom(int value)
	{
		if(value < MinZoom) value = MinZoom;
		else if(value > MaxZoom) value = MaxZoom;
		if(zoom != value)
		{
			zoom = value;
			Invalidate();
		}
	}
	__declspec(property(get = get_Zoom, put = set_Zoom)) int Zoom;

	void UpdateImage();
	void UpdateFeatures();

	void ZoomIn();
	void ZoomOut();

	static const int MinZoom;
	static const int MaxZoom;
};
