#include "System.h"
#include "VFImageFormatTIFF.h"

CVFImageFormatTIFF image_format_tiff;

bool CVFImageFormatTIFF::CanLoad()
{
	return true;
}

bool CVFImageFormatTIFF::CanSave()
{
	return false;
}

string CVFImageFormatTIFF::FileFilter()
{
	return TEXT("*.tif");
}

string CVFImageFormatTIFF::FileFilterName()
{
	return TEXT("TIFF files");
}

// TIFF tags
#define TIFF_TAG_NewSubfileType				254
#define TIFF_TAG_SubfileType				255
#define TIFF_TAG_ImageWidth					256
#define TIFF_TAG_ImageLength				257
#define TIFF_TAG_BitsPerSample				258
#define TIFF_TAG_Compression				259
#define TIFF_TAG_VALUE_Uncompressed				1
#define TIFF_TAG_VALUE_CCITT_1D					2
#define TIFF_TAG_VALUE_Group_3_FAX				3
#define TIFF_TAG_VALUE_Group_4_FAX				4
#define TIFF_TAG_VALUE_LZW						5
#define TIFF_TAG_VALUE_JPEG						6
#define TIFF_TAG_VALUE_PackBits					32773
#define TIFF_TAG_PhotometricInterpretation	262
#define TIFF_TAG_VALUE_WhiteIsZero				0
#define TIFF_TAG_VALUE_BlackIsZero				1
#define TIFF_TAG_VALUE_RGB						2
#define TIFF_TAG_VALUE_RGB_Palette				3
#define TIFF_TAG_VALUE_Transparency_mask		4
#define TIFF_TAG_VALUE_CMYK						5
#define TIFF_TAG_VALUE_YCbCr					6
#define TIFF_TAG_VALUE_CIELab					8
#define TIFF_TAG_Threshholding				263
#define TIFF_TAG_CellWidth					264
#define TIFF_TAG_CellLength					265
#define TIFF_TAG_FillOrder					266
#define TIFF_TAG_DocumentName				269
#define TIFF_TAG_ImageDescription			270
#define TIFF_TAG_Make						271
#define TIFF_TAG_Model						272
#define TIFF_TAG_StripOffsets				273
#define TIFF_TAG_Orientation				274
#define TIFF_TAG_SamplesPerPixel			277
#define TIFF_TAG_RowsPerStrip				278
#define TIFF_TAG_StripByteCounts			279
#define TIFF_TAG_MinSampleValue				280
#define TIFF_TAG_MaxSampleValue				281
#define TIFF_TAG_XResolution				282
#define TIFF_TAG_YResolution				283
#define TIFF_TAG_PlanarConfiguration		284
#define TIFF_TAG_PageName					285
#define TIFF_TAG_XPosition					286
#define TIFF_TAG_YPosition					287
#define TIFF_TAG_FreeOffsets				288
#define TIFF_TAG_FreeByteCounts				289
#define TIFF_TAG_GrayResponseUnit			290
#define TIFF_TAG_GrayResponseCurve			291
#define TIFF_TAG_T4Options					292
#define TIFF_TAG_T6Options					293
#define TIFF_TAG_ResolutionUnit				296
#define TIFF_TAG_PageNumber					297
#define TIFF_TAG_TransferFunction			301
#define TIFF_TAG_Software					305
#define TIFF_TAG_DateTime					306
#define TIFF_TAG_Artist						315
#define	TIFF_TAG_HostComputer				316
#define TIFF_TAG_Predictor					317
#define TIFF_TAG_WhitePoint					318
#define TIFF_TAG_PrimaryChromaticities		319
#define TIFF_TAG_ColorMap					320
#define TIFF_TAG_HalftoneHints				321
#define TIFF_TAG_TileWidth					322
#define TIFF_TAG_TileLength					323
#define TIFF_TAG_TileOffsets				324
#define TIFF_TAG_TileByteCounts				325
#define TIFF_TAG_InkSet						332
#define TIFF_TAG_InkNames					333
#define TIFF_TAG_NumberOfInks				334
#define TIFF_TAG_DotRange					336
#define TIFF_TAG_TargetPrinter				337
#define TIFF_TAG_ExtraSamples				338
#define TIFF_TAG_SampleFormat				339
#define TIFF_TAG_SMinSampleValue			340
#define TIFF_TAG_SMaxSampleValue			341	
#define TIFF_TAG_TransferRange				342
#define TIFF_TAG_JPEGProc					512
#define TIFF_TAG_JPEGInterchangeFormat		513
#define TIFF_TAG_JPEGInterchangeFormatLngth	514
#define TIFF_TAG_JPEGRestartInterval		515
#define TIFF_TAG_JPEGLosslessPredictors		517
#define TIFF_TAG_JPEGPointTransforms		518
#define TIFF_TAG_JPEGQTables				519
#define TIFF_TAG_JPEGDCTables				520
#define TIFF_TAG_JPEGACTables				521
#define TIFF_TAG_YCbCrCoefficients			529
#define TIFF_TAG_YCbCrSubSampling			530
#define TIFF_TAG_YCbCrPositioning			531
#define TIFF_TAG_ReferenceBlackWhite		532
#define TIFF_TAG_Copyright					33432

typedef struct TIFF_TAG
{
	USHORT  Tag;
	USHORT  Type;
	ULONG  Count;
	union
	{
		BYTE Byte[4];
		USHORT Short[2];
		ULONG Long;
	} Value;		
} Tag;

// TIFF types
#define TIFF_TYPE_Byte   1
#define TIFF_TYPE_Short  3
#define TIFF_TYPE_Long   4
#define TIFF_TYPE_SByte  6
#define TIFF_TYPE_SShort 8
#define TIFF_TYPE_SLong  9

// TIFF LWZ
#define TIFF_LZW_EoiCode	   257
#define TIFF_LZW_ClearCode	   256
#define TIFF_LZW_MaxTableEntry 8193

typedef struct TIFF_LZW_TableEntry
{
	USHORT Length;
	BYTE  Value;		
	USHORT OldCode;
} TIFF_LZW_TableEntry;

USHORT TIFFGetValueByOrder(USHORT value, int byteorder)
{
	return byteorder ? value : (USHORT)((value << 8) + (value >> 8));
}

ULONG TIFFGetValueByOrder(ULONG value, int byteorder)
{
	return byteorder ? value
		: ((value << 24)
		+ ((value & 0xFF00) << 8)
		+ ((value & 0xFF0000) >> 8)
		+ ((value & 0xFF000000) >> 24));
}

Tag* TIFFGetTag(USHORT TagType, USHORT TagsCount, Tag *Tags, int byteorder)
{
	USHORT i;
	for (i = 0; i < TagsCount; i++)
		if(TagType == TIFFGetValueByOrder(Tags[i].Tag, byteorder)) return (&Tags[i]);
	
	// default values	
	return (NULL);
}

ULONG TIFFGetTagValue(USHORT TagType, USHORT TagsCount, Tag *Tags, int byteorder)
{
	Tag* t = TIFFGetTag(TagType, TagsCount, Tags, byteorder);

	if (t != NULL)
	{
		if (TIFFGetValueByOrder(t->Count, byteorder) == 1)
		{
			switch(TIFFGetValueByOrder(t->Type, byteorder))
			{
				case TIFF_TYPE_Byte:
				case TIFF_TYPE_SByte:
					return ((ULONG)(t->Value.Byte[0]));

				case TIFF_TYPE_Short:
				case TIFF_TYPE_SShort:
					return ((ULONG)TIFFGetValueByOrder(t->Value.Short[0], byteorder));

				case TIFF_TYPE_Long:
				case TIFF_TYPE_SLong:
					return ((ULONG)TIFFGetValueByOrder(t->Value.Long, byteorder));			
			}
		}
	}

	// default values

	switch(TagType)
	{
	case TIFF_TAG_BitsPerSample:
		return (1);
	case TIFF_TAG_Compression:
		return (1);
	case TIFF_TAG_FillOrder:
		return (1);
	case TIFF_TAG_MaxSampleValue:
		return (0); //Default is 2**(BitsPerSample) - 1.
	case TIFF_TAG_MinSampleValue:
		return (0);
	case TIFF_TAG_NewSubfileType:
		return (0);
	case TIFF_TAG_Orientation:
		return (1);
	case TIFF_TAG_PlanarConfiguration:
		return (1);
	case TIFF_TAG_ResolutionUnit:
		return (2);
	case TIFF_TAG_SamplesPerPixel:
		return (1);
	case TIFF_TAG_Threshholding:
		return (1);
	}

	return 0;
}

ULONG* TIFFGetTagValues(USHORT TagType, FILE *file, USHORT TagsCount, Tag *Tags, int byteorder)
{
	Tag* t = TIFFGetTag(TagType, TagsCount, Tags, byteorder);
	ULONG* data = NULL;

	__try	
	{
		if (t != NULL)
		{
			if (TIFFGetValueByOrder(t->Count, byteorder) == 0)
				return (NULL); // no data
		
			data = (ULONG*)calloc(TIFFGetValueByOrder(t->Count, byteorder), sizeof(ULONG));
			if (data == NULL)
				return (NULL); // no memory

			ULONG i;

			switch(TIFFGetValueByOrder(t->Type, byteorder))
			{
				case TIFF_TYPE_Byte:
				case TIFF_TYPE_SByte:
					if (TIFFGetValueByOrder(t->Count, byteorder) <= 4)
					{
						for (i = 0; i < TIFFGetValueByOrder(t->Count, byteorder); ++i)
							data[i] = (ULONG)(t->Value.Byte[i]);
					}
					else
					{
						if (fseek(file, TIFFGetValueByOrder(t->Value.Long, byteorder), SEEK_SET))
							break;

						BYTE temp;
						for (i = 0; i < TIFFGetValueByOrder(t->Count, byteorder); ++i)
						{
							if (fread(&temp, sizeof(BYTE), 1, file) != 1)
								break;
							data[i] = temp;
						}
					}
					return (data);

				case TIFF_TYPE_Short:
				case TIFF_TYPE_SShort:
					if (TIFFGetValueByOrder(t->Count, byteorder) <= 2)
					{
						for (i = 0; i < TIFFGetValueByOrder(t->Count, byteorder); ++i)
							data[i] = (ULONG)(TIFFGetValueByOrder(t->Value.Short[i], byteorder));
					}
					else
					{
						if (fseek(file, TIFFGetValueByOrder(t->Value.Long, byteorder), SEEK_SET))
							break;

						USHORT temp;
						for (i = 0; i < TIFFGetValueByOrder(t->Count, byteorder); ++i)
						{
							if (fread(&temp, sizeof(USHORT), 1, file) != 1)
								break;
		  					data[i] = TIFFGetValueByOrder(temp, byteorder);
						}
					}
					return (data);

				case TIFF_TYPE_Long:
				case TIFF_TYPE_SLong:
					if (TIFFGetValueByOrder(t->Count, byteorder) <= 1)
					{
						data[0] = TIFFGetValueByOrder(t->Value.Long, byteorder);
					}
					else
					{
						if (fseek(file, TIFFGetValueByOrder(t->Value.Long, byteorder), SEEK_SET))
							break;

						ULONG temp;
						for (i = 0; i < TIFFGetValueByOrder(t->Count, byteorder); ++i)
						{
							if (fread(&temp, sizeof(ULONG), 1, file) != 1)
								break;
							data[i] = TIFFGetValueByOrder(temp, byteorder);
						}
					}
					return (data);
			}
		}
	}
	__except(EXCEPTION_EXECUTE_HANDLER){		
	}	
	
	free(data);
	return (NULL); // no tag or error
}

HRESULT TIFFReadHeaderAndTags(FILE * file, CVFImage *image, USHORT &TagsCount, Tag *&Tags, int &byteorder, ULONG *&colormap, ULONG *&bitpersample)
{

	__try	
	{
		rewind(file);
	
		size_t count;

		// read header
		BYTE  temp[2];
		ULONG IDF_Address;
	
		count = fread(temp, sizeof(BYTE), 2, file);

		if((count != 2) || (temp[1] != temp[0] && (temp[0] !='I' || temp[0] !='M')))
			return (E_FAIL); // bad header

		if (temp[0] == 'I')
			byteorder = 1;
		else
			byteorder = 0;

		if (fseek(file, 2L, SEEK_CUR)) // skip version
			return (E_FAIL);

		// read first IFD  address
		if (fread(&IDF_Address, sizeof(ULONG), 1, file) != 1)
			return (E_FAIL);
	
		// read tag count	
		if (fseek(file, TIFFGetValueByOrder(IDF_Address, byteorder), SEEK_SET))
			return (E_FAIL);

		if (fread(&TagsCount, sizeof(USHORT), 1, file) != 1)
			return (E_FAIL);

		TagsCount = TIFFGetValueByOrder(TagsCount, byteorder);	
		if (TagsCount <= 0)
			return (E_FAIL); // must be positive

		// read tags
		Tags = (Tag*)calloc(TagsCount, sizeof(Tag));	
		if (Tags == NULL)
			return (E_FAIL); // no memory

		USHORT i;
		for(i = 0; i < TagsCount; i++)
			if (fread(&Tags[i], sizeof(Tag), 1, file) != 1)
			{
				free(Tags);
				Tags = NULL;
				return (E_FAIL);
			}

		colormap = TIFFGetTagValues(TIFF_TAG_ColorMap, file, TagsCount, Tags, byteorder);	
		bitpersample = TIFFGetTagValues(TIFF_TAG_BitsPerSample, file, TagsCount, Tags, byteorder);	
		
		// get image size
		ULONG x = TIFFGetTagValue(TIFF_TAG_ImageWidth, TagsCount, Tags, byteorder);
		ULONG y = TIFFGetTagValue(TIFF_TAG_ImageLength, TagsCount, Tags, byteorder);
	
		if (x <= 0 || y <= 0)
		{
			free(Tags);
			Tags = NULL;
			free(colormap);
			colormap = NULL;
			free(bitpersample);
			bitpersample = NULL;
			return (E_FAIL);
		}

		image->SetSize(x, y);		
	}
	__except(EXCEPTION_EXECUTE_HANDLER){		
		free(Tags);
		Tags = NULL;
		free(colormap);
		colormap = NULL;
		free(bitpersample);
		bitpersample = NULL;
		return (E_FAIL);
	}	

	return S_OK;
}

HRESULT TIFFReadStipsInfo(FILE * file, USHORT TagsCount, Tag *Tags, int byteorder, ULONG &count, ULONG *&offsets, ULONG *&counts)
{
	//if (TIFFGetTagValue(TIFF_TAG_RowsPerStrip, TagsCount, Tags, byteorder) == 0)
	//	return (E_FAIL);

	offsets = TIFFGetTagValues(TIFF_TAG_StripOffsets, file, TagsCount, Tags, byteorder);	
	if (offsets == NULL)
		return (E_FAIL);
	
	counts = TIFFGetTagValues(TIFF_TAG_StripByteCounts, file, TagsCount, Tags, byteorder);
	if (counts == NULL)
	{
		free(offsets);
		offsets = NULL;
		return (E_FAIL);
	}

	count = TIFFGetValueByOrder(TIFFGetTag(TIFF_TAG_StripByteCounts, TagsCount, Tags, byteorder)->Count, byteorder);
	if (count <= 0)
	{
		free(counts);
		offsets = NULL;

		free(counts);
		offsets = NULL;

		return (E_FAIL);
	}

	return (S_OK);
}

HRESULT TIFFReadTilesInfo(FILE * file, USHORT TagsCount, Tag *Tags, int byteorder, ULONG &count, ULONG *&offsets, ULONG *&counts)
{
	offsets = TIFFGetTagValues(TIFF_TAG_TileOffsets, file, TagsCount, Tags, byteorder);	
	if (offsets == NULL)
		return (E_FAIL);
	
	counts = TIFFGetTagValues(TIFF_TAG_TileByteCounts, file, TagsCount, Tags, byteorder);
	if (counts == NULL)
	{
		free(offsets);
		offsets = NULL;
		return (E_FAIL);
	}

	count = TIFFGetValueByOrder(TIFFGetTag(TIFF_TAG_TileOffsets, TagsCount, Tags, byteorder)->Count, byteorder);
	if (count <= 0)
	{
		free(counts);
		offsets = NULL;

		free(counts);
		offsets = NULL;

		return (E_FAIL);
	}

	return (S_OK);
}

HRESULT ReadDataFromFile(FILE * file, BYTE* data, ULONG count, ULONG offset)
{
	if (fseek(file, offset, SEEK_SET))
		return (E_FAIL);

	if (fread(data, sizeof(BYTE), count, file) != count)
		return (E_FAIL);	

	return (S_OK);
}

BYTE GetBit(BYTE value, BYTE bit)
{
	return (BYTE)((value >> (7-bit)) & 0x01);
}

BYTE Get4Bit(BYTE value, BYTE bit)
{
	if (bit)
		return (BYTE)(value & 0x0F);
	else
		return (BYTE)((value >> (4)) & 0x0F);
}

void FillOrderTable(BYTE* table)
{
	BYTE i;
	for (i = 0; i < 0xFF; ++i)
		table[i] = (BYTE)(
			((i >> 7) & 0x01) +
			((i >> 6) & 0x01 << 1) +
			((i >> 5) & 0x01 << 2) +
			((i >> 4) & 0x01 << 3) +
			((i >> 3) & 0x01 << 4) +
			((i >> 2) & 0x01 << 5) +
			((i >> 1) & 0x01 << 6) +
			((i)      & 0x01 << 7));
}

void ChangeFillOrder(BYTE* table, BYTE* data, ULONG l)
{
	ULONG i;
	for (i = 0; i < l; ++i)
		data[i] = table[data[i]];
}

USHORT TIFFLZWGetNextCode(BYTE* d_in, ULONG* stream_byte, ULONG* stream_bit, ULONG* stream_codebits)
{	
	USHORT code = 0;
	ULONG i;
	
	for (i = 0; i < *stream_codebits; ++i)
	{
		code <<= 1;
		
		if (d_in[*stream_byte] & 0x80)
			code += 1;

		d_in[*stream_byte] <<= 1;

		++(*stream_bit);
		if (*stream_bit > 7)
		{
			++(*stream_byte);
			*stream_bit = 0;
		}
	}

	return (code);
}

void TIFFLZWInitializeTable(TIFF_LZW_TableEntry* lzw_table)
{
	USHORT i;

	for (i = 256; i < TIFF_LZW_MaxTableEntry; ++i)
	{
		lzw_table[i].Length = 0;
		lzw_table[i].OldCode = TIFF_LZW_EoiCode;		
	}	
}

void TIFFLZWWriteCode(BYTE* d_out, ULONG* writed, TIFF_LZW_TableEntry* lzw_table, USHORT Code, BYTE *firstchar, ULONG /*length*/)
{	
	ULONG i, l;
	l = lzw_table[Code].Length;

	for (i = 0; i < l; ++i)
	{
		d_out[(*writed)+l-i-1] = lzw_table[Code].Value;

		if (lzw_table[Code].OldCode != TIFF_LZW_EoiCode)
			Code = lzw_table[Code].OldCode;
		else
			break;	
	}	
	
	*firstchar = lzw_table[Code].Value;	
	(*writed) += l;	
}	

int TIFFLZWIsInTable(TIFF_LZW_TableEntry* lzw_table, USHORT Code)
{
	return (lzw_table[Code].Length>0);
}

void TIFFLZWAddStringToTable(TIFF_LZW_TableEntry* lzw_table, USHORT Code1, BYTE first_char, ULONG* stream_codebits, USHORT entry)
{	
	USHORT i;
	// search for free space
	for (i = entry; i < TIFF_LZW_MaxTableEntry; ++i)
	{
		if (lzw_table[i].Length == 0)
			break;
	}	
	
	lzw_table[i].Length = (WORD)(lzw_table[Code1].Length + 1);
	lzw_table[i].OldCode = Code1;

	lzw_table[i].Value = first_char;

	*stream_codebits = 9;
	if (i >= 511-1)
		*stream_codebits = 10;
	if (i >= 1023-1)
		*stream_codebits = 11;
	if (i >= 2047-1)
		*stream_codebits = 12;
}

HRESULT TIFFDecompress(BYTE* d_in, BYTE* d_out, ULONG l, ULONG BytesPerRow, ULONG compression)
{
	// common variables
	USHORT i;
	ULONG writed;

	// TIFF_TAG_VALUE_PackBits	
	ULONG readed;
	ULONG r;
	signed char byte;	

	// TIFF_TAG_VALUE_LZW
	USHORT Code;
	USHORT OldCode;
	ULONG stream_byte;
	ULONG stream_bit;
	ULONG stream_codebits;
	BYTE first_char;
	TIFF_LZW_TableEntry lzw_table[TIFF_LZW_MaxTableEntry];
	USHORT entry_count;

	for (i = 0; i < TIFF_LZW_MaxTableEntry; ++i)
	{
		lzw_table[i].Length = 0;
		lzw_table[i].OldCode = TIFF_LZW_EoiCode;
	}

	__try
	{

	switch (compression)
	{
	case TIFF_TAG_VALUE_Uncompressed:
		::CopyMemory(d_out, d_in, l);
		break;

	case TIFF_TAG_VALUE_PackBits:

		writed = 0;
		readed = 0;
		r	   = 0;
	
		while (writed < l)
		{
			writed = r*BytesPerRow;

			while (writed < (r+1)*BytesPerRow)
			{
				byte = d_in[readed]; ++readed;

				if (byte >= 0 && byte <= 127)
				{
					for (i = 0; i < byte+1; ++i)
					{
						d_out[writed] = d_in[readed];
						++readed; ++writed;
						if (writed >= (r+1)*BytesPerRow)
							break;
					}
				}
			
				if (writed >= (r+1)*BytesPerRow)
					break;
			
				if (byte >= -127 && byte <= -1)	
				{
					for (i = 0; i < -byte+1; ++i)
					{
						d_out[writed] = d_in[readed];
						++writed;
						if (writed >= (r+1)*BytesPerRow)
							break;
					}				
					++readed;
				}			
			}
			++r;
		}

		break;

	case TIFF_TAG_VALUE_LZW:

		stream_codebits = 9; // starting from 9 bit codes
		Code			= 0;
		OldCode			= 0;
		stream_byte		= 0;
		stream_bit		= 0;		
		entry_count		= 258;

		for (i = 0; i < 256; ++i)
		{
			lzw_table[i].Length = 1;
			lzw_table[i].Value  = (BYTE) i;
		}

		writed = 0;

		while (writed < l && (Code = TIFFLZWGetNextCode(d_in, &stream_byte, &stream_bit, &stream_codebits))
			!= TIFF_LZW_EoiCode)
		{
			if (Code == TIFF_LZW_ClearCode || writed == 0)
			{
				TIFFLZWInitializeTable(lzw_table); entry_count = 258; stream_codebits = 9;
				Code = TIFFLZWGetNextCode(d_in, &stream_byte, &stream_bit, &stream_codebits);
				if (Code == TIFF_LZW_EoiCode)
					break;
				TIFFLZWWriteCode(d_out, &writed, lzw_table, Code, &first_char, l);
				OldCode = Code;
			} /* end of ClearCode case */
			else
			{
				if (TIFFLZWIsInTable(lzw_table, Code))
				{
					TIFFLZWWriteCode(d_out, &writed, lzw_table, Code, &first_char, l);					
					TIFFLZWAddStringToTable(lzw_table, OldCode, first_char, &stream_codebits, entry_count);
					entry_count++;
					OldCode = Code;
				}
				else
				{
					TIFFLZWWriteCode(d_out, &writed, lzw_table, OldCode, &first_char, l);
					
					d_out[writed] = first_char; ++writed;					
					
					//TIFFLZWAddStringToTable(lzw_table, OldCode, first_char, &stream_codebits, entry_count);
					//entry_count++;
					TIFFLZWAddStringToTable(lzw_table, OldCode, first_char, &stream_codebits, Code);
					entry_count = (WORD)(Code + 1);
					OldCode = Code;
				}
			} /* end of not-ClearCode case */
		} /* end of while loop */

		return (S_OK);

	default:
		return (E_FAIL);
	}

	}
	__except(EXCEPTION_EXECUTE_HANDLER){						
		return (S_OK);
	}	

	return (S_OK);
}

HRESULT TIFFReadImage(FILE *file, CVFImage *image, ULONG BitsPerSample, int rgb, USHORT TagsCount, Tag *Tags, int byteorder, ULONG *colormap, ULONG &count, ULONG *&offsets, ULONG *&counts)
{
	ULONG i, y, x, j, z, tag_photo, tag_fillorder, tag_compression, tag_predictor;
	BYTE table[0xFF]; // for FillOrder
	int error = 0;

	BYTE* data = NULL;
	BYTE* uncompressed_data = NULL;	


	__try
	{
		tag_predictor = TIFFGetTagValue(TIFF_TAG_Predictor, TagsCount, Tags, byteorder);
		tag_photo        = TIFFGetTagValue(TIFF_TAG_PhotometricInterpretation, TagsCount, Tags, byteorder);
		tag_fillorder    = TIFFGetTagValue(TIFF_TAG_FillOrder, TagsCount, Tags, byteorder);
		tag_compression  = TIFFGetTagValue(TIFF_TAG_Compression, TagsCount, Tags, byteorder);

		if (tag_fillorder == 2)
			FillOrderTable(table);

		if (FAILED(TIFFReadStipsInfo(file, TagsCount, Tags, byteorder, count, offsets, counts)))
		{
			if (FAILED(TIFFReadTilesInfo(file, TagsCount, Tags, byteorder, count, offsets, counts)))
			{
				return (E_FAIL); // no image!
			}
			else // tiles
			{
				ULONG TileWidth    = TIFFGetTagValue(TIFF_TAG_TileWidth, TagsCount, Tags, byteorder);
				ULONG TileLength   = TIFFGetTagValue(TIFF_TAG_TileLength, TagsCount, Tags, byteorder);
				ULONG BytesPerRow = 0;

				if (BitsPerSample == 1)
					BytesPerRow = (TileWidth+7)/8;

				if (BitsPerSample == 4)
					BytesPerRow = (TileWidth+1)/2;

				if (BitsPerSample == 8)
					BytesPerRow = TileWidth;

				if (BitsPerSample == 16)
					BytesPerRow = TileWidth*2;

				if (rgb)
					BytesPerRow = 3 * BytesPerRow;

				ULONG TilePerX     = image->Width()/TileWidth + ((image->Width()%TileWidth)?1:0);
				/*ULONG TilePerY     = image->Height()/TileLength + ((image->Height()%TileLength)?1:0)*/;

				// find max tile
				ULONG maxtile = counts[0];
				for (i = 1; i < count; ++i)
					if (maxtile < counts[i])
						maxtile = counts[i];

				data = (BYTE*) calloc(maxtile, sizeof(BYTE));
				uncompressed_data = (BYTE*) calloc(TileLength*BytesPerRow, sizeof(BYTE));

				if (data != NULL)
				for (i = 0; i < count; ++i)
				{				
					if (FAILED(ReadDataFromFile(file, data, counts[i], offsets[i])))
					{
						error = 1;
						break;
					}

					// fill order
					if (tag_fillorder == 2)
						ChangeFillOrder(table, data, counts[i]);

					// decompress				

					if (FAILED(TIFFDecompress(data, uncompressed_data, BytesPerRow*TileLength, BytesPerRow, tag_compression)))
					{
						error = 1;
						break;
					}
					
					// put into buffer
					j = i/TilePerX*TileLength;
					z = i%TilePerX*TileWidth;

					if (rgb)
					{
						if (BitsPerSample == 8)
						{
       						for (y = j; y < ((j+TileLength)>(ULONG)image->Height()?(ULONG)image->Height():(j+TileLength)); ++y)
  								for (x = z; x < ((z+TileWidth)>(ULONG)image->Width()?(ULONG)image->Width():(z+TileWidth)); ++x)
								{
									if (tag_predictor == 2 && x > z) // Horizontal differencing
									{
										uncompressed_data[(y-j)*BytesPerRow+(x-z)*3] = (BYTE)(uncompressed_data[(y-j)*BytesPerRow+(x-z)*3] + uncompressed_data[(y-j)*BytesPerRow+(x-z-1)*3]);
										uncompressed_data[(y-j)*BytesPerRow+(x-z)*3+1] = (BYTE)(uncompressed_data[(y-j)*BytesPerRow+(x-z)*3+1] + uncompressed_data[(y-j)*BytesPerRow+(x-z-1)*3+1]);
										uncompressed_data[(y-j)*BytesPerRow+(x-z)*3+2] = (BYTE)(uncompressed_data[(y-j)*BytesPerRow+(x-z)*3+2] + uncompressed_data[(y-j)*BytesPerRow+(x-z-1)*3+2]);
									}

									image->Bits()[y * image->Width() + x] = (BYTE)((uncompressed_data[(y-j)*BytesPerRow+(x-z)*3]+
												uncompressed_data[(y-j)*BytesPerRow+(x-z)*3+1]+
												uncompressed_data[(y-j)*BytesPerRow+(x-z)*3+2])/3);
								}
							
							continue;
						}

						if (BitsPerSample == 16)
						{
       						for (y = j; y < ((j+TileLength)>(ULONG)image->Height()?(ULONG)image->Height():(j+TileLength)); ++y)
  								for (x = z; x < ((z+TileWidth)>(ULONG)image->Width()?(ULONG)image->Width():(z+TileWidth)); ++x)
								{
									if (tag_predictor == 2 && x > z) // Horizontal differencing
									{
										*(USHORT*)&uncompressed_data[(y-j)*BytesPerRow+(x-z)*3*2] = (WORD)(*(USHORT*)&uncompressed_data[(y-j)*BytesPerRow+(x-z)*3*2] + *(USHORT*)&uncompressed_data[(y-j)*BytesPerRow+(x-z-1)*3*2]);
										*(USHORT*)&uncompressed_data[(y-j)*BytesPerRow+(x-z)*3*2+1] = (WORD)(*(USHORT*)&uncompressed_data[(y-j)*BytesPerRow+(x-z)*3*2+1] + *(USHORT*)&uncompressed_data[(y-j)*BytesPerRow+(x-z-1)*3*2+1]);
										*(USHORT*)&uncompressed_data[(y-j)*BytesPerRow+(x-z)*3*2+2] = (WORD)(*(USHORT*)&uncompressed_data[(y-j)*BytesPerRow+(x-z)*3*2+2] + *(USHORT*)&uncompressed_data[(y-j)*BytesPerRow+(x-z-1)*3*2+2]);
									}

									image->Bits()[y * image->Width() + x] = (BYTE)
												((*(USHORT*)&uncompressed_data[(y-j)*BytesPerRow+(x-z)*3*2]+
												*(USHORT*)&uncompressed_data[(y-j)*BytesPerRow+(x-z)*3*2+1]+
												*(USHORT*)&uncompressed_data[(y-j)*BytesPerRow+(x-z)*3*2+2])/3/0x100);						
								}
							continue;
						}                        
					}

					if (BitsPerSample == 1)
					{
						for (y = j; y < ((j+TileLength)>(ULONG)image->Height()?(ULONG)image->Height():(j+TileLength)); ++y)
							for (x = z; x < ((z+TileWidth)>(ULONG)image->Width()?(ULONG)image->Width():(z+TileWidth)); ++x)
								image->Bits()[y * image->Width() + x] = (BYTE)(GetBit(uncompressed_data[(y-j)*BytesPerRow+(x-z)/8], (BYTE)((x - z) % 8)) * 0xFF);
						continue;
					}

					if (colormap == NULL)
					{
						if (BitsPerSample == 4)
							for (y = j; y < ((j+TileLength)>(ULONG)image->Height()?(ULONG)image->Height():(j+TileLength)); ++y)
								for (x = z; x < ((z+TileWidth)>(ULONG)image->Width()?(ULONG)image->Width():(z+TileWidth)); ++x)
									image->Bits()[y * image->Width() + x] = (BYTE)(Get4Bit(uncompressed_data[(y-j)*BytesPerRow+(x-z)/2], (BYTE)((x - z) % 2)) * 0x0F);

						if (BitsPerSample == 8)
							for (y = j; y < ((j+TileLength)>(ULONG)image->Height()?(ULONG)image->Height():(j+TileLength)); ++y)
								for (x = z; x < ((z+TileWidth)>(ULONG)image->Width()?(ULONG)image->Width():(z+TileWidth)); ++x)
								{
										if (tag_predictor == 2 && x > z) // Horizontal differencing
											uncompressed_data[(y-j)*BytesPerRow+(x-z)] = (BYTE)(uncompressed_data[(y-j)*BytesPerRow+(x-z)] + uncompressed_data[(y-j)*BytesPerRow+(x-z-1)]);
	                                
									image->Bits()[y * image->Width() + x] = (BYTE)(uncompressed_data[(y-j)*BytesPerRow+(x-z)]);
								}
					
						if (BitsPerSample == 16)
						{
							for (y = j; y < ((j+TileLength)>(ULONG)image->Height()?(ULONG)image->Height():(j+TileLength)); ++y)
								for (x = z; x < ((z+TileWidth)>(ULONG)image->Width()?(ULONG)image->Width():(z+TileWidth)); ++x)
								{
									if (tag_predictor == 2 && x > z) // Horizontal differencing
										*(USHORT*)&uncompressed_data[(y-j)*BytesPerRow+(x-z)*2] = (WORD)(*(USHORT*)&uncompressed_data[(y-j)*BytesPerRow+(x-z)*2] + *(USHORT*)&uncompressed_data[(y-j)*BytesPerRow+(x-z-1)*2]);

									image->Bits()[y * image->Width() + x] = (BYTE)(*(USHORT*)&uncompressed_data[(y-j)*BytesPerRow+(x-z)*2]/0x100);						
								}
						}
					}
					else
					{
						ULONG temp;
						ULONG rgbsize = 1 << BitsPerSample;

						if (BitsPerSample == 4)
							for (y = j; y < ((j+TileLength)>(ULONG)image->Height()?(ULONG)image->Height():(j+TileLength)); ++y)
								for (x = z; x < ((z+TileWidth)>(ULONG)image->Width()?(ULONG)image->Width():(z+TileWidth)); ++x)
								{
									temp = Get4Bit(uncompressed_data[(y-j)*BytesPerRow+(x-z)/2], (BYTE)((x - z) % 2));									
									image->Bits()[y * image->Width() + x] = (BYTE) ((colormap[temp]+ colormap[rgbsize+temp] + colormap[2*rgbsize+temp])/3/0x100);						
								}
									

						if (BitsPerSample == 8)
							for (y = j; y < ((j+TileLength)>(ULONG)image->Height()?(ULONG)image->Height():(j+TileLength)); ++y)
								for (x = z; x < ((z+TileWidth)>(ULONG)image->Width()?(ULONG)image->Width():(z+TileWidth)); ++x)
								{
									if (tag_predictor == 2 && x > z) // Horizontal differencing
										uncompressed_data[(y-j)*BytesPerRow+(x-z)] = (BYTE)(uncompressed_data[(y-j)*BytesPerRow+(x-z)] + uncompressed_data[(y-j)*BytesPerRow+(x-z-1)]);

									temp = uncompressed_data[(y-j)*BytesPerRow+(x-z)];
									image->Bits()[y * image->Width() + x] = (BYTE) ((colormap[temp]+ colormap[rgbsize+temp] + colormap[2*rgbsize+temp])/3/0x100);						
								}
					}
				}			
			}
		}
		else // strips
		{
			ULONG RowsPerStrip = TIFFGetTagValue(TIFF_TAG_RowsPerStrip, TagsCount, Tags, byteorder);
			ULONG BytesPerRow = 0;
			
			if (BitsPerSample == 1)
				BytesPerRow = (image->Width()+7)/8;

			if (BitsPerSample == 4)
				BytesPerRow = (image->Width()+1)/2;

			if (BitsPerSample == 8)
				BytesPerRow = image->Width();

			if (BitsPerSample == 16)
				BytesPerRow = image->Width()*2;

			if (rgb)
					BytesPerRow = 3 * BytesPerRow;

			if (RowsPerStrip == 0)
				RowsPerStrip = counts[0]/BytesPerRow;

			// find max strip
			ULONG maxstrip = counts[0];
			for (i = 1; i < count; ++i)
				if (maxstrip < counts[i])
					maxstrip = counts[i];

			data = (BYTE*) calloc(maxstrip, sizeof(BYTE));
			uncompressed_data = (BYTE*) calloc(RowsPerStrip*BytesPerRow, sizeof(BYTE));

			if (data != NULL)
			for (i = 0; i < count; ++i)			
			{				
				if (FAILED(ReadDataFromFile(file, data, counts[i], offsets[i])))
				{
					error = 1;
					break;
				}

				// fill order
				if (tag_fillorder == 2)
					ChangeFillOrder(table, data, counts[i]);

				// decompress				
				// put into buffer
				if (i+1 == count)
					j = image->Height()-RowsPerStrip*i;
				else
					j = RowsPerStrip;						

				if (FAILED(TIFFDecompress(data, uncompressed_data, BytesPerRow*j, BytesPerRow, tag_compression)))
				{
					error = 1;
					break;
				}						
				
				if (rgb)
				{
					if (BitsPerSample == 8)
					{                        
						for (y = 0; y < j; ++y)
							for (x = 0; x < (ULONG) image->Width(); ++x)
							{
								if (tag_predictor == 2 && x > 0) // Horizontal differencing
								{
									uncompressed_data[y*BytesPerRow+x*3] = (BYTE)(uncompressed_data[y*BytesPerRow+x*3] + uncompressed_data[y*BytesPerRow+(x-1)*3]);
									uncompressed_data[y*BytesPerRow+x*3+1] = (BYTE)(uncompressed_data[y*BytesPerRow+x*3+1] + uncompressed_data[y*BytesPerRow+(x-1)*3+1]);
									uncompressed_data[y*BytesPerRow+x*3+2] = (BYTE)(uncompressed_data[y*BytesPerRow+x*3+2] + uncompressed_data[y*BytesPerRow+(x-1)*3+2]);
								}

								image->Bits()[(RowsPerStrip*i+y)*image->Width() + x] = (BYTE)((uncompressed_data[y*BytesPerRow+x*3]+
															uncompressed_data[y*BytesPerRow+x*3+1]+
															uncompressed_data[y*BytesPerRow+x*3+2])/3);
							}
								
						continue;
					}

					if (BitsPerSample == 16)
					{   
						for (y = 0; y < j; ++y)
							for (x = 0; x < (ULONG) image->Width(); ++x)
							{
								if (tag_predictor == 2 && x > 0) // Horizontal differencing
								{
									*(USHORT*)&uncompressed_data[y*BytesPerRow+x*3*2] = (WORD)(*(USHORT*)&uncompressed_data[y*BytesPerRow+x*3*2] + *(USHORT*)&uncompressed_data[y*BytesPerRow+(x-1)*3*2]);
									*(USHORT*)&uncompressed_data[y*BytesPerRow+x*3*2+1] = (WORD)(*(USHORT*)&uncompressed_data[y*BytesPerRow+x*3*2+1] + *(USHORT*)&uncompressed_data[y*BytesPerRow+(x-1)*3*2+1]);
									*(USHORT*)&uncompressed_data[y*BytesPerRow+x*3*2+2] = (WORD)(*(USHORT*)&uncompressed_data[y*BytesPerRow+x*3*2+2] + *(USHORT*)&uncompressed_data[y*BytesPerRow+(x-1)*3*2+2]);
								}

								image->Bits()[(RowsPerStrip*i+y)*image->Width() + x] = (BYTE)((*(USHORT*)&uncompressed_data[y*BytesPerRow+x*2*3]+
															*(USHORT*)&uncompressed_data[y*BytesPerRow+x*2*3+1]+
															*(USHORT*)&uncompressed_data[y*BytesPerRow+x*2*3+2])/3/0x100);
							}							
						continue;
					}
				}

				if (BitsPerSample == 1)
				{
					for (y = 0; y < j; ++y)
						for (x = 0; x < (ULONG)image->Width(); ++x)
							image->Bits()[(RowsPerStrip*i+y)*image->Width() + x] = (BYTE)(GetBit(uncompressed_data[y*BytesPerRow+x/8], (BYTE)(x % 8)) * 0xFF);
					continue;
				}
				
				if (colormap == NULL)
				{
					if (BitsPerSample == 4)
						for (y = 0; y < j; ++y)
							for (x = 0; x < (ULONG)image->Width(); ++x)
								image->Bits()[(RowsPerStrip*i+y)*image->Width() + x] = (BYTE)(Get4Bit(uncompressed_data[y*BytesPerRow+x/2], (BYTE)(x % 2)) * 0x0F);

					if (BitsPerSample == 8)
					{
						for (y = 0; y < j; ++y)
							for (x = 0; x < (ULONG)image->Width(); ++x)
							{
								if (tag_predictor == 2 && x > 0) // Horizontal differencing
									uncompressed_data[y*BytesPerRow+x] = (BYTE)(uncompressed_data[y*BytesPerRow+x] + uncompressed_data[y*BytesPerRow+x-1]);

								image->Bits()[(RowsPerStrip*i+y)*image->Width() + x] = (BYTE)(uncompressed_data[y*BytesPerRow+x]);
							}
					}

					if (BitsPerSample == 16)
					{                        
						for (y = 0; y < j; ++y)
							for (x = 0; x < (ULONG) image->Width(); ++x)
							{
								if (tag_predictor == 2 && x > 0) // Horizontal differencing
									*(USHORT*)&uncompressed_data[y*BytesPerRow+2*(x-1)] = (WORD)(*(USHORT*)&uncompressed_data[y*BytesPerRow+2*(x-1)] + *(USHORT*)&uncompressed_data[y*BytesPerRow+2*(x-1)]);

								image->Bits()[(RowsPerStrip*i+y)*image->Width() + x] = (BYTE)(*(USHORT*)&uncompressed_data[y*BytesPerRow+2*x]/0x100);
							}
					}
				}
				else
				{
					ULONG temp;
					ULONG rgbsize = 1 << BitsPerSample;

					if (BitsPerSample == 4)
						for (y = 0; y < j; ++y)
							for (x = 0; x < (ULONG)image->Width(); ++x)
							{

									temp = Get4Bit(uncompressed_data[y*BytesPerRow+x/2], (BYTE)(x % 2));
									image->Bits()[(RowsPerStrip*i+y)*image->Width() + x] = (BYTE) ((colormap[temp]+ colormap[rgbsize+temp] + colormap[2*rgbsize+temp])/3/0x100);						
							}				
								

					if (BitsPerSample == 8)
						for (y = 0; y < j; ++y)
							for (x = 0; x < (ULONG)image->Width(); ++x)
							{
								if (tag_predictor == 2 && x > 0) // Horizontal differencing
										uncompressed_data[y*BytesPerRow+x] = (BYTE)(uncompressed_data[y*BytesPerRow+x] + uncompressed_data[y*BytesPerRow+x-1]);

									temp = uncompressed_data[y*BytesPerRow+x];													
									image->Bits()[(RowsPerStrip*i+y)*image->Width() + x] = (BYTE) ((colormap[temp]+ colormap[rgbsize+temp] + colormap[2*rgbsize+temp])/3/0x100);						
							}												
				}
			}		
		}

		// TIFF_TAG_VALUE_WhiteIsZero
		if (tag_photo == TIFF_TAG_VALUE_WhiteIsZero)
			for (y = 0; y < (ULONG)image->Height(); ++y)
				for (x = 0; x < (ULONG)image->Width(); ++x)
					image->Bits()[y * image->Width() + x] = (BYTE)(0xFF - image->Bits()[y * image->Width() + x]);

		free(data);
		data = NULL;

		free(uncompressed_data);
		uncompressed_data = NULL;

		if (error)
			return (E_FAIL);
		else
			return (S_OK);
	}
	__except(EXCEPTION_EXECUTE_HANDLER){				
		free(data);
		free(uncompressed_data);
		return (E_FAIL);
	}	
}

HRESULT TIFFReadData(FILE * file, CVFImage *image, USHORT TagsCount, Tag *Tags, int byteorder, ULONG *colormap, ULONG *bitpersample, ULONG &count, ULONG *&offsets, ULONG *&counts)
{
	// detect type of image

	ULONG BitsPerSample   = TIFFGetTagValue(TIFF_TAG_BitsPerSample, TagsCount, Tags, byteorder);
	ULONG SamplesPerPixel = TIFFGetTagValue(TIFF_TAG_SamplesPerPixel, TagsCount, Tags, byteorder);
	ULONG PhotometricInterpretation = TIFFGetTagValue(TIFF_TAG_PhotometricInterpretation, TagsCount, Tags, byteorder);
	
	// bilevel?
	if (BitsPerSample == 1 &&
		SamplesPerPixel == 1 &&
		(PhotometricInterpretation == TIFF_TAG_VALUE_WhiteIsZero ||
		PhotometricInterpretation == TIFF_TAG_VALUE_BlackIsZero))
	{
		return TIFFReadImage(file, image, BitsPerSample, 0, TagsCount, Tags, byteorder, colormap, count, offsets, counts);
	}

	// grayscale?
	if ((BitsPerSample == 4 || BitsPerSample == 8 || BitsPerSample == 16) &&
		SamplesPerPixel == 1 &&
		(PhotometricInterpretation == TIFF_TAG_VALUE_WhiteIsZero ||
		PhotometricInterpretation == TIFF_TAG_VALUE_BlackIsZero))
	{
		return TIFFReadImage(file, image, BitsPerSample, 0, TagsCount, Tags, byteorder, colormap, count, offsets, counts);
	}

	// color map?
	if ((BitsPerSample == 4 || BitsPerSample == 8) &&
		colormap != NULL &&
		TIFFGetTagValue(TIFF_TAG_PhotometricInterpretation, TagsCount, Tags, byteorder) == TIFF_TAG_VALUE_RGB_Palette)
	{
		return TIFFReadImage(file, image, BitsPerSample, 0, TagsCount, Tags, byteorder, colormap, count, offsets, counts);
	}

	// true color? 8bit?
	if (bitpersample != NULL &&
		TIFFGetTagValue(TIFF_TAG_SamplesPerPixel, TagsCount, Tags, byteorder) == 3 &&
		TIFFGetTagValue(TIFF_TAG_PhotometricInterpretation, TagsCount, Tags, byteorder) == TIFF_TAG_VALUE_RGB)
	{
		if (bitpersample[0] == 8 && bitpersample[1] == 8 &&
			bitpersample[2] == 8)
			
			return TIFFReadImage(file, image, 8, 1, TagsCount, Tags, byteorder, colormap, count, offsets, counts);

		if (bitpersample[0] == 16 && bitpersample[1] == 16 &&
			bitpersample[2] == 16)
			
			return TIFFReadImage(file, image, 16, 1, TagsCount, Tags, byteorder, colormap, count, offsets, counts);
	}

	return E_FAIL; // image type is wrong or unknown
}

bool CVFImageFormatTIFF::Load(CVFImage *image, const string &filename)
{
	FILE *file = NULL;

	int byteorder;
	USHORT TagsCount;
	Tag *Tags;
	ULONG *colormap = NULL;
	ULONG *bitpersample = NULL;
	ULONG count;
	ULONG *offsets = NULL;
	ULONG *counts = NULL;

	__try
	{		
		file = _tfopen(filename, TEXT("rb"));
		if(file == NULL) return false;
		
		int error = 0;

		if(FAILED(TIFFReadHeaderAndTags(file, image, TagsCount, Tags, byteorder, colormap, bitpersample))) error = 1;
		else if(FAILED(TIFFReadData(file, image, TagsCount, Tags, byteorder, colormap, bitpersample, count, offsets, counts))) error = 1;
		
		if(error)
		{
			image->Clear();
			return false;
		}
	}
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
		image->Clear();
		return false;
	}
	
	free(Tags);
	free(offsets);
	free(counts);
	free(colormap);
	free(bitpersample);

	fclose(file);			

	return true;
}

bool CVFImageFormatTIFF::Save(CVFImage * /*image*/, const string & /*filename*/)
{
	return false;
}
