#include "System.h"
#include "WaitForm.h"

///////////////
//           //
// CWaitForm //
//           //
///////////////

CWaitForm::CWaitForm()
	: CForm(), thread(NULL)
{
	SetMinimizeBox(false);
	SetMaximizeBox(false);
	SetSystemMenu(false);
	SetToolWindow(true);
	SetFormPosition(fpOwnerCenter);
	SetText(STRING("Please wait"));
}

LRESULT CWaitForm::OnMessage(UINT message, WPARAM wParam, LPARAM lParam)
{
	if(message == CM_THREAD_FINISHED)
	{
		thread = NULL;
		Close();
		return 0;
	}
	else return CForm::OnMessage(message, wParam, lParam);
}

void CWaitForm::OnFirstShow()
{
	if(thread) thread->Resume();
	else Close();
}

LRESULT CWaitForm::OnPaint(WPARAM /*wParam*/, LPARAM /*lParam*/)
{
	PAINTSTRUCT ps;
	HDC dc = BeginPaint(Handle(), &ps);
	HFONT old_font = SelectFont(dc, Font());
	INT old_bk_mode = SetBkMode(dc, TRANSPARENT);
	RECT rect;
	GetClientRect(rect);
	SIZE size;
	GetTextExtentPoint32(dc, message, message.Length(), &size);
	rect.left = (rect.right - size.cx) / 2;
	rect.top = (rect.bottom - size.cy) / 2;
	rect.right = rect.left + size.cx;
	rect.bottom = rect.top + size.cy;
	DrawText(dc, message, -1, &rect, DT_CENTER | DT_NOPREFIX);
	SetBkMode(dc, old_bk_mode);
	SelectFont(dc, old_font);
	EndPaint(Handle(), &ps);
	return 0;
}

bool CWaitForm::OnCloseQuery()
{
	return !thread;
}

void CWaitForm::SetThread(CGUIThread * value)
{
	thread = value;
	thread->Control = this;
}

void CWaitForm::SetMessage(const string &value)
{
	message = value;
	HDC dc = CreateCompatibleDC(NULL);
	HFONT old_font = SelectFont(dc, Font());
	SIZE size;
	GetTextExtentPoint32(dc, message, message.Length(), &size);
	SelectFont(dc, old_font);
	DeleteDC(dc);
	SetClientSize(max(size.cx + 16, 150), max(size.cy + 16, 70));
}

/////////////
//         //
// WaitBox //
//         //
/////////////

void WaitBox(CForm * owner, CGUIThread * thread, const string &message)
{
	CWaitForm *form = new CWaitForm();
	form->SetParent(owner);
	form->SetThread(thread);
	form->SetMessage(message);
	form->ShowModal();
	delete form;
}
