package VeriFinger;

/**
 * VeriFinger library wrapper, connection to native code.<br/>
 * Allows usage of VeriFinger fingerprint matching technology
 * using Java programming language.
 *
 * @version 2.1 05/31/2006
 */

public class VeriFingerWrapper {

        /**
         * VeriFinger software protection types
         * */

		// Registration types
		public static final int VF_RT_NOT_PROTECTED         = 0;
		public static final int VF_RT_UNREGISTERED          = 6;
		public static final int VF_RT_HASP                  = 1;
		public static final int VF_RT_PC                    = 2;
		public static final int VF_RT_UAREU                 = 4;
		public static final int VF_RT_LAN                   = 8;

        // image
        public static final int VF_IMAGE_RESOLUTION         = 500;
        public static final int VF_MIN_IMAGE_RESOLUTION     = 50;
        public static final int VF_MAX_IMAGE_RESOLUTION     = 5000;
        public static final int VF_MIN_IMAGE_DIMENSION      = 16;
        public static final int VF_MAX_IMAGE_DIMENSION      = 2048;
        public static final int VF_MAX_MINUTIA_COUNT        = 1024;
        public static final int VF_MAX_SINGULAR_POINT_COUNT = 64;
        public static final int VF_BLOCK_SIZE               = 16;
        public static final int VF_MAX_BLOCKED_ORIENTS_DIMENSION = (VF_MAX_IMAGE_DIMENSION / VF_BLOCK_SIZE);
        public static final int VF_MAX_FEATURES_SIZE        = 10000;

		public static final int VF_GENERALIZE_COUNT         = 3;

        // Directions
        public static final int VFDIR_0                     = 0;
        public static final int VFDIR_45                    = 30;
        public static final int VFDIR_90                    = (VFDIR_45 * 2);
        public static final int VFDIR_135                   = (VFDIR_45 * 3);
        public static final int VFDIR_180                   = (VFDIR_45 * 4);
        public static final int VFDIR_225                   = (VFDIR_45 * 5);
        public static final int VFDIR_270                   = (VFDIR_45 * 6);
        public static final int VFDIR_315                   = (VFDIR_45 * 7);
        public static final int VFDIR_360                   = (VFDIR_45 * 8);
        public static final int VFDIR_UNKNOWN               = 127;
        public static final int VFDIR_BACKGROUND            = 255;

        // Parameter types
        public static final int VF_TYPE_VOID                = 0;
        public static final int VF_TYPE_BYTE                = 1;
        public static final int VF_TYPE_SBYTE               = 2;
        public static final int VF_TYPE_WORD                = 3;
        public static final int VF_TYPE_SHORT               = 4;
        public static final int VF_TYPE_DWORD               = 5;
        public static final int VF_TYPE_INT                 = 6;
        public static final int VF_TYPE_BOOL                = 10;
        public static final int VF_TYPE_CHAR                = 20;
        public static final int VF_TYPE_STRING              = 100;

        // Singular point types
        public static final int VFFEAT_SP_DELTA             = 3;
        public static final int VFFEAT_SP_CORE              = 1;
        public static final int VFFEAT_SP_DOUBLECORE        = 2;

        // Parameters: General
        public static final int VFP_NAME                    = 10; // String
        public static final int VFP_VERSION_HIGH            = 11; // DWord
        public static final int VFP_VERSION_LOW             = 12; // DWord
        public static final int VFP_COPYRIGHT               = 13; // String

        // Parameters: Features extraction
        public static final int VFP_EXTRACT_FEATURES							= 110; // Int
        public static final int VF_ALL_FEATURES									= 0;
        public static final int VF_OMIT_BLOCKED_ORIENTATIONS                    = 1;
        public static final int VF_OMIT_BLOCKED_ORIENTATIONS_AND_CURVATURES     = 2;

        // Parameters: Features Matching
        public static final int VFP_MATCHING_THRESHOLD          = 200; // Int
        public static final int VFP_MAXIMAL_ROTATION            = 201; // Int | DIR_0..DIR_180
        public static final int VFP_MATCH_FEATURES              = 210; // Int | The values are the same as for the VFP_EXTRACT_FEATURES
        public static final int VFP_MATCHING_SPEED              = 220; // Int | Can be one of the following: Low, High
        public static final int VF_MATCHING_SPEED_LOW           = 0;
        public static final int VF_MATCHING_SPEED_HIGH          = 256;

        // Parameters: Features generalization
        public static final int VFP_GENERALIZATION_THRESHOLD    = 300; // Int | >= 0

        // VeriFinger specific
        public static final int VFP_MODE                                = 1000; // Int | Can be one of the following:
        public static final int VF_MODE_GENERAL                         = 0;
        public static final int VF_MODE_DIGITALPERSONA_URU              = 100;
        public static final int VF_MODE_BIOMETRIKA_FX2000               = 200;
        public static final int VF_MODE_KEYTRONIC_SECUREDESKTOP         = 300;
        public static final int VF_MODE_IDENTIX_TOUCHVIEW               = 400;
        public static final int VF_MODE_PRECISEBIOMETRICS_100CS         = 500;
        public static final int VF_MODE_STMICROELECTRONICS_TOUCHCHIP    = 600;
        public static final int VF_MODE_IDENTICATORTECHNOLOGY_DF90      = 700;
        public static final int VF_MODE_AUTHENTEC_AFS2                  = 800;
        public static final int VF_MODE_AUTHENTEC_AES4000               = 810;
        public static final int VF_MODE_ATMEL_FINGERCHIP                = 900;
        public static final int VF_MODE_BMF_BLP100                      = 1000;
        public static final int VF_MODE_SECUGEN_HAMSTER                 = 1100;
        public static final int VF_MODE_ETHENTICA                       = 1200;

        // FAR and matching threshold ratio
        public static final int      VF_FARMT_COUNT                     = 4;
        public static final double[] VF_FARMT_FAR                       = {0.0, 1.0, 2.0, 3.0}; // FAR 10^-X %
        public static final int[]    VF_FARMT_MT						= {22, 34, 47, 63}; // Matching threshold

		/**
         * Method VFRegistrationType. Returns VeriFinger protection type.
         * @return int protection type
         */
        public static native int VFRegistrationType() throws VeriFingerException, Exception;

        /**
         * Generate registration id from serial number
         * @param serial number
         * @return String registration id
         */
        public static native String VFGenerateId(String serial) throws VeriFingerException, Exception;

        /**
         * Register VeriFinger library
         * @param serial serial number
         * @param key registration key (generated by Neurotechnologija)
         * @return int registration result, 0 indicates failure
         */
        public static native void VFRegister(String serial, String key) throws VeriFingerException, Exception;

		// Initialization
		public static native int VFInitialize() throws VeriFingerException, Exception;
		public static native int VFFinalize() throws VeriFingerException, Exception;

		// Contexts
		public static final int VF_DEFAULT_CONTEXT = 0;
		public static native int VFCreateContext() throws VeriFingerException, Exception;
		public static native void VFFreeContext(int context) throws VeriFingerException, Exception;

		/**
         * Extract features from grayscale image
         * @param width image width
         * @param height image height
         * @param image image data array
         */

		public static native VeriFingerExtractionResult VFExtract(int width, int height, byte[] image, int resolution, byte[] features, int context) throws VeriFingerException, IllegalArgumentException, Exception, ClassNotFoundException;

		// Features generalization
		public static native VeriFingerGeneralizationResult VFGeneralize(int count, byte[][] gen_features, byte[] features, int context) throws VeriFingerException, Exception;

        // Verification
        public static native VeriFingerMatchDetails VFVerify(byte[] features1, byte[] features2, boolean extended_match_info, int context) throws VeriFingerException, Exception;

        // Identification
        public static native void VFIdentifyStart(byte[] test_features, int context) throws VeriFingerException, Exception, IllegalArgumentException;
        public static native VeriFingerMatchDetails VFIdentifyNext(byte[] sample_features, boolean extendedMatchInfo, int context) throws VeriFingerException, Exception, IllegalArgumentException, ClassNotFoundException;
        public static native void VFIdentifyEnd(int context) throws VeriFingerException, Exception, IllegalArgumentException;

        // Conversion from VFDIR_XXX to degrees and vice a versa
        public static double VFDirToDeg(double dir)
        {
                return ((dir) * 180 / VFDIR_180);
        }

        public static double VFDegToDir(double deg)
        {
                return ((deg) * VFDIR_180 / 180);
        }

        /**
         * Get VeriFinger parameter value
         * @param name parameter name <br/>
         * */
        public static native String getParameterValue(int parameter, int context) throws VeriFingerException, Exception;

        /**
         * Set VeriFinger parameter value
         *
         * */
        public static native void setParameterValue(int parameter, String value, int context) throws VeriFingerException, Exception;

        /**
         * Get VeriFinger parameter type
         *
         * */
        public static native int getParameterType(int parameter) throws VeriFingerException, Exception;

    public static native VeriFingerFeatures VFDecompressFeatures(byte[] features) throws VeriFingerException, Exception;
    public static native int VFCompressFeatures(VeriFingerFeatures f, byte[] features) throws VeriFingerException, Exception;

        private VeriFingerWrapper() {
        }

        static
        {
        try {
                System.loadLibrary("VFJavaW42");
                }
                catch (UnsatisfiedLinkError e) {
                        System.out.println("Error loading VeriFinger - " + e.getMessage());
                        e.printStackTrace();
                }
        }
};