unit FormFileListEntry;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, VFFileList, VFinger, Utils;

type
  TFLEFMode = (flefmEditEntry = 0, flefmNewEntry, flefmEnroll, flefmIdentify);

  TFileListEntryForm = class(TForm)
    LabelSource: TLabel;
    EditSource: TEdit;
    LabelResolution: TLabel;
    EditResolution: TEdit;
    LabelDPI: TLabel;
    LabelFingerId: TLabel;
    EditFingerId: TEdit;
    LabelPreview: TLabel;
    ButtonOk: TButton;
    ButtonCancel: TButton;
    ButtonDefault: TButton;
    ButtonHelp: TButton;
    function FormHelp(Command: Word; Data: Integer;
      var CallHelp: Boolean): Boolean;
    procedure ButtonHelpClick(Sender: TObject);
    procedure ButtonDefaultClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormCreate(Sender: TObject);
    procedure EditSourceChange(Sender: TObject);
    procedure EditFingerIdChange(Sender: TObject);
  private
    FDifFID: Boolean;
    FManyEntries: Boolean;
    FDifRes: Boolean;
    FFromScanner: Boolean;
    FSource: string;
    FMode: TFLEFMode;
    procedure SetDifFID(const Value: Boolean);
    procedure SetDifRes(const Value: Boolean);
    procedure SetFingerID(const Value: string);
    procedure SetFromScanner(const Value: Boolean);
    procedure SetManyEntries(const Value: Boolean);
    procedure SetMode(const Value: TFLEFMode);
    procedure SetResolution(const Value: Integer);
    procedure SetSource(const Value: string);
    procedure UpdateMode;
    procedure UpdateSourceEnabled;
    procedure UpdateFromScanner;
    procedure UpdateSource;
    procedure UpdatePreview;
    procedure SetDefaults;
    function GetFingerID: string;
    function GetResolution: Integer;
  public
    property Mode: TFLEFMode read FMode write SetMode;
    property ManyEntries: Boolean read FManyEntries write SetManyEntries;
    property FromScanner: Boolean read FFromScanner write SetFromScanner;
    property DifRes: Boolean read FDifRes write SetDifRes;
    property DifFID: Boolean read FDifFID write SetDifFID;
    property Source: string read FSource write SetSource;
    property Resolution: Integer read GetResolution write SetResolution;
    property FingerID: string read GetFingerID write SetFingerID;
  end;

const
  FLEB_EDIT_ENTRY   = $00000000;
  FLEB_NEW_ENTRY    = $00000001;
  FLEB_ENROLL       = $00000002;
  FLEB_IDENTIFY     = $00000003;
  FLEB_MODE_MASK    = $00000003;
  FLEB_MANY_ENTRIES = $00000004;
  FLEB_FROM_SCANNER = $00000008;
  FLEB_DIF_RES      = $00010000;
  FLEB_DIF_FID      = $00020000;

function FileListEntryBox(Owner: TForm; var Flags: LongWord; var Source: string;
  var Resolution: Integer; var FingerId: string): Boolean;

implementation

{$R *.dfm}

{ FileListEntryBox }

function FileListEntryBox(Owner: TForm; var Flags: LongWord; var Source: string;
  var Resolution: Integer; var FingerId: string): Boolean;
var
  Form: TFileListEntryForm;
begin
  Form := TFileListEntryForm.Create(Owner);
  Form.Mode := TFLEFMode(Flags and FLEB_MODE_MASK);
  Form.ManyEntries := IsFlagSet(Flags, FLEB_MANY_ENTRIES);
  Form.FromScanner := IsFlagSet(Flags, FLEB_FROM_SCANNER);
  Form.DifRes := IsFlagSet(Flags, FLEB_DIF_RES);
  Form.DifFID := IsFlagSet(flags, FLEB_DIF_FID);
  Form.Source := Source;
  if not Form.FromScanner then if not Form.DifRes then Form.Resolution := Resolution;
  Form.FingerID := FingerId;
  Result := Form.ShowModal = mrOk;
  if Result then
  begin
    if not Form.ManyEntries then Source := Form.Source;
    if not Form.FromScanner then if not Form.DifRes then Resolution := Form.Resolution;
    if not Form.DifFID then FingerId := Form.FingerId;
    SetFlag(Flags, FLEB_DIF_RES, Form.DifRes);
    SetFlag(Flags, FLEB_DIF_FID, Form.DifFID);
  end;
  Form.Free;
end;

{ TFileListEntryForm }

function TFileListEntryForm.FormHelp(Command: Word; Data: Integer;
  var CallHelp: Boolean): Boolean;
begin
  MessageDlg(GetFingerIDHelp(FFromScanner), mtInformation, [mbOk], 0);
  Result := True;
end;

procedure TFileListEntryForm.ButtonHelpClick(Sender: TObject);
var
  CallHelp: Boolean;
begin
  FormHelp(0, 0, CallHelp);
end;

procedure TFileListEntryForm.SetDifFID(const Value: Boolean);
begin
  if FDifFID <> Value then
  begin
    FDifFID := Value;
    if FDifFID then
    begin
      EditFingerID.Text := EmptyStr;
      UpdatePreview;
    end;
  end;
end;

procedure TFileListEntryForm.SetDifRes(const Value: Boolean);
begin
  if FDifRes <> Value then
  begin
    FDifRes := Value;
    if FDifRes then EditResolution.Text := EmptyStr;
  end;
end;

procedure TFileListEntryForm.SetFingerID(const Value: string);
begin
  if FDifFID then EditFingerId.Text := EmptyStr
  else EditFingerId.Text := Value;
  UpdatePreview;
end;

procedure TFileListEntryForm.SetFromScanner(const Value: Boolean);
begin
  if FFromScanner <> Value then
  begin
    FFromScanner := Value;
    UpdateFromScanner;
  end;
end;

procedure TFileListEntryForm.SetManyEntries(const Value: Boolean);
begin
  if FManyEntries <> Value then
  begin
    FManyEntries := Value;
    UpdateMode;
    UpdateSourceEnabled;
  end;
end;

procedure TFileListEntryForm.SetMode(const Value: TFLEFMode);
begin
  if FMode <> Value then
  begin
    FMode := Value;
    UpdateMode;
  end;
end;

procedure TFileListEntryForm.SetResolution(const Value: Integer);
begin
  if FDifRes then EditResolution.Text := EmptyStr
  else EditResolution.Text := IntToStr(Value);
end;

procedure TFileListEntryForm.SetSource(const Value: string);
begin
  FSource := Value;
  UpdateSource;
  UpdatePreview;
end;

procedure TFileListEntryForm.UpdateMode;
var
  S: string;
begin
  if FMode = flefmIdentify then S := 'Identify'
  else if FMode = flefmEnroll then S := 'Enroll'
  else begin
    if FMode = flefmNewEntry then S := 'New '
    else S := 'Edit ';
    if FManyEntries then S := S + 'entries' else S := S + 'entry';
  end;
  Caption := S;
end;

procedure TFileListEntryForm.UpdateSourceEnabled;
begin
  EditSource.Enabled := not FManyEntries and not FFromScanner;
  if EditSource.Enabled then ActiveControl := EditSource
  else ActiveControl := EditFingerId;
end;

procedure TFileListEntryForm.UpdateFromScanner;
begin
  if FFromScanner then LabelSource.Caption := '&Scanner'
  else LabelSource.Caption := '&Filename';
  LabelResolution.Visible := not FFromScanner;
  EditResolution.Visible := not FFromScanner;
  LabelDPI.Visible := not FFromScanner;
  UpdateSourceEnabled;
  UpdateSource;
end;

procedure TFileListEntryForm.UpdateSource;
var
  Text: string;
begin
  Text := Source;
  if FManyEntries and not FFromScanner then
  begin
    if Text <> EmptyStr then Text := Text + ' ';
    Text := Text + '...';
  end;
  EditSource.Text := Text;
  UpdatePreview;
end;

procedure TFileListEntryForm.UpdatePreview;
begin
  try
    LabelPreview.Caption := MakeFingerID(EditFingerId.Text, FSource, FFromScanner);
  except
    on E: Exception do
      LabelPreview.Caption := Format('Error: %s', [E.Message]);
  end;
end;

procedure TFileListEntryForm.SetDefaults;
begin
  if not FFromScanner then SetResolution(VF_IMAGE_RESOLUTION);
  if FFromScanner then SetFingerID(DefaultScannerFingerId)
  else SetFingerID(DefaultFileFingerId);
end;

procedure TFileListEntryForm.ButtonDefaultClick(Sender: TObject);
begin
  SetDefaults;
end;

function TFileListEntryForm.GetFingerID: string;
begin
  Result := EditFingerId.Text;
end;

function TFileListEntryForm.GetResolution: Integer;
begin
  if FDifRes then Result := -1
  else Result := StrToInt(EditResolution.Text);
end;

procedure TFileListEntryForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  if FDifRes then FDifRes := EditResolution.Text = EmptyStr;
  if FDifFID then FDifFID := EditFingerId.Text = EmptyStr;
end;

procedure TFileListEntryForm.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
var
  I: Integer;
begin
  try
    if ModalResult = mrOk then
    begin
      if (EditSource.Text = EmptyStr) and not FManyEntries then
      begin
        EditSource.SetFocus;
        raise Exception.Create('Please specify file name');
      end;
      if not FFromScanner then
      begin
        if (EditResolution.Text = EmptyStr) and not (FManyEntries and FDifRes) then
        begin
          EditResolution.SetFocus;
          raise Exception.Create('Please specify resolution');
        end
        else if EditResolution.Text <> EmptyStr then
        begin
          try
            I := StrToInt(EditResolution.Text);
          except
            EditResolution.SetFocus;
            raise Exception.Create('Resolution have to be an integer value');
          end;
          if I <= 0 then
          begin
            EditResolution.SetFocus;
            raise Exception.Create('Resolution have to be greater than zero');
          end;
        end;
      end;
      try
        CheckFingerID(EditFingerId.Text, FFromScanner);
      except
        on E: Exception do
        begin
          EditFingerId.SetFocus;
          raise Exception.CreateFmt('Error processing finger id:'#13'%s', [E.Message]);
        end;
      end;
    end;
    CanClose := True;
  except
    on E: Exception do
    begin
      MessageDlg(E.Message, mtError, [mbOk], 0);
      CanClose := False;
    end;
  end;
end;

procedure TFileListEntryForm.FormCreate(Sender: TObject);
begin
  UpdateMode;
  UpdateFromScanner;
end;

procedure TFileListEntryForm.EditSourceChange(Sender: TObject);
begin
  if not FFromScanner and not FManyEntries then
    FSource := EditSource.Text;
  UpdatePreview;
end;

procedure TFileListEntryForm.EditFingerIdChange(Sender: TObject);
begin
  UpdatePreview;
end;

end.
