unit FormMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Menus, ActnList, ExtCtrls, StdCtrls, ActiveX,
  Utils, Log, VFinger, ScanMan, VFImage, VFFeatures, VFControls, VFOptions, VFDatabase,
  FormFileList, VFFileList, FormFileListEntry, FormWait,
  ComCtrls, DB, ADODB, FormOptions, FormAbout;

const
  IDR_DATABASE = 1;

  UM_FIRST_SHOW = WM_USER;

  IDEN_PROGRESS_COUNT = 1000;

type
  TTaskStateEvent = procedure(Sender: TObject; State: string) of object;
  TTaskProgressEvent = function(Sender: TObject; Started: Boolean; Value: Integer): Boolean of object;

  TTaskThread = class(TThread)
  private
    FTaskName: string;
    FTaskString: string;

    FState: string;
    FStarted: Boolean;
    FValue: Integer;
    FContinue: Boolean;
    FOnState: TTaskStateEvent;
    FOnProgress: TTaskProgressEvent;

    procedure SynchronizedState;
    procedure SynchronizedProgress;

  protected
    FFailed: Boolean;

    procedure CheckRunning;

    procedure DoState(State: string);
    function DoProgress(Started: Boolean; Value: Integer): Boolean;

  public
    constructor Create(TaskName: string; TaskString: string);

    property TaskName: string read FTaskName;
    property TaskString: string read FTaskString;

    property Failed: Boolean read FFailed;

    property OnState: TTaskStateEvent read FOnState write FOnState;
    property OnProgress: TTaskProgressEvent read FOnProgress write FOnProgress;

  end;

  TDBThread = class(TTaskThread)
  private
    FError: string;
    procedure SetDatabase(const Value: TVFDatabase);
  protected
    FDatabase: TVFDatabase;
    procedure Execute; override;
    procedure DBExecute; virtual; abstract;
  public
    constructor Create(Database: TVFDatabase; OperationName, OperationString: string);
    property Database: TVFDatabase read FDatabase write SetDatabase; { CheckRunning(); database = value; }
    property Error: string read FError;
  end;

  TFileListThread = class(TTaskThread)
  private
    FFileList: TVFFileList;
  protected
    procedure Execute; override;
  public
    constructor Create(FileList: TVFFileList);
  end;

  TDBOpenThread = class(TDBThread, IVFDBOpenCallback)
  private
    FProvider: string;
    FDataSource: string;
    FUserID: string;
    FPassword: string;
    function _AddRef: Integer; stdcall;
    function _Release: Integer; stdcall;
    function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
    procedure OnOpenState(State: string);
    function OnOpenProgress(Started: Boolean; Value: Integer): Boolean;
  protected
    procedure DBExecute; override;
  public
    constructor Create(Database: TVFDatabase = nil;
      Provider: string = ''; DataSource: string = ''; UserID: string = ''; Password: string = '');
  end;

  TDBCloseThread = class(TDBThread)
  protected
    procedure DBExecute; override;
  public
    constructor Create(Database: TVFDatabase = nil);
  end;

  TDBClearThread = class(TDBThread)
  protected
    procedure DBExecute; override;
  public
    constructor Create(Database: TVFDatabase = nil);
  end;

  TExtractionThread = class(TThread)
  private
    FResolution: Integer;
    FImage: TVFImage;
    FImageAfter: TVFImage;
    FFeatures: array[0..VF_MAX_FEATURES_SIZE - 1] of Byte;
    FFeaturesSize: LongWord;
    FTime: LongWord;
    FResult: Integer;
    function GetFeatures: PByte;
  protected
    procedure Execute; override;
  public
    constructor Create(Resolution: Integer; Image, ImageAfter: TVFImage);
    property Features: PByte read GetFeatures;
    property FeaturesSize: LongWord read FFeaturesSize;
    property Time: LongWord read FTime;
    property Result: Integer read FResult;
  end;

  PIdenProgress = ^TIdenProgress;
  TIdenProgress = record
    Count: Integer;
    Index: array[0..IDEN_PROGRESS_COUNT - 1] of Integer;
    Result: array[0..IDEN_PROGRESS_COUNT - 1] of Integer;
    Similarity: array[0..IDEN_PROGRESS_COUNT - 1] of Integer;
  end;

  TIdenProgressEvent = function(Sender: TObject; Started: Boolean; Value: PIdenProgress): Boolean of object;

  TIdentificationThread = class(TThread)
  private
    FFeatures: PByte;
    FFeaturesSize: LongWord;
    FDatabase: TVFDatabase;
    FUseG: Boolean;
    FDatabasePercent: Integer;
    FMatchAll: Boolean;
    FResult: Integer;
    FIndex: Integer;
    FSimilarity: Integer;
    FTime: Double;
    FStarted: Boolean;
    FValue: PIdenProgress;
    FContinue: Boolean;
    FOnProgress: TIdenProgressEvent;
    function DoProgress(Started: Boolean; Value: PIdenProgress): Boolean;
    procedure SynchronizedProgress;
  protected
     procedure Execute; override;
  public
    constructor Create(Features: PByte; FeaturesSize: LongWord; Database: TVFDatabase; UseG: Boolean;
      DatabasePercent: Integer; MatchAll: Boolean);
    destructor Destroy; override;
    property Result: Integer read FResult;
    property Index: Integer read FIndex;
    property Similarity: Integer read FSimilarity;
    property Time: Double read FTime;
    property OnProgress: TIdenProgressEvent read FOnProgress write FOnProgress;
  end;

  TMainForm = class(TForm)
    Actions: TActionList;
    MainMenu: TMainMenu;
    FileClear: TAction;
    FileOpen: TAction;
    FileOpenFileList: TAction;
    FileSave: TAction;
    FileSaveRight: TAction;
    FileStop: TAction;
    FileExit: TAction;
    miFile: TMenuItem;
    miFileClear: TMenuItem;
    msFile1: TMenuItem;
    miFileOpen: TMenuItem;
    miFileOpenFileList: TMenuItem;
    msFile2: TMenuItem;
    miFileSave: TMenuItem;
    miFileSaveRight: TMenuItem;
    msFile3: TMenuItem;
    miFileStop: TMenuItem;
    msFile4: TMenuItem;
    miFileExit: TMenuItem;
    miMode: TMenuItem;
    miView: TMenuItem;
    miTools: TMenuItem;
    miHelp: TMenuItem;
    ModeEnrollment: TAction;
    ModeEnrollmentWithGen: TAction;
    ModeVerification: TAction;
    ModeIdentification: TAction;
    ViewZoomIn: TAction;
    ViewZoomOut: TAction;
    ViewZoom1To1: TAction;
    ViewShowFeatures: TAction;
    ViewShowSingularPoints: TAction;
    ViewShowBlockedOrientations: TAction;
    ToolsFileListEditor: TAction;
    ToolsClearDatabase: TAction;
    ToolsClearMainLog: TAction;
    ToolsClearScannerLog: TAction;
    ToolsOptions: TAction;
    HelpAbout: TAction;
    miModeEnrollment: TMenuItem;
    miModeEnrollmentWithGen: TMenuItem;
    miModeVerification: TMenuItem;
    miModeIdentification: TMenuItem;
    miViewZoomIn: TMenuItem;
    miViewZoomOut: TMenuItem;
    miViewZoom1To1: TMenuItem;
    msView: TMenuItem;
    miViewShowFeatures: TMenuItem;
    miViewShowSingularPoints: TMenuItem;
    miViewShowBlockedOrientations: TMenuItem;
    miToolsFileListEditor: TMenuItem;
    msTools1: TMenuItem;
    miToolsClearDatabase: TMenuItem;
    msTools2: TMenuItem;
    miToolsClearMainLog: TMenuItem;
    miToolsClearScannerLog: TMenuItem;
    msTools3: TMenuItem;
    miToolsOptions: TMenuItem;
    miHelpAbout: TMenuItem;
    SplitterVert: TSplitter;
    SplitterHorz: TSplitter;
    PanelBottom: TPanel;
    StatusBar: TStatusBar;
    CaptionPanelMain: TPanel;
    SplitterV1: TSplitter;
    CaptionPanelScanner: TPanel;
    SplitterV2: TSplitter;
    CaptionPanelResults: TPanel;
    CaptionLabelMain: TLabel;
    CaptionLabelScanner: TLabel;
    CaptionLabelResults: TLabel;
    PanelMain: TPanel;
    PanelResults: TPanel;
    LabelMain: TLabel;
    ProgressBarMain: TProgressBar;
    ButtonMainStop: TButton;
    LabelMainTime: TLabel;
    ProgressBarResults: TProgressBar;
    ButtonResultsStop: TButton;
    LabelResultsTime: TLabel;
    RichEditMain: TRichEdit;
    RichEditScanner: TRichEdit;
    ListViewResults: TListView;
    OpenDialog: TOpenDialog;
    SaveDialog: TSaveDialog;
    HelpRegister: TAction;
    HelpClearRegistration: TAction;
    OpenListDialog: TOpenDialog;
    msMode1: TMenuItem;
    msMode2: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure SplitterVertMoved(Sender: TObject);
    procedure SplitterHorzMoved(Sender: TObject);
    procedure SplitterV1Moved(Sender: TObject);
    procedure SplitterV2CanResize(Sender: TObject; var NewSize: Integer;
      var Accept: Boolean);
    procedure FormResize(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FileClearExecute(Sender: TObject);
    procedure FileOpenExecute(Sender: TObject);
    procedure ToolsFileListEditorExecute(Sender: TObject);
    procedure FileStopExecute(Sender: TObject);
    procedure ButtonMainStopClick(Sender: TObject);
    procedure ButtonResultsStopClick(Sender: TObject);
    procedure FileExitExecute(Sender: TObject);
    procedure FileOpenFileListExecute(Sender: TObject);
    procedure FileSaveExecute(Sender: TObject);
    procedure FileSaveRightExecute(Sender: TObject);
    procedure ModeEnrollmentExecute(Sender: TObject);
    procedure ModeEnrollmentWithGenExecute(Sender: TObject);
    procedure ModeVerificationExecute(Sender: TObject);
    procedure ModeIdentificationExecute(Sender: TObject);
    procedure ViewZoomInExecute(Sender: TObject);
    procedure ViewZoomOutExecute(Sender: TObject);
    procedure ViewZoom1To1Execute(Sender: TObject);
    procedure ViewShowFeaturesExecute(Sender: TObject);
    procedure ViewShowSingularPointsExecute(Sender: TObject);
    procedure ViewShowBlockedOrientationsExecute(Sender: TObject);
    procedure ToolsClearDatabaseExecute(Sender: TObject);
    procedure ToolsClearMainLogExecute(Sender: TObject);
    procedure ToolsClearScannerLogExecute(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure ToolsOptionsExecute(Sender: TObject);
    procedure HelpAboutExecute(Sender: TObject);
  private
    ShowedOnce: Boolean;
    ViewLeft, ViewRight: TVFView;
    SplitterVertRatio, SplitterHorzRatio, SplitterV1Ratio, SplitterV2Ratio: Double;

    MainLog, ScannerLog: TLog;

    Database: TVFDatabase;

    TaskThread: TTaskThread;
    ExtractionThread: TExtractionThread;
    IdentificationThread: TIdentificationThread;
    FStopTask, FStopIdentification: Boolean;

    Resolution: Integer;
    Image: TVFImage;
    ImageAfter: TVFImage;
    Features: TVFFeatures;
    FingerID: string;
    ImageFromScanner: Boolean;
    ScannerID: string;

    GenIndex: Integer;
    GenResolution: array[0..VF_GENERALIZE_COUNT - 1] of Integer;
    GenImage: array[0..VF_GENERALIZE_COUNT - 1] of TVFImage;
    GenImageAfter: array[0..VF_GENERALIZE_COUNT - 1] of TVFImage;
    GenFeatures: array[0..VF_GENERALIZE_COUNT - 1] of TVFFeatures;
    GenFingerID: array[0..VF_GENERALIZE_COUNT - 1] of string;

    VerIndex: Integer;
    VerResolution: array[0..1] of Integer;
    VerImage: array[0..1] of TVFImage;
    VerImageAfter: array[0..1] of TVFImage;
    VerFeatures: array[0..1] of TVFFeatures;
    VerFingerID: array[0..1] of string;

    Mode, LastMode: TVFDemoMode;
    ChangeBackMode: Boolean;

    FileList: TVFFileList;

    Options: TVFOptions;

    TaskTimeStarted: LongWord;
    TaskUITime: LongWord;
    IdentificationTimeStarted: Double;

    FingerFound: Boolean;
    Count, CountRej: Integer;
    CountIden, TotalIden: Integer;
    TotalSize: LongWord;
    TotalExtTime: LongWord;
    TotalIdenTime: Double;
    FingersFound: Integer;
    FalseAccepted, FalseRejected: Integer;

    procedure UMFirstShow(var Message: TMessage); message UM_FIRST_SHOW;
    procedure FormFirstShow(Sender: TObject);

    procedure AddVersionToLog(Log: TLog; Name: string; VersionHigh, VersionLow: LongWord);
    function FingerprintCountToString(Value: Integer; No: string = ''): string;
    function MatchingSpeedToString(Count: Integer; Time: Double; Unknown: string = 'unknown'): string;
    function PercentToString(Count, Total: Integer; Unknown: string = 'Unknown'): string;

    procedure SetMode(Value: TVFDemoMode);
    function GetModeString: string;
    procedure UpdateMode;

    procedure UpdateStatus;

    procedure StartTask(Thread: TTaskThread);
    procedure OnTaskState(Sender: TObject; State: string);
    function OnTaskProgress(Sender: TObject; Started: Boolean; Value: Integer): Boolean;
    procedure OnTaskComplete(Sender: TObject);
    function GetTaskName: string;
    function GetTaskString: string;
    procedure UpdateTask;
    procedure UpdateImagesMenu;
    procedure UpdateViewMenu;
    procedure UpdateTaskProgress(Started: Boolean; Value: Integer; Text: string = '');
    procedure UpdateIdentificationProgress(Started: Boolean; Value: Integer);
    function TaskRunning: Boolean;
    function CanStopTask: Boolean;
    function StoppingTask: Boolean;
    function MainTaskRunning: Boolean;
    function DBOpenRunning: Boolean;
    function FileListRunning: Boolean;
    function IdentificationRunning: Boolean;
    function StoppingIdentification: Boolean;
    procedure Clear;
    procedure StopTask;
    procedure StopIdentification;
    function StopTaskPrompt: Boolean;
    function ErrorPrompt(Reason: string; Warning: Boolean = False): Boolean;

    procedure CreateDefaultDatabase(Filename: string);
    procedure StartDBOpen;
    procedure OnDBOpenComplete(Time: LongWord);

    procedure Enroll;
    procedure EnrollWithGen;
    procedure Verify;
    function StartIdentification: Boolean;

    procedure StartFileList;
    procedure OnFileListImage(FileIndex: Integer);
    procedure OnFileListComplete(Time: LongWord);

    function OnFileImage(Filename: string; Resolution: Integer; FingerID: string): Boolean;

    procedure OnSMMonitor(ScannerID: string; Event: LongWord);
    procedure OnScannerImage(ScannerID: string; Width, Height: Integer; Image: PByte; Resolution: Integer);
    procedure OnScannerState(ScannerID: string; State: LongWord);

    procedure OnImage;
    procedure OnExtractionComplete(Sender: TObject);

    function OnIdentificationProgress(Sender: TObject; Started: Boolean; Progress: PIdenProgress): Boolean;
    procedure OnIdentificationComplete(Sender: TObject);
    procedure SaveImage(Image: TVFImage);

  end;

var
  MainForm: TMainForm;

implementation

{$R *.dfm}

{ TTaskThread }

constructor TTaskThread.Create(TaskName, TaskString: string);
begin
  inherited Create(True);
  FreeOnTerminate := True;
  FTaskName := TaskName;
  FTaskString := TaskString;
  FFailed := False;
end;

function TTaskThread.DoProgress(Started: Boolean; Value: Integer): Boolean;
begin
  if Assigned(FOnState) then
  begin
    FStarted := Started;
    FValue := Value;
    Synchronize(SynchronizedProgress);
    Result := FContinue;
  end
  else Result := True;
end;

procedure TTaskThread.DoState(State: string);
begin
  if Assigned(FOnState) then
  begin
    FState := State;
    Synchronize(SynchronizedState);
  end;
end;

procedure TTaskThread.SynchronizedState;
begin
  FOnState(Self, FState);
end;

procedure TTaskThread.SynchronizedProgress;
begin
  FContinue := FOnProgress(Self, FStarted, FValue)
end;

procedure TTaskThread.CheckRunning;
begin
  if not Suspended then raise Exception.Create('Thread running');
end;

{ TDBThread }

constructor TDBThread.Create(Database: TVFDatabase; OperationName,
  OperationString: string);
begin
  inherited Create(OperationName, OperationString);
  FDatabase := Database;
end;

procedure TDBThread.Execute;
begin
  try
    CoInitialize(nil);
    DBExecute;
    CoUninitialize;
  except
    on E: Exception do
    begin
      FError := E.Message;
      FFailed := True;
    end;
  end;
end;

procedure TDBThread.SetDatabase(const Value: TVFDatabase);
begin
  CheckRunning;
  FDatabase := Value;
end;

{ TDBOpenThread }

constructor TDBOpenThread.Create(Database: TVFDatabase; Provider,
  DataSource, UserID, Password: string);
begin
  inherited Create(Database, 'Database open', 'Opening database');
  FDatabase := Database;
  FProvider := Provider;
  FDataSource := DataSource;
  FUserID := UserID;
  FPassword := Password;
end;

procedure TDBOpenThread.DBExecute;
begin
  Database.Open(FProvider, FDataSource, FUserID, FPassword, Self);
end;

function TDBOpenThread.OnOpenProgress(Started: Boolean; Value: Integer): Boolean;
begin
  Result := DoProgress(Started, Value);
end;

procedure TDBOpenThread.OnOpenState(State: string);
begin
  DoState(State);
end;

function TDBOpenThread._AddRef: Integer;
begin
  Result := 0;
end;

function TDBOpenThread._Release: Integer;
begin
  Result := 0;
end;

function TDBOpenThread.QueryInterface(const IID: TGUID; out Obj): HResult;
begin
  Result := E_NOINTERFACE;
end;

{ TDBCloseThread }

constructor TDBCloseThread.Create(Database: TVFDatabase);
begin
  inherited Create(Database, 'Database close', 'Closing database');
end;

procedure TDBCloseThread.DBExecute;
begin
  Database.Close;
end;

{ TDBClearThread }

constructor TDBClearThread.Create(Database: TVFDatabase);
begin
  inherited Create(Database, 'Database clear', 'Clearing database');
end;

procedure TDBClearThread.DBExecute;
begin
  Database.Clear;
end;

{ TFileListThread }

constructor TFileListThread.Create(FileList: TVFFileList);
begin
  inherited Create('File list', 'Processing file list');
  FFileList := FileList;
end;

procedure TFileListThread.Execute;
var
  I: Integer;
begin
  DoProgress(False, FFileList.Count);
  for I := 0 to FFileList.Count - 1 do
    if not DoProgress(True, I) then Break;
end;

{ TExtractionThread }

constructor TExtractionThread.Create(Resolution: Integer; Image, ImageAfter: TVFImage);
begin
  inherited Create(True);
  FreeOnTerminate := True;
  FResult := VFE_FAILED;
  FTime := 0;
  FResolution := Resolution;
  FImage := Image;
  FImageAfter := ImageAfter;
end;

procedure TExtractionThread.Execute;
var
  TC: LongWord;
begin
  FImageAfter.Assign(FImage);
  TC := GetTickCount;
  FResult := VFExtract(FImage.Width, FImage.Height, FImageAfter.Bits, FResolution, @FFeatures[0], FFeaturesSize, nil);
  FTime := GetTickCount - TC;
end;

function TExtractionThread.GetFeatures: PByte;
begin
  Result := @FFeatures[0];
end;

{ TIdentificationThread }

constructor TIdentificationThread.Create(Features: PByte; FeaturesSize: LongWord;
  Database: TVFDatabase; UseG: Boolean; DatabasePercent: Integer;
  MatchAll: Boolean);
begin
  inherited Create(True);
  FreeOnTerminate := True;
  FUseG := UseG;
  FDatabasePercent := DatabasePercent;
  FMatchAll := MatchAll;
  FResult := VFE_FAILED;
  FIndex := -1;
  FSimilarity := 0;
  FTime := 0;
  GetMem(FFeatures, FeaturesSize);
  FFeaturesSize := FeaturesSize;
  Move(Features^, FFeatures^, FeaturesSize);
  FDatabase := Database;
end;

destructor TIdentificationThread.Destroy;
begin
  FreeMem(FFeatures);
  inherited;
end;

function TIdentificationThread.DoProgress(Started: Boolean;
  Value: PIdenProgress): Boolean;
begin
  if Assigned(FOnProgress) then
  begin
    FStarted := Started;
    FValue := Value;
    Synchronize(SynchronizedProgress);
    Result := FContinue;
  end
  else Result := True;
end;

procedure TIdentificationThread.Execute;
var
  FD: TVFDBFindData;
  C: Integer;
  CountThreshold: Integer;
  Count: Integer;
  IP: TIdenProgress;
  MD: TVFMatchDetails;
  Stop: Boolean;
  I: Integer;
  DT: LongWord;
begin
  FResult := VFIdentifyStart(FFeatures, nil);
  if VFSucceeded(FResult) then
  begin
    if FUseG then FDatabase.FindStart(FFeatures, FD);
    C := FDatabase.Count;
    CountThreshold := C * FDatabasePercent div 100;
    Count := 0;
    DoProgress(False, PIdenProgress(C));
    IP.Count := 0;
    MD.Size := SizeOf(MD);
    Stop := False;
    I := -1;
    DT := GetTickCount;
    repeat begin
      if FUseG then
      begin
        I := FDatabase.FindNext(FD);
        if (I = -1) or (Count > CountThreshold) then I := C;
      end
      else Inc(I);
      if I < C then
      begin
        FTime := FTime - GetTickCountEx;
        FResult := VFIdentifyNext(FDatabase.Features[I], @MD, nil);
        FTime := FTime + GetTickCountEx;
        IP.Index[IP.Count] := I;
        IP.Result[IP.Count] := FResult;
        IP.Similarity[IP.Count] := MD.Similarity;
        Inc(IP.Count);
        if VFSucceeded(FResult) then
        begin
          if (Index = -1) or (MD.Similarity > FSimilarity) then
          begin
            FIndex := I;
            FSimilarity := MD.Similarity;
            Stop := not FMatchAll;
          end;
        end;
        Inc(Count);
      end
      else Stop := True;
      if (IP.Count = IDEN_PROGRESS_COUNT) or (Stop) or (GetTickCount - DT > 300) then
      begin
        if not DoProgress(True, @IP) then Break;
        IP.Count := 0;
        DT := GetTickCount;
      end;
    end until Stop;
    FResult := VFIdentifyEnd(nil);
    if FUseG then FDatabase.FindEnd(FD);
    if VFSucceeded(FResult) then
      if Index = -1 then FResult := VFE_FAILED else FResult := VFE_OK;
  end;
end;

procedure TIdentificationThread.SynchronizedProgress;
begin
  FContinue := FOnProgress(Self, FStarted, FValue)
end;

{ TMainForm }

procedure SMMonitorProc(const ScannerID: PAnsiChar; Event: LongWord; Param: Pointer); stdcall;
begin
  TMainForm(Param).OnSMMonitor(ScannerID, Event);
end;

procedure SMImageProc(const ScannerID: PAnsiChar; Width, Height: Integer;
  const Image: PByte; Resolution: Integer; Param: Pointer); stdcall;
begin
  TMainForm(Param).OnScannerImage(ScannerID, Width, Height, Image, Resolution);
end;

procedure SMStateProc(const ScannerID: PAnsiChar; State: LongWord; Param: Pointer); stdcall;
begin
  TMainForm(Param).OnScannerState(ScannerID, State);
end;

procedure TMainForm.FormCreate(Sender: TObject);
var
  I: Integer;
  Result: Integer;
  IDs: TStringList;
begin
  ShowedOnce := False;
  ViewLeft := TVFView.Create(Self);
  ViewLeft.Parent := Self;
  ViewLeft.Align := alLeft;
  ViewRight := TVFView.Create(Self);
  ViewRight.Parent := Self;
  ViewRight.Align := alClient;
  ViewRight.ColorMin := RGB(0, 230, 0);
  SplitterVertRatio := 0.5;
  SplitterHorzRatio := 0.4;
  SplitterV1Ratio := 0.4;
  SplitterV2Ratio := 0.4;

  OpenDialog.Filter := ImageIO.GetLoadFileFilterString();
  SaveDialog.Filter := ImageIO.GetSaveFileFilterString();
  OpenListDialog.Filter := TVFFileList.FilterString;
  OpenListDialog.Title := 'Open file list';

  MainLog := TLog.Create;
  MainLog.Control := RichEditMain;
  ScannerLog := TLog.Create;
  ScannerLog.Control := RichEditScanner;

  ImageFromScanner := False;
  TaskThread := nil;
  ExtractionThread := nil;
  IdentificationThread := nil;
  FStopTask := False;
  FStopIdentification := False;
  Mode := vfdmNone;

  Database := TVFDatabase.Create;

  Image := TVFImage.Create;
  ImageAfter := TVFImage.Create;
  Features := TVFFeatures.Create;

  for I := 0 to VF_GENERALIZE_COUNT - 1 do
  begin
    GenImage[i] := TVFImage.Create;
    GenImageAfter[i] := TVFImage.Create;
    GenFeatures[i] := TVFFeatures.Create;
  end;
  GenIndex := 0;

  for I := 0 to 1 do
  begin
    VerImage[I] := TVFImage.Create;
    VerImageAfter[I] := TVFImage.Create;
    VerFeatures[I] := TVFFeatures.Create;
  end;
  VerIndex := 0;

  FileList := TVFFileList.Create;

  AddVersionToLog(MainLog, Application.Title, ApplicationVersionHigh, ApplicationVersionLow);
  MainLog.Add('OS: %s. Version %u.%u (Build %u). Service Pack Version %u.%u',
    [OSName, OSMajorVersion, OSMinorVersion, OSBuild,
    OSSPMajorVersion, OSSPMinorVersion]);
  AddVersionToLog(MainLog, VFGetName, VFGetVersionHigh, VFGetVersionLow);
  MainLog.AddEmpty;

  AddVersionToLog(ScannerLog, SMGetName, SMGetVersionHigh, SMGetVersionLow);
  ScannerLog.AddEmpty;

  UpdateTaskProgress(False, 0);
  UpdateIdentificationProgress(False, 0);

  LoadOptions(Options, ApplicationRegistryKey);
  UpdateOptions(Options);
  LastMode := Options.StartupMode;
  ChangeBackMode := True;

  UpdateStatus;

  UpdateMode;
  UpdateViewMenu;

  Result := SMSetMonitor(SMMonitorProc, Self);
  ScannerLog.Add('SetMonitor: %s', [SMResultToString(Result)]);
  ScannerLog.AddEmpty;
  IDs := TStringList.Create;
  SMGetScannerIds(IDs);
  ScannerLog.Add(IntToString(IDs.Count, 'scanner', 'scanners') + ' found');
  ScannerLog.AddEmpty;
  for I := 0 to IDs.Count - 1 do
    ScannerLog.Add('Starting capturing from scanner %s: %s', [IDs[I],
      SMResultToString(SMStartCapturing(IDs[I], SMImageProc, SMStateProc, Self))]);
  if ScannerLog.LastLine <> EmptyStr then ScannerLog.AddEmpty;
  IDs.Clear;
  IDs.Destroy;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
var
  I: Integer;
begin
  SMStopCapturing;
  SMRemoveMonitor;

  SaveOptions(Options, ApplicationRegistryKey);

  FileList.Free;

  for I := 0 to VF_GENERALIZE_COUNT - 1 do
  begin
    GenImage[I].Free;
    GenImageAfter[I].Free;
    GenFeatures[I].Free;
  end;

  for I := 0 to 1 do
  begin
    VerImage[I].Free;
    VerImageAfter[I].Free;
    VerFeatures[I].Free;
  end;

  Image.Free;
  ImageAfter.Free;
  Features.Free;

  Database.Free;

  ScannerLog.Free;
  MainLog.Free;
end;

procedure TMainForm.SplitterVertMoved(Sender: TObject);
begin
  if ClientWidth <> 0 then SplitterVertRatio := ViewLeft.Width / ClientWidth
  else SplitterVertRatio := 0;
end;

procedure TMainForm.SplitterHorzMoved(Sender: TObject);
begin
  if ClientHeight <> 0 then SplitterHorzRatio := PanelBottom.Height / ClientHeight
  else SplitterHorzRatio := 0;
end;

procedure TMainForm.SplitterV1Moved(Sender: TObject);
begin
  if ClientWidth <> 0 then SplitterV1Ratio := CaptionPanelMain.Width / ClientWidth
  else SplitterV1Ratio := 0;
end;

procedure TMainForm.SplitterV2CanResize(Sender: TObject;
  var NewSize: Integer; var Accept: Boolean);
begin
  if ClientWidth <> 0 then SplitterV2Ratio := CaptionPanelResults.Width / ClientWidth
  else SplitterV1Ratio := 0;
end;

procedure TMainForm.FormResize(Sender: TObject);
begin
  ViewLeft.Width := Round(SplitterVertRatio * ClientWidth);
  ViewLeft.Left := 0;
  PanelBottom.Height := Round(SplitterHorzRatio * ClientHeight);
  CaptionPanelMain.Width := Round(SplitterV1Ratio * ClientWidth);
  CaptionPanelResults.Width := Round(SplitterV2Ratio * ClientWidth);
  StatusBar.Top := ClientHeight;
end;

procedure TMainForm.FormShow(Sender: TObject);
begin
  if not ShowedOnce then
  begin
    ShowedOnce := True;
    PostMessage(Handle, UM_FIRST_SHOW, 0, 0);
  end;
end;

procedure TMainForm.UMFirstShow(var Message: TMessage);
begin
  FormFirstShow(Self);
end;

procedure TMainForm.FormFirstShow(Sender: TObject);
var
  RegistrationType: Integer;
begin
  StartDBOpen;
end;

procedure TMainForm.FileClearExecute(Sender: TObject);
begin
  Clear;
end;

procedure TMainForm.FileOpenExecute(Sender: TObject);
var
  Ok: Boolean;
  Flags: LongWord;
  Filename: string;
  I: Integer;
begin
  if not TaskRunning and (Mode <> vfdmNone) then
  begin
    OpenDialog.FileName := EmptyStr;
    if OpenDialog.Execute then
    begin
      if Mode = vfdmEnrollmentWithGen then
        if OpenDialog.Files.Count <> VF_GENERALIZE_COUNT then
        begin
          MessageDlg(Format('%s requires %d images', [GetModeString, VF_GENERALIZE_COUNT]), mtError, [mbOk], 0);
          Exit;
        end
        else GenIndex := 0
      else if Mode = vfdmVerification then
        if OpenDialog.Files.Count <> 2 then
        begin
          MessageDlg(Format('%s verification requires 2 images', [GetModeString]), mtError, [mbOk], 0);
          Exit;
        end
        else VerIndex := 0;
      if Options.FilePrompt and ((Mode = vfdmEnrollment) or (Mode = vfdmEnrollmentWithGen) or (Mode = vfdmIdentification)) then
      begin
        if Mode = vfdmIdentification then Flags := FLEB_IDENTIFY else Flags := FLEB_ENROLL;
        if OpenDialog.Files.Count > 1 then SetFlag(Flags, FLEB_MANY_ENTRIES);
        Filename := OpenDialog.FileName;
        Ok := FileListEntryBox(Self, Flags, Filename, Options.FileResolution, Options.FileFingerId);
      end
      else Ok := True;
      if Ok then
      begin
        FileList.Clear;
        for I := 0 to OpenDialog.Files.Count - 1 do
          FileList.Add(OpenDialog.Files[I], Options.FileResolution, MakeFingerID(Options.FileFingerId, OpenDialog.Files[I], False));
        StartFileList;
      end;
    end;
  end;
end;

procedure TMainForm.Clear;
var
  I: Integer;
begin
  if not TaskRunning then
  begin
    ViewLeft.Image := nil;
    ViewRight.Image := nil;
    ViewRight.Features := nil;
    Image.Clear;
    ImageAfter.Clear;
    Features.Clear;
    FingerID := EmptyStr;
    for I := 0 to VF_GENERALIZE_COUNT - 1 do
    begin
      GenImage[I].Clear;
      GenImageAfter[I].Clear;
      GenFeatures[I].Clear;
      GenFingerID[I] := EmptyStr;
    end;
    GenIndex := 0;
    for I := 0 to 1 do
    begin
      VerImage[I].Clear;
      VerImageAfter[I].Clear;
      VerFeatures[I].Clear;
      VerFingerID[I] := EmptyStr;
    end;
    VerIndex := 0;
    ListViewResults.Items.Clear;
    UpdateTask;
  end;
end;

function TMainForm.TaskRunning: Boolean;
begin
  Result := (TaskThread <> nil) or (ExtractionThread <> nil) or (IdentificationThread <> nil);
end;

function TMainForm.OnFileImage(Filename: string; Resolution: Integer; FingerID: string): Boolean;
begin
  Self.Resolution := Resolution;
  MainLog.Add('Loading image: %s', [Filename]);
  if not Image.LoadFromFile(Filename) then
  begin
    ViewLeft.Image := nil;
    ViewRight.Image := nil;
    ViewRight.Features := nil;
    MainLog.Add('Can not load image, further processing not performed');
    MainLog.AddEmpty;
    ErrorPrompt(Format('Can not load image from file ''%s''', [Filename]));
    Result := True;
  end
  else begin
    Self.FingerID := FingerID;
    OnImage;
    Result := True;
  end;
end;

procedure TMainForm.OnImage;
begin
  // if not processing file list then show the input image and clear output view
  if not FileListRunning then
  begin
    ViewLeft.Image := Image;
    ViewLeft.Resolution := Resolution;
    ViewRight.Image := nil;
    ViewRight.Features := nil;
    ListViewResults.Items.Clear;
  end;
  MainLog.Add('Size: %dx%d. Resolution: %d dpi', [Image.Width, Image.Height, Resolution]);
  MainLog.Append('. Extraction started...');
  ExtractionThread := TExtractionThread.Create(Resolution, Image, ImageAfter);
  ExtractionThread.OnTerminate := OnExtractionComplete;
  UpdateTask;
  ExtractionThread.Resume;
end;

function TMainForm.ErrorPrompt(Reason: string; Warning: Boolean = False): Boolean;
var
  ErrorTime: LongWord;
  DlgType: TMsgDlgType;
  DlgButtons: TMsgDlgButtons;
  R: Integer;
begin
  ErrorTime := GetTickCount;
  if Warning then DlgType := mtWarning
  else DlgType := mtError;
  if FileListRunning then
  begin
    Reason := Reason + #13'Press Cancel to stop file list';
    if Warning then DlgButtons := [mbYes, mbNo, mbCancel]
    else DlgButtons := [mbOk, mbCancel]
  end
  else
    if Warning then DlgButtons := [mbYes, mbNo]
    else DlgButtons := [mbOk];
  R := MessageDlg(Reason, DlgType, DlgButtons, 0);
  Inc(TaskUITime, GetTickCount() - ErrorTime);
  case R of
    IDOK, IDYES: Result := True;
    IDNO: Result := False;
  else
    StopTask;
    Result := False;
  end;
end;

function TMainForm.FileListRunning: Boolean;
begin
  Result := TaskThread is TFileListThread;
end;

procedure TMainForm.StopTask;
begin
  if MainTaskRunning then
  begin
    FStopTask := True;
    UpdateTask;
  end;
  StopIdentification;
end;

procedure TMainForm.OnExtractionComplete(Sender: TObject);
var
  Result: Integer;
  OK: Boolean;
  Resume: Boolean;
begin
  Result := ExtractionThread.Result;
  if VFSucceeded(Result) or (Result = VFE_LOW_QUALITY_IMAGE) then
  begin
    MainLog.Add('Extraction complete: %s. Time: %s', [VFResultToString(Result),
      TimeToString(ExtractionThread.Time)]);
    OK := (Result <> VFE_LOW_QUALITY_IMAGE) or Options.UseLowQualityImages;
  end
  else begin
    MainLog.Add('Extraction error: %s. Futher processing not performed',
      [VFErrorToString(Result)]);
    OK := False;
    ErrorPrompt(Format('Extraction error: %s', [VFErrorToString(Result)]));
  end;
  if FileListRunning then
  begin
    Inc(Count);
    Inc(TotalExtTime, ExtractionThread.Time);
    if OK then
    begin
      Inc(TotalSize, ExtractionThread.FeaturesSize);
    end
    else Inc(CountRej);
  end;
  Features.SetData(ExtractionThread.Features, ExtractionThread.FeaturesSize);
  // if processing file list then input image is not shown yet and results not cleared
  if FileListRunning then
  begin
    ViewLeft.Image := Image;
    ViewLeft.Resolution := Resolution;
    ListViewResults.Items.Clear;
  end;
  // show output
  ViewRight.Image := ImageAfter;
  ViewRight.Resolution := Resolution;
  ViewRight.Features := Features;
  if Mode = vfdmEnrollmentWithGen then
  begin
    if not OK then
    begin
      MainLog.AddEmpty;
      MainLog.Add('Generalization aborted');
      GenIndex := 0;
      if FileListRunning then FStopTask := True;
    end
    else begin
      GenResolution[GenIndex] := Resolution;
      GenImage[GenIndex].Assign(Image);
      GenImageAfter[GenIndex].Assign(ImageAfter);
      GenFeatures[GenIndex].Assign(Features);
      GenFingerID[GenIndex] := FingerID;
      Inc(GenIndex);
      if GenIndex = VF_GENERALIZE_COUNT then
      begin
        GenIndex := 0;
      end
      else OK := False;
    end;
  end
  else if Mode = vfdmVerification then
  begin
    if not OK then
    begin
      MainLog.AddEmpty;
      MainLog.Add('Verification aborted');
      VerIndex := 0;
      if FileListRunning then FStopTask := true;
    end
    else begin
      VerResolution[VerIndex] := Resolution;
      VerImage[VerIndex].Assign(Image);
      VerImageAfter[VerIndex].Assign(Image);
      VerFeatures[VerIndex].Assign(Features);
      VerFingerID[VerIndex] := FingerID;
      Inc(VerIndex);
      if VerIndex = 2 then
      begin
        VerIndex := 0;
      end
      else OK := False;
    end;
  end;
  Resume := True;
  if OK then
    case Mode of
      vfdmEnrollment: Enroll;
      vfdmEnrollmentWithGen: EnrollWithGen;
      vfdmVerification: Verify;
      vfdmIdentification: Resume := not StartIdentification;
    end;
  ImageFromScanner := False;
  ExtractionThread := nil;
  if Resume then
  begin
    UpdateTask;
    MainLog.AddEmpty;
    if FileListRunning then TaskThread.Resume;
  end;
end;

procedure TMainForm.UpdateTask;
var
  Running, CanStop, Stopping: Boolean;
  RT: Integer;
  F: Boolean;
  S, SS: string;
begin
  Running := TaskRunning;
  CanStop := CanStopTask;
  Stopping := StoppingTask;
  if Running then
  begin
    FileClear.Enabled := False;
    FileOpen.Enabled := False;
    FileOpenFileList.Enabled := False;
    ModeEnrollment.Enabled := False;
    ModeEnrollmentWithGen.Enabled := False;
    ModeVerification.Enabled := False;
    ModeIdentification.Enabled := False;
    ToolsClearDatabase.Enabled := False;
    ToolsClearMainLog.Enabled := False;
    ToolsClearScannerLog.Enabled := False;
    ToolsOptions.Enabled := False;
    HelpRegister.Enabled := False;
    HelpClearRegistration.Enabled := False;
  end
  else begin
    FileClear.Enabled := True;
    FileOpen.Enabled := Mode <> vfdmNone;
    FileOpenFileList.Enabled := (Mode = vfdmEnrollment) or (Mode = vfdmIdentification);
    ModeEnrollment.Enabled := True;
    ModeEnrollmentWithGen.Enabled := True;
    ModeVerification.Enabled := True;
    ModeIdentification.Enabled := True;
    ToolsClearDatabase.Enabled := True;
    ToolsClearMainLog.Enabled := True;
    ToolsClearScannerLog.Enabled := True;
    ToolsOptions.Enabled := True;
    RT := VFRegistrationType;
    F := (RT <> VF_RT_NOT_PROTECTED) and (RT <> VF_RT_HASP);
    HelpRegister.Enabled := F;
    HelpClearRegistration.Enabled := F;
  end;
  FileStop.Enabled := Running and CanStop and not Stopping;
  if Running then
  begin
    if Stopping then S := 'Stopping '
    else S := 'S&top ';
    S := S + GetTaskName;
  end
  else S := 'Stop';
  FileStop.Caption := S;

  S := GetTaskString;
  if Running then
  begin
    if Stopping then S := S + ' [stopping]';
  end;
  if Mode <> vfdmNone then SS := ' - ' + GetModeString else SS := EmptyStr;
  if Running then SS := SS + ' - ' + S;
  Caption := Application.Title + SS;
  LabelMain.Caption := S;
  ButtonMainStop.Enabled := CanStop and not Stopping;
  LabelMainTime.Visible := MainTaskRunning;

  ButtonResultsStop.Enabled := IdentificationRunning and not StoppingIdentification;
  LabelResultsTime.Visible := IdentificationRunning;

  UpdateImagesMenu;
end;

procedure TMainForm.AddVersionToLog(Log: TLog; Name: string; VersionHigh, VersionLow: LongWord);
begin
  Log.Add('%s    Version %u.%u.%u.%u', [Name, HiWord(VersionHigh), LoWord(VersionHigh),
    HiWord(VersionLow), LoWord(VersionLow)]);
end;

function TMainForm.FingerprintCountToString(Value: Integer;
  No: string): string;
begin
  Result := IntToString(Value, 'fingerprint', 'fingerprints', No);
end;

function TMainForm.MatchingSpeedToString(Count: Integer; Time: Double;
  Unknown: string): string;
begin
  if Time <> 0 then Result := FingerprintCountToString(Round(Count / Time)) + ' per second'
  else Result := Unknown;
end;

function TMainForm.PercentToString(Count, Total: Integer;
  Unknown: string): string;
begin
  if Total <> 0 then Result := Format('%d%%', [Percent(Count, Total)])
  else Result := Unknown;
end;

procedure TMainForm.Enroll;
var
  Flags: LongWord;
  Thread: TIdentificationThread;
  Index: Integer;
  DupFingerId: string;
  Feat: array[0..VF_MAX_FEATURES_SIZE] of Byte;
  Size: LongWord;
begin
  if Options.CheckMinMinutiaCount and (Features.Minutiae.Count < Options.MinMinutiaCount) then
  begin
    MainLog.Add('Minutia count (%d) is less than minimal accepted value (%d)', [Features.Minutiae.Count, Options.MinMinutiaCount]);
    ErrorPrompt(Format('Minutia count (%d) is less than minimal accepted value (%d)',
      [Features.Minutiae.Count, Options.MinMinutiaCount]));
    Exit;
  end;
  if ImageFromScanner then
  begin
    if Options.ScannerPrompt then
    begin
      Flags := FLEB_ENROLL or FLEB_FROM_SCANNER;
      if not FileListEntryBox(Self, Flags, ScannerId, Resolution, Options.ScannerFingerId) then Exit;
    end;
    FingerId := MakeFingerID(Options.ScannerFingerId, ScannerId, True);
  end;
  Size := Features.GetData(@Feat[0]);
  if Options.SearchForDuplicates then
  begin
    Thread := TIdentificationThread.Create(@Feat[0], Size, Database, True, 100, False);
    Thread.FreeOnTerminate := False;
    WaitBox(Self, Thread, 'Searching for duplicates');
    Index := Thread.Index;
    Thread.Free;
    if Index <> -1 then
    begin
      DupFingerId := Database.FingerID[Index];
      if not Options.WarnDifferentFingerId or (FingerId <> DupFingerId) then
      begin
        MainLog.Add('Duplicate found. Finger ID: ''%s''. Record ID: %d', [DupFingerId, Database.ID[Index]]);
        if not ErrorPrompt(Format('Duplicate found. Finger ID: ''%s''. Record ID: %d'#13'Add to the database?',
          [DupFingerId, Database.ID[Index]]), True) then Exit;
      end;
    end;
  end;
  MainLog.Add('Adding to the database. Features size: %s. Finger ID: ''%s''',
    [SizeToString(Size), FingerId]);
  try
    Database.Add(@Feat[0], Size, FingerId);
    UpdateStatus;
  except
    on E: Exception do
    begin
      MainLog.Add('Can not add to the database');
      ErrorPrompt(Format('Can not add to the database'#13'%s', [E.Message]));
    end;
  end;
end;

procedure TMainForm.EnrollWithGen;
var
  GenFeat: array[0..VF_GENERALIZE_COUNT - 1] of PByte;
  I: Integer;
  Feat: array[0..VF_MAX_FEATURES_SIZE - 1] of Byte;
  Size: LongWord;
  TC: Double;
  Result: Integer;
  S: string;
begin
  MainLog.AddEmpty;
  for I := 0 to VF_GENERALIZE_COUNT - 1 do
  begin
    GetMem(GenFeat[I], VF_MAX_FEATURES_SIZE);
    GenFeatures[I].GetData(GenFeat[I]);
  end;
  TC := GetTickCountEx;
  Result := VFGeneralize(VF_GENERALIZE_COUNT, @GenFeat[0], @Feat[0], Size, nil);
  TC := GetTickCountEx - TC;
  for I := 0 to VF_GENERALIZE_COUNT - 1 do
  begin
    FreeMem(GenFeat[I]);
  end;
  if VFSucceeded(Result) or (Result = VFE_FAILED) then
  begin
    if Result = VFE_FAILED then S := 'failed'
    else S := Format('succeeded (based on image %d)', [Result + 1]);
    MainLog.Add('Generalization %s. Time: %s', [S, TimeToString(TC)]);
    if Result <> VFE_FAILED then
    begin
      Resolution := GenResolution[Result];
      Image.Assign(GenImage[Result]);
      ImageAfter.Assign(GenImageAfter[Result]);
      Features.SetData(@Feat[0], Size);
      FingerId := GenFingerId[Result];
      ViewLeft.Resolution := Resolution;
      ViewLeft.UpdateImage;
      ViewRight.Resolution := Resolution;
      ViewRight.UpdateImage;
      ViewRight.UpdateFeatures;
      Enroll;
    end;
  end
  else begin
    S := Format('Generalization error: %s', [VFErrorToString(Result)]);
    MainLog.Add(S);
    ErrorPrompt(S);
  end;
end;

procedure TMainForm.Verify;
var
  MD: TVFMatchDetails;
  Result: Integer;
  S: string;
  TC: Double;
  Feat1: array[0..VF_MAX_FEATURES_SIZE] of Byte;
  Feat2: array[0..VF_MAX_FEATURES_SIZE] of Byte;
begin
  MainLog.AddEmpty;
  MD.Size := SizeOf(MD);
  VerFeatures[0].GetData(@feat1[0]);
  VerFeatures[1].GetData(@feat2[0]);
  TC := GetTickCountEx;
  Result := VFVerify(@feat1[0], @feat2[0], @MD, nil);
  TC := GetTickCountEx - TC;
  if VFSucceeded(Result) or (Result = VFE_FAILED) then
  begin
    if Result = VFE_FAILED then S := 'failed' else S := 'succeeded';
    MainLog.Add('Verification %s. Similarity: %d. Time: %s',
      [S, MD.Similarity, TimeToString(TC)]);
  end
  else begin
    S := Format('Verification error: %s', [VFErrorToString(result)]);
    MainLog.Add(S);
    ErrorPrompt(S);
  end;
end;

function TMainForm.StartIdentification: Boolean;
var
  Flags: LongWord;
  Feat: array[0..VF_MAX_FEATURES_SIZE] of Byte;
  Size: LongWord;
begin
  if Database.Count = 0 then
  begin
    MainLog.Add('Database is empty. Do not starting identification');
    ErrorPrompt('Database is empty');
    Result := False;
    Exit;
  end;
  if ImageFromScanner then
  begin
    if Options.ScannerPrompt then
    begin
      Flags := FLEB_ENROLL or FLEB_FROM_SCANNER;
      if not FileListEntryBox(Self, Flags, ScannerId, Resolution, Options.ScannerFingerId) then
      begin
        Result := False;
        Exit;
      end;
    end;
    FingerId := MakeFingerID(Options.ScannerFingerId, ScannerId, True);
  end;
  MainLog.Add('Finger ID: ''%s''', [FingerId]);
  if Options.UseG then
  begin
    MainLog.Append('. G: %d', [Features.G]);
  end;
  MainLog.Append('. Identification started...');
  CountIden := 0;
  FingerFound := False;
  Size := Features.GetData(@Feat[0]);
  IdentificationThread := TIdentificationThread.Create(@Feat[0], Size, Database, Options.UseG, Options.DatabasePercent, Options.MatchAll);
  IdentificationThread.OnProgress := OnIdentificationProgress;
  IdentificationThread.OnTerminate := OnIdentificationComplete;
  UpdateIdentificationProgress(FALSE, 0);
  UpdateTask;
  IdentificationTimeStarted := GetTickCountEx;
  IdentificationThread.Resume;
  Result := True;
end;

procedure TMainForm.StopIdentification;
begin
  if IdentificationRunning and not StoppingIdentification then
  begin
    FStopIdentification := True;
    UpdateTask;
  end;
end;

function TMainForm.CanStopTask: Boolean;
begin
  Result := (TaskThread <> nil) or (IdentificationThread <> nil);
end;

procedure TMainForm.CreateDefaultDatabase(Filename: string);
var
  Res: HRSRC;
  Size: LongWord;
  ResData: HGLOBAL;
  Data: Pointer;
  Stream: TStream;
begin
  Res := FindResource(0, PChar(IDR_DATABASE), 'DATABASE');
  if Res = 0 then raise Exception.Create('Resource not found');
  Size := SizeofResource(0, Res);
  ResData := LoadResource(0, Res);
  if ResData = 0 then raise Exception.Create('Can not load resource');
  Data := LockResource(ResData);
  if Data = nil then raise Exception.Create('Can not lock resource');
  Stream := TFileStream.Create(Filename, fmCreate or fmShareDenyWrite);
  try
    Stream.WriteBuffer(Data^, Size);
  finally
    Stream.Free;
  end;
end;

function TMainForm.DBOpenRunning: Boolean;
begin
  Result := TaskThread is TDBOpenThread;
end;

function TMainForm.GetModeString: string;
begin
  Result := VFDemoModeToString(Mode);
end;

function TMainForm.GetTaskName: string;
begin
  if TaskThread <> nil then Result := TaskThread.TaskName
  else if ExtractionThread <> nil then Result := 'Extraction'
  else if IdentificationThread <> nil then Result := 'Identification'
  else Result := EmptyStr;
end;

function TMainForm.GetTaskString: string;
begin
  if TaskThread <> nil then Result := TaskThread.TaskString
  else if ExtractionThread <> nil then Result := 'Extracting'
  else if IdentificationThread <> nil then Result := 'Identificating'
  else if Mode = vfdmEnrollmentWithGen then if GenIndex > 0 then Result := Format('Waiting for image %d of %d', [GenIndex + 1, VF_GENERALIZE_COUNT]) else Result := Format('Waiting for %d images', [VF_GENERALIZE_COUNT])
  else if Mode = vfdmVerification then if VerIndex > 0 then Result := Format('Waiting for image %d of %d', [VerIndex + 1, 2]) else Result := Format('Waiting for %d images', [2])
  else Result := 'Waiting for image(s)';
end;

function TMainForm.IdentificationRunning: Boolean;
begin
  Result := IdentificationThread <> nil;
end;

function TMainForm.MainTaskRunning: Boolean;
begin
  Result := TaskThread <> nil;
end;

procedure TMainForm.OnDBOpenComplete(Time: LongWord);
var
  DBOpenThread: TDBOpenThread;
  S: string;
  C: Integer;
begin
  DBOpenThread := TaskThread as TDBOpenThread;
  if DBOpenThread.Failed then
  begin
    MainLog.Append('. Database not opened');
    MessageDlg(DBOpenThread.Error, mtError, [mbOk], 0);
  end
  else begin
    C := Database.Count;
    MainLog.Append(': %s loaded', [IntToString(C, 'record', 'records', 'no')]);
    if C <> 0 then
    begin
      if Time <> 0 then S := Format(' (%u records per second)', [LongWord(C) * 1000 div Time]) else S := EmptyStr;
      MainLog.Add('Time: %s%s', [TimeToString(Time), S]);
    end;
  end;
  UpdateStatus;
end;

procedure TMainForm.OnFileListComplete(Time: LongWord);
var
  CountLeft: Integer;
  P: Double;
begin
  MainLog.Append(': %s proceeded', [IntToString(Count, 'file', 'files', 'no')]);
  if Count <> 0 then
  begin
    if (Mode = vfdmEnrollment) or (Mode = vfdmIdentification) then
    begin
      MainLog.Add('Average time per file: %s (%s total)', [TimeToString(Time div LongWord(Count)),
        TimeToString(Time)]);
      MainLog.Add('Average extraction time: %s (%s total)', [TimeToString(TotalExtTime div LongWord(Count)),
        TimeToString(TotalExtTime)]);
      CountLeft := Count - CountRej;
      MainLog.Add('Images rejected: %d (%d%%)', [CountRej, Percent(CountRej, Count)]);
      if CountLeft <> 0 then
      begin
        MainLog.Append('. For %d non-rejected:', [CountLeft]);
        if Mode = vfdmEnrollment then
        begin
          MainLog.Add('Average features size: %s (%s total)',
            [SizeToString(TotalSize div LongWord(CountLeft)),
            SizeToString(TotalSize)]);
        end
        else begin
          MainLog.Add('Average matching time: %s (%s total)',
            [TimeToString(TotalIdenTime / CountLeft),
            TimeToString(TotalIdenTime)]);
          MainLog.Add('Average matching speed: %s',
            [MatchingSpeedToString(TotalIden, TotalIdenTime)]);
          if Options.UseG then
          begin
            MainLog.Add('Average effective matching speed: %s',
              [MatchingSpeedToString(CountLeft * Database.Count, TotalIdenTime)]);
            MainLog.Add('Average fingerprints matched: %d', [TotalIden div CountLeft]);
            if Database.Count <> 0 then
            begin
              P := PercentF(TotalIden, CountLeft * Database.Count);
              if Options.MatchAll and (P = 100.0) then MainLog.Append(' (all database)')
              else MainLog.Append(' (%d%% of database size)', [Round(P)]);
            end;
          end;
          MainLog.Add('FAR: %s (%d of %d)', [PercentToString(FalseAccepted, CountLeft - FingersFound),
            FalseAccepted, CountLeft - FingersFound]);
          MainLog.Add('FRR: %s (%d of %d)', [PercentToString(FalseRejected, FingersFound),
            FalseRejected, FingersFound]);
        end;
      end;
    end
    else MainLog.Add('Time: %s', [TimeToString(Time)]);
  end;
end;

procedure TMainForm.OnFileListImage(FileIndex: Integer);
var
  Item: PVFFileListItem;
begin
  TaskThread.Suspend;
  if FileIndex = 0 then MainLog.AddEmpty;
  MainLog.Add('File %d of %d', [FileIndex + 1, FileList.Count]);
  Item := FileList[FileIndex];
  if OnFileImage(Item.Filename, Item.Resolution, Item.FingerId) then
  begin
    // if this is the first file in the file list then show input image and clear output view
    if FileIndex = 0 then
    begin
      ViewLeft.Image := Image;
      ViewLeft.Resolution := Resolution;
      ViewRight.Image := nil;
      ViewRight.Features := nil;
      ListViewResults.Items.Clear;
    end;
  end
  else TaskThread.Resume;
end;

procedure TMainForm.OnIdentificationComplete(Sender: TObject);
var
  Time: Double;
  Result: Integer;
  S: string;
  P: Double;
  Index: Integer;
  Similarity: Integer;
  DatabaseFingerId: string;
begin
  Time := GetTickCountEx - IdentificationTimeStarted;
  Result := IdentificationThread.Result;
  if VFSucceeded(Result) or (Result = VFE_FAILED) then
  begin
    if FStopIdentification then S := 'stopped' else S := 'complete';
    MainLog.Add('Identification %s: %s matched', [S, FingerprintCountToString(CountIden, 'no')]);
    if Options.UseG and (Database.Count <> 0) then
    begin
      P := PercentF(CountIden, Database.Count);
      if Options.MatchAll and (P = 100.0) then MainLog.Append(' (all database)')
      else MainLog.Append(' (%d%% of database size)', [Round(P)]);
    end;
    MainLog.Append('. Time: %s', [TimeToString(Time)]);
    if FingerFound and FileListRunning then Inc(FingersFound);
    Index := IdentificationThread.Index;
    Similarity := IdentificationThread.Similarity;
    if Index = -1 then
    begin
      MainLog.Add('Not identified');
      if FingerFound and (FingerId <> EmptyStr) then
      begin
        MainLog.Append('. ID found in the database. False Rejection?');
        if FileListRunning then Inc(FalseRejected);
      end;
    end
    else begin
      MainLog.Add('Identified finger ID: ''%s''', [Database.FingerID[Index]]);
      if Options.UseG then
      begin
        MainLog.Append('. G: %d', [VFFeatGetG(Database.Features[Index])]);
      end;
      MainLog.Append('. Similarity: %d', [Similarity]);
      DatabaseFingerId := Database.FingerID[Index];
      if (FingerId <> EmptyStr) and (DatabaseFingerId <> EmptyStr) and (FingerId <> DatabaseFingerId) then
      begin
        MainLog.Append('. IDs do not match. False Acceptance?');
        if FileListRunning then Inc(FalseAccepted);
      end;
    end;
    MainLog.Add('Matching time: %s', [TimeToString(IdentificationThread.Time)]);
    if CountIden <> 0 then
    begin
      MainLog.Add('Matching speed: %s', [MatchingSpeedToString(CountIden, IdentificationThread.Time)]);
      if Options.UseG then
      begin
        MainLog.Add('Effective matching speed: %s', [MatchingSpeedToString(Database.Count, IdentificationThread.Time)]);
      end;
    end;
  end
  else begin
    MainLog.Add('Identification error: %s', [VFErrorToString(Result)]);
    ErrorPrompt(Format('Identification error: %s', [VFErrorToString(Result)]));
  end;
  if FileListRunning then
  begin
    Inc(TotalIden, CountIden);
    TotalIdenTime := TotalIdenTime + IdentificationThread.Time;
  end;
  MainLog.AddEmpty;
  IdentificationThread := nil;
  FStopIdentification := False;
  UpdateIdentificationProgress(False, 0);
  UpdateTask;
  if FileListRunning then TaskThread.Resume;
end;

function TMainForm.OnIdentificationProgress(Sender: TObject;
  Started: Boolean; Progress: PIdenProgress): Boolean;
var
  I: Integer;
  LI: TListItem;
begin
  if FStopIdentification then
  begin
    Result := False;
    Exit;
  end;
  if Started then UpdateIdentificationProgress(Started, Progress.Count)
  else UpdateIdentificationProgress(Started, Integer(Progress));
  if Started then
  begin
    Inc(CountIden, Progress.Count);
    ListViewResults.Items.BeginUpdate;
    for I := 0 to Progress.Count - 1 do
    begin
      if Options.ShowAllResults or VFSucceeded(Progress.Result[I]) then
      begin
        LI := ListViewResults.Items.Add;
        LI.Caption := IntToStr(VFFeatGetG(Database.Features[Progress.Index[I]]));
        LI.Subitems.Add(VFResultToString(Progress.Result[I]));
        LI.Subitems.Add(IntToStr(Progress.Similarity[I]));
        LI.Subitems.Add(IntToStr(Database.ID[Progress.Index[I]]));
        LI.Subitems.Add(Database.FingerID[Progress.Index[I]]);
      end;
      if not FingerFound then
      begin
        if FingerId = Database.FingerID[Progress.Index[I]] then
        begin
          FingerFound := True;
        end;
      end;
    end;
    ListViewResults.Items.EndUpdate;
  end;
  Result := True;
end;

procedure TMainForm.OnTaskComplete(Sender: TObject);
var
  Time: LongWord;
  S: string;
begin
  Time := GetTickCount - TaskTimeStarted - TaskUITime;
  if TaskThread.Failed then S := ' failed'
  else if FStopTask then S := ' stopped' else S := ' complete';
  MainLog.Add(TaskThread.TaskName + S);
  if DBOpenRunning then OnDBOpenComplete(Time)
  else if FileListRunning then OnFileListComplete(Time);
  MainLog.AddEmpty;
  TaskThread := nil;
  FStopTask := False;
  UpdateTaskProgress(False, 0);
  UpdateTask;
  if ChangeBackMode then
  begin
    SetMode(LastMode);
    ChangeBackMode := False;
  end;
end;

function TMainForm.OnTaskProgress(Sender: TObject; Started: Boolean;
  Value: Integer): Boolean;
begin
  if FStopTask then Result := False
  else begin
    UpdateTaskProgress(Started, Value);
    if DBOpenRunning and Started then UpdateStatus
    else if FileListRunning and Started then OnFileListImage(Value);
    Result := True;
  end;
end;

procedure TMainForm.OnTaskState(Sender: TObject; State: string);
begin
  UpdateTaskProgress(False, 0, State);
end;

procedure TMainForm.SaveImage(Image: TVFImage);
begin
  if Image.IsEmpty then Exit;
  SaveDialog.FileName := EmptyStr;
  if SaveDialog.Execute then
  begin
    if not Image.SaveToFile(SaveDialog.FileName, SaveDialog.FilterIndex - 1) then
      MessageDlg(Format('Can not save image to file ''%s''', [SaveDialog.FileName]), mtError, [mbOk], 0);
  end;
end;

procedure TMainForm.SetMode(Value: TVFDemoMode);
begin
  if not TaskRunning and (Mode <> Value) then
  begin
    Mode := Value;
    MainLog.Add('Switching to mode: %s', [GetModeString]);
    MainLog.AddEmpty;
    Clear;
    UpdateMode;
  end;
end;

procedure TMainForm.StartDBOpen;
var
  DBName: string;
begin
  DBName := GetFullFileName('VFDemo.mdb');
  if not FileExists(DBName) then
  begin
    MainLog.Add('Database file not found. Creating default database');
    try
      CreateDefaultDatabase(DBName);
    except
      on E: Exception do
      begin
        MainLog.Add('Can not create default database');
        MessageDlg(E.Message, mtError, [mbOk], 0);
      end;
    end;
    MainLog.AddEmpty;
  end;
  StartTask(TDBOpenThread.Create(Database, 'Microsoft.Jet.OLEDB.4.0', DBName));
end;

procedure TMainForm.StartFileList;
var
  Item: PVFFileListItem;
begin
  if FileList.Count = 1 then
  begin
    Item := FileList[0];
    OnFileImage(Item.Filename, Item.Resolution, Item.FingerId);
    Exit;
  end;
  MainLog.Add('%d files selected', [FileList.Count]);
  Count := 0;
  CountRej := 0;
  TotalSize := 0;
  TotalExtTime := 0;
  TotalIden := 0;
  TotalIdenTime := 0;
  FingersFound := 0;
  FalseAccepted := 0;
  FalseRejected := 0;
  StartTask(TFileListThread.Create(FileList));
end;

procedure TMainForm.StartTask(Thread: TTaskThread);
begin
  TaskThread := Thread;
  UpdateTaskProgress(False, 0);
  UpdateTask;
  if MainLog.LastLine = EmptyStr then MainLog.AddEmpty
  else MainLog.Append('. ');
  MainLog.Append(TaskThread.TaskName + ' started...');
  TaskUITime := 0;
  TaskTimeStarted := GetTickCount;
  TaskThread.OnTerminate := OnTaskComplete;
  TaskThread.OnState := OnTaskState;
  TaskThread.OnProgress := OnTaskProgress;
  TaskThread.Resume;
end;

function TMainForm.StoppingIdentification: Boolean;
begin
  Result := FStopIdentification;
end;

function TMainForm.StoppingTask: Boolean;
begin
  Result := ((TaskThread <> nil) and FStopTask) or ((TaskThread = nil) and (IdentificationThread <> nil) and FStopIdentification);
end;

function TMainForm.StopTaskPrompt: Boolean;
begin
  if TaskRunning and CanStopTask and not StoppingTask then
  begin
    if MessageDlg(GetTaskString + ' is in progress.'#13'Do you want to stop it?',
      mtConfirmation, [mbYes, mbNo], 0) = mrYes then StopTask;
    Result := False;
    Exit;
  end;
  Result := not TaskRunning;
end;

var
  UIPProgress, UIPTotal: Integer;
  TimeStarted: Double;

procedure TMainForm.UpdateIdentificationProgress(Started: Boolean;
  Value: Integer);
var
  TimeElapsed, TimeEstimated: Double;
  S: string;
begin
  if Started then
  begin
    UIPProgress := UIPProgress + Value;
    ProgressBarResults.Position := UIPProgress;
    TimeElapsed := GetTickCountEx - TimeStarted;
    if (UIPProgress <> 0) and (UIPTotal <> 0) then
      TimeEstimated := TimeElapsed / UIPProgress * UIPTotal
    else TimeEstimated := 0;
    if UIPTotal <> 0 then
    begin
      if UIPProgress <> 0 then S := Format('; remaining: %s', [TimeToString(TimeEstimated - TimeElapsed)])
      else S := EmptyStr;
      LabelResultsTime.Caption := Format('Time elapsed: %s', [TimeToString(TimeElapsed)]) + S;
    end
    else LabelResultsTime.Caption := EmptyStr;
  end
  else begin
    UIPTotal := Value;
    UIPProgress := 0;
    ProgressBarResults.Max := UIPTotal;
    ProgressBarResults.Visible := UIPTotal <> 0;
    LabelResultsTime.Caption := EmptyStr;
    TimeStarted := GetTickCountEx;
  end;
end;

procedure TMainForm.UpdateImagesMenu;
begin
  FileSave.Enabled := not TaskRunning and not Image.IsEmpty;
  FileSaveRight.Enabled := not TaskRunning and not ImageAfter.IsEmpty;
end;

procedure TMainForm.UpdateMode;
begin
  ModeEnrollment.Checked := Mode = vfdmEnrollment;
  ModeEnrollmentWithGen.Checked := Mode = vfdmEnrollmentWithGen;
  ModeIdentification.Checked := Mode = vfdmIdentification;
  ModeVerification.Checked := Mode = vfdmVerification;
  UpdateTask;
end;

var
  UTPProgress, UTPTotal: Integer;
  UTPTimeStarted: LongWord;

procedure TMainForm.UpdateTaskProgress(Started: Boolean; Value: Integer;
  Text: string);
var
  TimeElapsed, TimeEstimated: LongWord;
  S: string;
begin
  if Started then
  begin
    UTPProgress := Value;
    ProgressBarMain.Position := UTPProgress;
    TimeElapsed := GetTickCount - UTPTimeStarted - TaskUITime;
    if (UTPProgress <> 0) and (UTPTotal <> 0) then TimeEstimated := Round(TimeElapsed / UTPProgress * UTPTotal)
    else TimeEstimated := 0;
    if UTPTotal <> 0 then
    begin
      if UTPProgress <> 0 then S := Format('; remaining: %s', [TimeToString(TimeEstimated - TimeElapsed)]) else S := EmptyStr;
      LabelMainTime.Caption := Format('Time elapsed: %s', [TimeToString(TimeElapsed)]) + S;
    end
    else LabelResultsTime.Caption := EmptyStr;
  end
  else begin
    UTPTotal := Value;
    UTPProgress := 0;
    ProgressBarMain.Max := UTPTotal;
    ProgressBarMain.Visible := UTPTotal <> 0;
    LabelMainTime.Caption := Text;
    UTPTimeStarted := GetTickCount;
  end;
end;

procedure TMainForm.ToolsFileListEditorExecute(Sender: TObject);
begin
  FileListBox(Self);
end;

procedure TMainForm.FileStopExecute(Sender: TObject);
begin
  StopTask;
end;

procedure TMainForm.ButtonMainStopClick(Sender: TObject);
begin
  StopTask;
end;

procedure TMainForm.ButtonResultsStopClick(Sender: TObject);
begin
  StopIdentification;
end;

procedure TMainForm.UpdateViewMenu;
begin
  ViewZoomIn.Enabled := ViewRight.Zoom <> TVFView.MaxZoom;
  ViewZoomOut.Enabled := ViewRight.Zoom <> TVFView.MinZoom;
  ViewShowFeatures.Checked := ViewRight.ShowFeatures;
  ViewShowSingularPoints.Checked := ViewRight.ShowSingularPoints;
  ViewShowBlockedOrientations.Checked := ViewRight.ShowBlockedOrientations;
end;

procedure TMainForm.FileExitExecute(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.FileOpenFileListExecute(Sender: TObject);
begin
  if not TaskRunning and ((Mode = vfdmEnrollment) or (Mode = vfdmIdentification)) then
  begin
    OpenListDialog.FileName := EmptyStr;
    if OpenListDialog.Execute then
    begin
      if FileList.Load(OpenListDialog.FileName) then
      begin
        if FileList.Count = 0 then MessageDlg('File list is empty', mtError, [mbOk], 0)
        else StartFileList;
      end
      else MessageDlg(Format('Can not open file list ''%s''', [OpenListDialog.FileName]), mtError, [mbOk], 0);
    end;
  end;
end;

procedure TMainForm.FileSaveExecute(Sender: TObject);
begin
  if not TaskRunning then
  begin
    SaveImage(Image);
  end;
end;

procedure TMainForm.FileSaveRightExecute(Sender: TObject);
begin
  if not TaskRunning then
  begin
    SaveImage(ImageAfter);
  end;
end;

procedure TMainForm.ModeEnrollmentExecute(Sender: TObject);
begin
  SetMode(vfdmEnrollment);
end;

procedure TMainForm.ModeEnrollmentWithGenExecute(Sender: TObject);
begin
  SetMode(vfdmEnrollmentWithGen);
end;

procedure TMainForm.ModeVerificationExecute(Sender: TObject);
begin
  SetMode(vfdmVerification);
end;

procedure TMainForm.ModeIdentificationExecute(Sender: TObject);
begin
  SetMode(vfdmIdentification);
end;

procedure TMainForm.UpdateStatus;
var
  S: string;
begin
  if Database.Opened then S := EmptyStr else S := ' (Offline)';
  StatusBar.Panels[0].Text := 'Records in the database: ' + IntToStr(Database.Count) + S;
end;

procedure TMainForm.ViewZoomInExecute(Sender: TObject);
begin
  ViewLeft.ZoomIn;
  ViewRight.ZoomIn;
  UpdateViewMenu;
end;

procedure TMainForm.ViewZoomOutExecute(Sender: TObject);
begin
  ViewLeft.ZoomOut;
  ViewRight.ZoomOut;
  UpdateViewMenu;
end;

procedure TMainForm.ViewZoom1To1Execute(Sender: TObject);
begin
  ViewLeft.Zoom := 100;
  ViewRight.Zoom := 100;
end;

procedure TMainForm.ViewShowFeaturesExecute(Sender: TObject);
begin
  ViewRight.ShowFeatures := not ViewRight.ShowFeatures;
  UpdateViewMenu;
end;

procedure TMainForm.ViewShowSingularPointsExecute(Sender: TObject);
begin
  ViewRight.ShowSingularPoints := not ViewRight.ShowSingularPoints;
  UpdateViewMenu;
end;

procedure TMainForm.ViewShowBlockedOrientationsExecute(Sender: TObject);
begin
  ViewRight.ShowBlockedOrientations := not ViewRight.ShowBlockedOrientations;
  UpdateViewMenu;
end;

procedure TMainForm.ToolsClearDatabaseExecute(Sender: TObject);
var
  Thread: TDBClearThread;
  TC: LongWord;
begin
  if not TaskRunning then
  begin
    if MessageDlg('Do you really want to clear database?', mtConfirmation, [mbYes, mbNo], 0) = mrYes then
    begin
      Thread := TDBClearThread.Create(Database);
      Thread.FreeOnTerminate := False;
      TC := GetTickCount;
      WaitBox(Self, Thread, Thread.TaskString);
      TC := GetTickCount - TC;
      if Thread.Failed then
      begin
        MainLog.Add(Thread.TaskName + ' failed');
        MessageDlg(Thread.Error, mtError, [mbOk], 0);
      end
      else MainLog.Add('Database cleared. Time: %s', [TimeToString(TC)]);
      Thread.Free;
      MainLog.AddEmpty;
      UpdateStatus;
    end;
  end;
end;

procedure TMainForm.ToolsClearMainLogExecute(Sender: TObject);
begin
  if not TaskRunning then
  begin
    MainLog.Clear;
  end;
end;

procedure TMainForm.ToolsClearScannerLogExecute(Sender: TObject);
begin
  if not TaskRunning then
  begin
    ScannerLog.Clear;
  end;
end;

procedure TMainForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  CanClose := StopTaskPrompt;
end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
var
  Thread: TDBCloseThread;
  TC: LongWord;
begin
  Thread := TDBCloseThread.Create(Database);
  TC := GetTickCount;
  WaitBox(Self, Thread, Thread.TaskString);
  TC := GetTickCount - TC;
  if Thread.Failed then
  begin
     MainLog.Add(Thread.TaskName + ' failed');
     MessageDlg(Thread.Error, mtError, [mbOk], 0);
  end
  else MainLog.Add('Database closed. Time: %s', [TimeToString(TC)]);
  MainLog.AddEmpty;
end;

procedure TMainForm.ToolsOptionsExecute(Sender: TObject);
begin
  if not TaskRunning then
  begin
    if OptionsBox(Self, Options) then
    begin
      UpdateOptions(Options);
    end;
  end;
end;

procedure TMainForm.HelpAboutExecute(Sender: TObject);
begin
  AboutBox(Self);
end;

procedure TMainForm.OnScannerImage(ScannerID: string; Width,
  Height: Integer; Image: PByte; Resolution: Integer);
begin
  if not TaskRunning and (Mode <> vfdmNone) then
  begin
    Self.ScannerID := ScannerID;
    Self.Resolution := Resolution;
    FingerID := EmptyStr;
    Self.Image.Assign(Width, Height, Image);
    ImageFromScanner := True;
    MainLog.Add('Image from scanner: %s', [Self.ScannerID]);
    OnImage;
  end;
end;

procedure TMainForm.OnScannerState(ScannerID: string; State: LongWord);
begin
  ScannerLog.Add(SMStateToString(State, ScannerID));
end;

procedure TMainForm.OnSMMonitor(ScannerID: string; Event: LongWord);
begin
  if ScannerLog.LastLine <> EmptyStr then ScannerLog.AddEmpty;
  ScannerLog.Add(SMEventToString(Event, ScannerID));
  ScannerLog.AddEmpty;
end;

end.

