unit VFControls;

interface

uses Windows, Messages, Classes, Controls, Graphics, Math, VFinger, VFImage, VFFeatures;

type
  TVFView = class(TCustomControl)
  private
    FShowSingularPoints: Boolean;
    FShowBlockedOrientations: Boolean;
    FZoom: Integer;
    FResolution: Integer;
    FColorMax: TColor;
    FMinutiaColor: TColor;
    FSPColor: TColor;
    FColorMin: TColor;
    FBOColor: TColor;
    FFeatures: TVFFeatures;
    FImage: TVFImage;
    FBitmap: HBITMAP;
    FBitmapWidth, FBitmapHeight: Integer;
    FShowFeatures: Boolean;

    procedure FreeBitmap;
    procedure GetImageRect(var X, Y, W, H: Integer);
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure PaintBO(X, Y: Integer; Dir: Byte);
    procedure PaintDir(X, Y: Integer; Dir: Byte; Length: Integer);

    procedure SetBOColor(const Value: TColor);
    procedure SetColorMax(const Value: TColor);
    procedure SetColorMin(const Value: TColor);
    procedure SetFeatures(const Value: TVFFeatures);
    procedure SetImage(const Value: TVFImage);
    procedure SetMinutiaColor(const Value: TColor);
    procedure SetResolution(Value: Integer);
    procedure SetShowBlockedOrientations(const Value: Boolean);
    procedure SetShowSingularPoints(const Value: Boolean);
    procedure SetSPColor(const Value: TColor);
    procedure SetZoom(Value: Integer);
    procedure SetShowFeatures(const Value: Boolean);

  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure Paint; override;

  public
    constructor Create(AOwner: TComponent); override;

    procedure UpdateImage;
    procedure UpdateFeatures;

    procedure ZoomIn;
    procedure ZoomOut;

    property Image: TVFImage read FImage write SetImage;
    property ColorMin: TColor read FColorMin write SetColorMin;
    property ColorMax: TColor read FColorMax write SetColorMax;
    property Resolution: Integer read FResolution write SetResolution;
    property Features: TVFFeatures read FFeatures write SetFeatures;
    property ShowFeatures: Boolean read FShowFeatures write SetShowFeatures;
    property MinutiaColor: TColor read FMinutiaColor write SetMinutiaColor;
    property ShowSingularPoints: Boolean read FShowSingularPoints write SetShowSingularPoints;
    property SPColor: TColor read FSPColor write SetSPColor;
    property ShowBlockedOrientations: Boolean read FShowBlockedOrientations write SetShowBlockedOrientations;
    property BOColor: TColor read FBOColor write SetBOColor;
    property Zoom: Integer read FZoom write SetZoom;
    class function MaxZoom: Integer;
    class function MinZoom: Integer;

  end;

implementation

{ TVFView }

constructor TVFView.Create(AOwner: TComponent);
begin
  inherited;
  FImage := nil;
  FBitmap := 0;
  FBitmapWidth := 0;
  FBitmapHeight := 0;
  FColorMin := clBlack;
  FColorMax := clWhite;
  FResolution := VF_IMAGE_RESOLUTION;
  FFeatures := nil;
  FShowFeatures := True;
  FMinutiaColor := clRed;
  FShowSingularPoints := True;
  FSPColor := RGB(255, 127, 0);
  FShowBlockedOrientations := False;
  FBOColor := RGB(0, 0, 255);
  FZoom := 100;

  Width := 300;
  Height := 200;
end;

procedure TVFView.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.ExStyle := Params.ExStyle or WS_EX_CLIENTEDGE;
end;

procedure TVFView.FreeBitmap;
begin
  if FBitmap <> 0 then
  begin
    DeleteObject(FBitmap);
    FBitmap := 0;
    FBitmapWidth := 0;
    FBitmapHeight := 0;
  end;
end;

procedure TVFView.GetImageRect(var X, Y, W, H: Integer);
begin
  if FBitmap <> 0 then
  begin
    W := FBitmapWidth * VF_IMAGE_RESOLUTION div FResolution * FZoom div 100;
    H := FBitmapHeight * VF_IMAGE_RESOLUTION div FResolution * FZoom div 100;
  end
  else begin
    W := 0;
    H := 0;
  end;
  X := (ClientWidth - W) div 2;
  Y := (ClientHeight - H) div 2;
end;

class function TVFView.MaxZoom: Integer;
begin
  Result := 800;
end;

class function TVFView.MinZoom: Integer;
begin
  Result := 25;
end;

procedure TVFView.Paint;
var
  X, Y, W, H: Integer;
  BmpDC: HDC;
  OldBitmap: HBITMAP;
  I, J: Integer;
  XX, YY: Integer;
  BS, HBS: Integer;
  Dir: Byte;
  S: Integer;
  LastBrushStyle: TBrushStyle;
begin
  GetImageRect(X, Y, W, H);
  if FBitmap <> 0 then
  begin
    BmpDC := CreateCompatibleDC(Canvas.Handle);
    OldBitmap := SelectObject(BmpDC, FBitmap);
    StretchBlt(Canvas.Handle, X, Y, W, H, BmpDC, 0, 0, FBitmapWidth, FBitmapHeight, SRCCOPY);
    SelectObject(BmpDC, OldBitmap);
    DeleteDC(BmpDC);
  end;
  if ShowFeatures and (FFeatures <> nil) then
  begin
    if FShowBlockedOrientations then
    begin
      Canvas.Pen.Color := FBOColor;
      Canvas.Pen.Width := Max(1, Round(FZoom / 100.0));
      BS := VF_BLOCK_SIZE * FZoom div 100;
      HBS := BS div 2;
      YY := Y + HBS;
      for I := 0 to FFeatures.BlockedOrientations.Height - 1 do
      begin
        XX := X + HBS;
        for J := 0 to FFeatures.BlockedOrientations.Width - 1 do
        begin
          Dir := FFeatures.BlockedOrientations.Pixels[j, i];
          if Dir <> VFDIR_BACKGROUND then
            if Dir <> VFDIR_UNKNOWN then PaintBO(XX, YY, Dir)
            else begin
              PaintBO(XX, YY, VFDIR_0);
              PaintBO(XX, YY, VFDIR_45);
              PaintBO(XX, YY, VFDIR_90);
              PaintBO(XX, YY, VFDIR_135);
            end;
          Inc(XX, BS);
        end;
        Inc(YY, BS);
      end;
    end;
    LastBrushStyle := Canvas.Brush.Style;
    Canvas.Brush.Style := bsClear;
    if FShowSingularPoints then
    begin
      Canvas.Pen.Color := FSPColor;
      Canvas.Pen.Width := Max(1, Round(FZoom / 100.0));
      S := Max(1, Round(5 * FZoom / 100.0));
      for I := 0 to FFeatures.SingularPoints.Count - 1 do
      begin
        XX := X + FFeatures.SingularPoints[i].X * FZoom div 100;
        YY := Y + FFeatures.SingularPoints[i].Y * FZoom div 100;
        case FFeatures.SingularPoints[i].T of
          vfsptCore: begin
            Canvas.Rectangle(XX - S, YY - S, XX + S + 1, YY + S + 1);
            PaintDir(XX, YY, FFeatures.SingularPoints[i].D, 20);
          end;
          vfsptDoubleCore: begin
            Canvas.Ellipse(XX - S, YY - S, XX + S + 1, YY + S + 1);
          end;
          vfsptDelta: begin
            Canvas.MoveTo(XX - S, YY + S);
            Canvas.LineTo(XX, YY - S);
            Canvas.LineTo(XX + S, YY + S);
            Canvas.LineTo(XX - S, YY + S);
          end;
        end;
      end;
    end;
    Canvas.Pen.Color := FMinutiaColor;
    Canvas.Pen.Width := Max(1, Round(FZoom / 100.0));
    S := Max(1, Round(3 * FZoom / 100.0));
    for I := 0 to FFeatures.Minutiae.Count - 1 do
    begin
      XX := X + FFeatures.Minutiae[i].X * FZoom div 100;
      YY := Y + FFeatures.Minutiae[i].Y * FZoom div 100;
      Canvas.Ellipse(XX - S, YY - S, XX + S + 1, YY + S + 1);
      PaintDir(XX, YY, FFeatures.Minutiae[i].D, 10);
    end;
    Canvas.Brush.Style := LastBrushStyle;
  end;
end;

procedure TVFView.PaintBO(X, Y: Integer; Dir: Byte);
var
  HBS, A: Double;
  DX, DY: Integer;
begin
  HBS := VF_BLOCK_SIZE * FZoom / 100.0 / 2.0;
  A := VFDirToRad(Dir);
  DX := Round(Cos(A) * HBS);
  DY := Round(Sin(A) * HBS);
  Canvas.MoveTo(X, Y);
  Canvas.LineTo(X + DX, Y + DY);
  Canvas.MoveTo(X, Y);
  Canvas.LineTo(X - DX, Y - DY);
end;

procedure TVFView.PaintDir(X, Y: Integer; Dir: Byte; Length: Integer);
var
  A: Double;
  DX, DY: Integer;
begin
  Length := Length * FZoom div 100;
  A := VFDirToRad(Dir);
  DX := Round(Cos(A) * Length);
  DY := Round(Sin(A) * Length);
  Canvas.MoveTo(X, Y);
  Canvas.LineTo(X + DX, Y + DY);
end;

procedure TVFView.SetBOColor(const Value: TColor);
begin
  if FBOColor <> Value then
  begin
    FBOColor := Value;
    Invalidate;
  end;
end;

procedure TVFView.SetColorMax(const Value: TColor);
begin
  if FColorMax <> Value then
  begin
    FColorMax := Value;
    Invalidate;
  end;
end;

procedure TVFView.SetColorMin(const Value: TColor);
begin
  if FColorMin <> Value then
  begin
    FColorMin := Value;
    Invalidate;
  end;
end;

procedure TVFView.SetFeatures(const Value: TVFFeatures);
begin
  FFeatures := Value;
  UpdateFeatures;
end;

procedure TVFView.SetImage(const Value: TVFImage);
begin
  FImage := Value;
  UpdateImage;
end;

procedure TVFView.SetMinutiaColor(const Value: TColor);
begin
  if FMinutiaColor <> Value then
  begin
    FMinutiaColor := Value;
    Invalidate;
  end;
end;

procedure TVFView.SetResolution(Value: Integer);
begin
  if Value < 1 then Value := 1;
  if FResolution <> Value then
  begin
    FResolution := Value;
    Invalidate;
  end;
end;

procedure TVFView.SetShowBlockedOrientations(const Value: Boolean);
begin
  if FShowBlockedOrientations <> Value then
  begin
    FShowBlockedOrientations := Value;
    Invalidate;
  end;
end;

procedure TVFView.SetShowFeatures(const Value: Boolean);
begin
  if FShowFeatures <> Value then
  begin
    FShowFeatures := Value;
    Invalidate;
  end;
end;

procedure TVFView.SetShowSingularPoints(const Value: Boolean);
begin
  if FShowSingularPoints <> Value then
  begin
    FShowSingularPoints := Value;
    Invalidate;
  end;
end;

procedure TVFView.SetSPColor(const Value: TColor);
begin
  if FSPColor <> Value then
  begin
    FSPColor := Value;
    Invalidate;
  end;
end;

procedure TVFView.SetZoom(Value: Integer);
begin
  if Value < MinZoom then Value := MinZoom;
  if Value > MaxZoom then Value := MaxZoom;
  if FZoom <> Value then
  begin
    FZoom := Value;
    Invalidate;
  end;
end;

procedure TVFView.UpdateFeatures;
begin
  Invalidate;
end;

procedure TVFView.UpdateImage;
begin
  FreeBitmap;
  if FImage <> nil then
  begin
    FBitmap := FImage.GetBitmap(FColorMin, FColorMax);
    FBitmapWidth := FImage.Width;
    FBitmapHeight := FImage.Height;
  end;
  Invalidate;
end;

procedure TVFView.WMEraseBkgnd(var Message: TWMEraseBkgnd);
var
  X, Y, W, H: Integer;
  R: TRect;
begin
  Canvas.Brush.Color := Color;
  if Image <> nil then
  begin
    GetImageRect(X, Y, W, H);
    R.Left := 0;
    R.Top := 0;
    R.Right := ClientWidth;
    R.Bottom := Y;
    Canvas.FillRect(R);
    R.Top := R.Bottom;
    Inc(R.Bottom, H);
    R.Right := X;
    Canvas.FillRect(R);
    R.Left := R.Right + W;
    R.Right := ClientWidth;
    Canvas.FillRect(R);
    R.Left := 0;
    R.Top := R.Bottom;
    R.Bottom := ClientHeight;
    Canvas.FillRect(R);
  end
  else begin
    R := ClientRect;
    Canvas.FillRect(R);
  end;
  Message.Result := Ord(True);
end;

procedure TVFView.ZoomIn;
begin
  SetZoom(FZoom * 2);
end;

procedure TVFView.ZoomOut;
begin
  SetZoom(FZoom div 2);
end;

end.

