unit VFImageFormatBMP;

interface

uses Classes, SysUtils, Windows, VFImage;

type
  TVFImageFormatBMP = class(TVFImageFormat)
  protected
    function GetCanLoad: Boolean; override;
    function GetCanSave: Boolean; override;
    function GetFileFilter: string; override;
    function GetFileFilterName: string; override;

    procedure Save(Image: TVFImage; Stream: TStream); override;

  public
    function Load(Image: TVFImage; Filename: string): Boolean; override;

  end;

implementation

var
  ImageFormatBMP: TVFImageFormatBMP;

const
  BitmapFileSignature = $4D42;

{ TVFImageFormatBMP }

function TVFImageFormatBMP.GetCanLoad: Boolean;
begin
  Result := True;
end;

function TVFImageFormatBMP.GetCanSave: Boolean;
begin
  Result := True;
end;

function TVFImageFormatBMP.GetFileFilter: string;
begin
  Result := '*.bmp';
end;

function TVFImageFormatBMP.GetFileFilterName: string;
begin
  Result := 'Bitmap files';
end;

function GetLeftShiftCount(Value: LongWord): Integer;
var
  I: Integer;
begin
  Result := 0;
    for I := 0 to Sizeof(LongWord) * 8 - 1 do
    begin
      if Value and 1 <> 0 then Inc(Result);
      Value := Value shr 1;
    end;
  Result := 8 - Result;
end;

function GetRightShiftCount(Value: LongWord): Integer;
var
  I: Integer;
begin
  for I := 0 to SizeOf(LongWord) * 8 - 1 do
  begin
    if Value and 1 <> 0 then
    begin
      Result := I;
      Exit;
    end;
    Value := Value shr 1;
  end;
  Result := -1;
end;

function TVFImageFormatBMP.Load(Image: TVFImage; Filename: string): Boolean;
var
  Bmp: THandle;
  DS: TDIBSection;
  DC: HDC;
  Pad: Integer;
  BI: TBitmapInfo;
  Buffer: PByte;
  P, PP, Bits: LongWord;
  I, J: Integer;
  B, G, R: Byte;
  RMask, GMask, BMask: LongWord;
  RRight, GRight, BRight: Integer;
  RLeft, GLeft, BLeft: Integer;
  LWColor: LongWord;
  WColor: Word;
begin
  Bmp := LoadImage(0, PChar(Filename), IMAGE_BITMAP, 0, 0,
    LR_DEFAULTCOLOR or LR_CREATEDIBSECTION or LR_DEFAULTSIZE or LR_LOADFROMFILE);
  if Bmp = 0 then
  begin
    Result := False;
    Exit;
  end;
  if GetObject(Bmp, SizeOf(DS), @DS) = 0 then
  begin
    DeleteObject(bmp);
    Result := False;
    Exit;
  end;

  Image.SetSize(DS.dsBm.bmWidth, DS.dsBm.bmHeight);
  P := LongWord(Image.Bits) + LongWord(Image.Height - 1) * LongWord(Image.Width);
  Pad := DS.dsBm.bmWidthBytes - (((ds.dsBmih.biWidth * ds.dsBmih.biBitCount) + 7) div 8);

  case DS.dsBmih.biBitCount of
    1, 4, 8: begin
      DC := GetDC(0);
      Buffer := AllocMem(DS.dsBmih.biWidth * 3 + 4);

      FillChar(BI, SizeOf(BI), 0);
      BI.bmiHeader.biSize := SizeOf(TBitmapInfoHeader);
      BI.bmiHeader.biWidth := DS.dsBmih.biWidth;
      BI.bmiHeader.biHeight := DS.dsBmih.biHeight;
      BI.bmiHeader.biPlanes := 1;
      BI.bmiHeader.biBitCount := 24;
      BI.bmiHeader.biCompression := BI_RGB;

      for I := 0 to DS.dsBmih.biHeight - 1 do
      begin
        GetDIBits(DC, Bmp, I, 1, Buffer, BI, DIB_RGB_COLORS);
        Bits := LongWord(Buffer);
        PP := P;
        for J := 0 to DS.dsBmih.biWidth - 1 do
        begin
          B := PByte(Bits)^; Inc(Bits);
          G := PByte(Bits)^; Inc(Bits);
          R := PByte(Bits)^; Inc(Bits);
          PByte(PP)^ := (B + G + R) div 3;
          Inc(PP);
        end;
        Dec(P, Image.Width);
      end;

      FreeMem(Buffer);
      ReleaseDC(0, DC);
    end;

    16: begin
      if DS.dsBmih.biCompression = BI_BITFIELDS then
      begin
        RMask := DS.dsBitfields[0];
        GMask := DS.dsBitfields[1];
        BMask := DS.dsBitfields[2];
      end
      else begin
        RMask := $7C00;
        GMask := $03E0;
        BMask := $001F;
      end;

      RRight := GetRightShiftCount(RMask);
      GRight := GetRightShiftCount(GMask);
      BRight := GetRightShiftCount(BMask);

      RLeft := GetLeftShiftCount(RMask);
      GLeft := GetLeftShiftCount(GMask);
      BLeft := GetLeftShiftCount(BMask);

      Bits := LongWord(DS.dsBm.bmBits);
      for I := 0 to DS.dsBmih.biHeight - 1 do
      begin
        PP := P;
        for J := 0 to DS.dsBmih.biWidth - 1 do
        begin
          WColor := PWord(Bits)^; Inc(Bits, SizeOf(Word));
          B := Byte(((WColor and Word(BMask)) shr BRight) shl BLeft);
          G := Byte(((WColor and Word(GMask)) shr GRight) shl GLeft);
          R := Byte(((WColor and Word(RMask)) shr RRight) shl RLeft);
          PByte(PP)^ := (B + G + R) div 3;
          Inc(PP);
        end;
        Inc(Bits, Pad);
        Dec(P, Image.Width);
      end;
    end;

    24: begin
      Bits := LongWord(DS.dsBm.bmBits);
      for I := 0 to DS.dsBmih.biHeight - 1 do
      begin
        PP := P;
        for J := 0 to DS.dsBmih.biWidth - 1 do
        begin
          B := PByte(Bits)^; Inc(Bits);
          G := PByte(Bits)^; Inc(Bits);
          R := PByte(Bits)^; Inc(Bits);
          PByte(PP)^ := (B + G + R) div 3;
          Inc(PP);
        end;
        Inc(Bits, Pad);
        Dec(P, Image.Width);
      end;
    end;

    32: begin
      if DS.dsBmih.biCompression = BI_BITFIELDS then
      begin
        RMask := DS.dsBitfields[0];
        GMask := DS.dsBitfields[1];
        BMask := DS.dsBitfields[2];
      end
      else begin
        RMask := $00FF0000;
        GMask := $0000FF00;
        BMask := $000000FF;
      end;

      RRight := GetRightShiftCount(RMask);
      GRight := GetRightShiftCount(GMask);
      BRight := GetRightShiftCount(BMask);

      Bits := LongWord(DS.dsBm.bmBits);
      for I := 0 to DS.dsBmih.biHeight - 1 do
      begin
        PP := P;
        for J := 0 to DS.dsBmih.biWidth - 1 do
        begin
          LWColor := PLongWord(Bits)^; Inc(Bits, SizeOf(LongWord));
          B := Byte((LWColor and BMask) shr BRight);
          G := Byte((LWColor and GMask) shr GRight);
          R := Byte((LWColor and RMask) shr RRight);
          PByte(PP)^ := (B + G + R) div 3;
        end;
        Inc(Bits, Pad);
        Dec(P, Image.Width);
      end;
    end;

  else
    DeleteObject(Bmp);
    Image.Clear;
    Result := False;
    Exit;
  end;

  DeleteObject(Bmp);
  Result := True;
end;

procedure TVFImageFormatBMP.Save(Image: TVFImage; Stream: TStream);
var
  BFH: TBitmapFileHeader;
  BIH: TBitmapInfoHeader;
  WidthBytes: LongWord;
  Color: TRGBQuad;
  I: Integer;
  Bits: LongWord;
begin
  BFH.bfType := BitmapFileSignature;
  BFH.bfReserved1 := 0;
  BFH.bfReserved2 := 0;
  BFH.bfOffBits := SizeOf(BFH) + SizeOf(BIH) + SizeOf(TRGBQuad) * 256;

  BIH.biSize := SizeOf(BIH);
  BIH.biWidth := Image.Width;
  BIH.biHeight := Image.Height;
  BIH.biPlanes := 1;
  BIH.biBitCount := 8;
  BIH.biCompression := BI_RGB;
  WidthBytes := Image.Width div 4;
  if Image.Width mod 4 <> 0 then Inc(WidthBytes);
  WidthBytes := WidthBytes * 4;
  BIH.biSizeImage := WidthBytes * LongWord(Image.Height);
  BIH.biXPelsPerMeter := 0;
  BIH.biYPelsPerMeter := 0;
  BIH.biClrUsed := 0;
  BIH.biClrImportant := 0;

  BFH.bfSize := BFH.bfOffBits + BIH.biSizeImage;

  Stream.WriteBuffer(BFH, SizeOf(BFH));
  Stream.WriteBuffer(BIH, SizeOf(BIH));

  Color.rgbReserved := 0;
  for I := 0 to 255 do
  begin
    Color.rgbBlue := Byte(I);
    Color.rgbGreen := Byte(I);
    Color.rgbRed := Byte(I);
    Stream.WriteBuffer(Color, SizeOf(Color));
  end;

  Dec(WidthBytes, Image.Width);
  Bits := LongWord(Image.Bits) + LongWord(Image.Height - 1) * LongWord(Image.Width);
  for I := 0 to Image.Height - 1 do
  begin
    Stream.WriteBuffer(PByte(Bits)^, image.Width);
    Stream.Seek(WidthBytes, soFromCurrent);
    Dec(Bits, Image.Width);
  end;
end;

initialization
  ImageFormatBMP := TVFImageFormatBMP.Create;

finalization
  ImageFormatBMP.Free;

end.
