/* $Id: fs80.c 8360 2006-08-17 12:36:43Z edgaras $ */
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>

#include "bfpsdk.h"

#include "scanner.h"

//#define DEBUG

#ifdef DEBUG
#define logf(MSG, ...) printf(MSG, ##__VA_ARGS__)
#else
#define logf(MSG, ...) do { } while (0)
#endif

static unsigned char *scanner_read (int *wdth, int *hght);
static int scanner_init (void);
static void scanner_close (void);

struct scanner_info scanner_fs80 =
{
	.dpi 		500,
	.name 		"Futronic FS80",
	.version 	1,
	.close 		&scanner_close,
	.init 		&scanner_init,
	.read 		&scanner_read,
	.timeout	500
};


static unsigned char *img_buf = NULL;
static int   img_buf_size = 0;

static struct bfp_hardware_info  bfp_hw_info;
static struct bfp_software_info  bfp_soft_info;


static int
scanner_init ()
{
	int res;


	res = bfp_init();
	if (res != 0) {
		logf("bfp_init() failed: %d\n", res);
		return -1;
	}

	memset(&bfp_hw_info, 0, sizeof(bfp_hw_info));
	res = bfp_get_hardware_info(&bfp_hw_info, 0);
	if (res != 0)
	{
		logf("bfp_get_hardware_param() failed: %d\n", res);
		return -1;
	}

	/* get image dimensions */
	memset(&bfp_soft_info, 0, sizeof(bfp_soft_info));
	res = bfp_get_software_info(&bfp_soft_info);
	if (res != 0)
	{
		logf("bfp_get_software_info() failed: %d\n", res);
		return -1;
	}

	/* allocate buffer for image */
	img_buf_size = bfp_soft_info.image_size;
	img_buf = malloc(img_buf_size);
	if (img_buf == NULL)
	{
		logf("Unable to allocate image buffer\n");
		img_buf_size = 0;
		return -1;
	}

	return 0;
}

static void scanner_close ()
{
	int res;


	if (img_buf != NULL)
	{
		free(img_buf);
		img_buf = NULL;
		img_buf_size = 0;
	}

	res = bfp_deinit();
	if (res)
	{
		logf("bfp_deinit() failed: %d\n", res);
	}
}


static unsigned char *scanner_read(int *wdth, int *hght)
{
	int res;
	int i;


	if (img_buf == NULL)
	{
		logf("error: scanner not initialized\n");
		return NULL;
	}

	res = bfp_get_image_timeo(img_buf, img_buf_size, (scanner_fs80.timeout * 1.) / 1000, 0);
	if (res)
	{
		logf("bfp_get_image_timeo() failed: %d\n", res);
		return NULL;
	}
	logf("bfp_get_image_timeo() OK\n");

	*wdth = bfp_soft_info.image_width;
	*hght = bfp_soft_info.image_height;

	/* Invert image */
	i = *wdth * *hght - 1;
	for (; i >= 0; i--)
	{
		img_buf[i] = ~(img_buf[i]);
	}

	return img_buf;
}

#if 0
static void
dump_scanner_info(void)
{
	logf("\n"
	       "bfp_hardware_info:\n"
	       "   next_link:	%p\n"
	       "   country:	%u\n"
	       "   manufacturer:	%u\n"
	       "   parameters:	%u\n"
	       "   serial_num_low:	%u\n"
	       "   serial_num_mid:	%u\n"
	       "   serial_num_high:	%u\n"
	       "   major_hardware_version:	%u\n"
	       "   minor_hardware_version:	%u\n"
	       "   major_firmware_version:	%u\n"
	       "   minor_firmware_version:	%u\n"
	       "   x_size_high:	%u\n"
	       "   x_size_low:	%u\n"
	       "   y_size_high:	%u\n"
	       "   y_size_low:	%u\n"
	       "   rotate_image:	%u\n"
	       "   mirror_image:	%u\n"
	       "   x_pixel_size:	%u\n"
	       "   another_light:	%u\n"
	       "   serial_num_option:	%u\n"
	       "   y_pixel_size:	%u\n"
	       "   calibration_option:	%u\n\n",
	       bfp_hw_info.next_link,
	       bfp_hw_info.country,
	       bfp_hw_info.manufacturer,
	       bfp_hw_info.parameters,
	       bfp_hw_info.serial_num_low,
	       bfp_hw_info.serial_num_mid,
	       bfp_hw_info.serial_num_high,
	       bfp_hw_info.major_hardware_version,
	       bfp_hw_info.minor_hardware_version,
	       bfp_hw_info.major_firmware_version,
	       bfp_hw_info.minor_firmware_version,
	       bfp_hw_info.x_size_high,
	       bfp_hw_info.x_size_low,
	       bfp_hw_info.y_size_high,
	       bfp_hw_info.y_size_low,
	       bfp_hw_info.rotate_image,
	       bfp_hw_info.mirror_image,
	       bfp_hw_info.x_pixel_size,
	       bfp_hw_info.another_light,
	       bfp_hw_info.serial_num_option,
	       bfp_hw_info.y_pixel_size,
	       bfp_hw_info.calibration_option);

	logf("bfp_soft_param:\n"
	       "   model_size:	%u\n"
	       "   total_size:	%u\n"
	       "   quality:	%hu\n"
	       "   measure:	%hu\n"
	       "   internal_version:	%s\n"
	       "   image_height:	%u\n"
	       "   image_width:	%u\n"
	       "   image_size:	%u\n"
	       "   samples_num:	%hu\n"
	       "   models_num:	%hu\n"
	       "   max_models:	%hu\n"
	       "   major_software_version:	%u\n"
	       "   minor_software_version:	%u\n\n",
	       bfp_soft_info.model_size,
	       bfp_soft_info.total_size,
	       bfp_soft_info.quality,
	       bfp_soft_info.measure,
	       bfp_soft_info.internal_version,
	       bfp_soft_info.image_height,
	       bfp_soft_info.image_width,
	       bfp_soft_info.image_size,
	       bfp_soft_info.samples_num,
	       bfp_soft_info.models_num,
	       bfp_soft_info.max_models,
	       bfp_soft_info.major_software_version,
	       bfp_soft_info.minor_software_version);
}
#endif


